/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2007 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include "xesam-g-dbussearcher.h"
#include "xesam-g-searcher.h"
#include "xesam-glib-globals.h"
#include "xesam-g-debug-private.h"
#include "xesam-g-utils.h"
#include "xesam-marshal.h"

/**
 * SECTION:xesam-g-dbussearcher
 * @short_description: A #XesamGSearcher talking to a search engine over DBus
 *                     on a Xesam Search 1.0 interface
 *
 * Implementation of a #XesamGSearcher communicating with a search engine over
 * DBus on a Xesam Search 1.0 interface.
 *
 * To connect to the default Xesam Search interface call
 * xesam_g_dbus_searcher_new_default().
 *	 
 * Generally you should only keep one #XesamGDBusSearcher around and use that
 * to spawn you #XesamGSession<!-- -->s on.
 *
 * #XesamGDBusSearcher is available since 1.0
 */

struct _XesamGDBusSearcherPrivate {
	DBusGConnection			*connection;
	DBusGProxy				*proxy;
	gchar					*id;
};

typedef struct {
	gpointer		user_data;
	GCallback		callback;
	XesamGSearcher	*searcher;
	const gchar		*method_name;
	gchar			*prop_name;
} RequestCtx;

enum  {
	XESAM_G_DBUS_SEARCHER_DUMMY_PROPERTY
};

static void xesam_g_dbus_searcher_real_new_session 		(XesamGSearcher		*base,
														 XesamGSearcherGotHandle callback,
														 gpointer			user_data);

static void xesam_g_dbus_searcher_real_close_session	(XesamGSearcher		*base,
														 const gchar			*session_handle,
														 XesamGSearcherVoidResponse callback,
														 gpointer			user_data);

static void xesam_g_dbus_searcher_real_get_property 	(XesamGSearcher		*base,
														 const gchar		*session_handle,
														 const gchar		*prop_name,
														 XesamGSearcherGotProperty callback,
														 gpointer			user_data);

static void xesam_g_dbus_searcher_real_set_property 	(XesamGSearcher		 *base,
														 const gchar		*session_handle,
														 const gchar		*prop_name,
														 const GValue		*value,
														 XesamGSearcherGotProperty callback,
														 gpointer			user_data);

static void xesam_g_dbus_searcher_real_new_search 		(XesamGSearcher		*base,
														 const gchar		*session_handle,
														 const gchar		*query,
														 XesamGSearcherGotHandle callback,
														 gpointer			user_data);

static void xesam_g_dbus_searcher_real_start_search 	(XesamGSearcher		*base,
														 const gchar			*search_handle,
														 XesamGSearcherVoidResponse callback,
														 gpointer			user_data);

static void xesam_g_dbus_searcher_real_close_search 	(XesamGSearcher		*base,
														 const gchar			*search_handle,
														 XesamGSearcherVoidResponse callback,
														 gpointer			user_data);

static void xesam_g_dbus_searcher_real_get_hits			(XesamGSearcher		*base,
														 const gchar			*search_handle,
														 guint 				count,
														 XesamGSearcherGotHits callback,
														 gpointer			user_data);

static void xesam_g_dbus_searcher_real_get_hit_data		(XesamGSearcher		*base,
														 const gchar		*search_handle,
														 GArray				*hit_ids,
														 GStrv				field_names,														 
														 XesamGSearcherGotHits callback,
														 gpointer user_data);

static void xesam_g_dbus_searcher_real_get_hit_count	(XesamGSearcher		*base,
														 const gchar			*search_handle,
														 XesamGSearcherGotHitCount callback,
														 gpointer user_data);

static void xesam_g_dbus_searcher_real_get_state		(XesamGSearcher		*base,
														 XesamGSearcherGotState callback,
														 gpointer			user_data);

static void _register_dbus_signal_marshallers 			(void);

static void	_dispatch_hits_added 					(DBusGProxy 		*proxy,
													 const gchar		*search_id,
													 const guint		count,
													 XesamGDBusSearcher	*self);

static void	_dispatch_hits_removed 					(DBusGProxy 		*proxy,
													 const gchar		*search_id,
													 const GArray		*ids,
													 XesamGDBusSearcher	*self);

static void	_dispatch_hits_modified 				(DBusGProxy 		*proxy,
													 const gchar		*search_id,
													 const GArray		*ids,
													 XesamGDBusSearcher	*self);

static void	_dispatch_search_done					(DBusGProxy 		*proxy,
													 const gchar		*search_id,
													 XesamGDBusSearcher	*self);

static void	_dispatch_state_changed					(DBusGProxy 		*proxy,
													 const GStrv		state_info,
													 XesamGDBusSearcher	*self);


static RequestCtx*
				request_ctx_new						(XesamGSearcher		*searcher,
													 gpointer 			user_data,
													 const gchar		*method_name,
													 GCallback 			callback);

static void
			request_ctx_free						(RequestCtx			*ctx);


static gpointer xesam_g_dbus_searcher_parent_class = NULL;
static XesamGSearcherIface* xesam_g_dbus_searcher_xesam_g_searcher_parent_iface = NULL;

/* Extra types needed. Also defined in dbus-searcher-stub.c - bad, bad, me  */
#define XESAM_TYPE_STRV_ARRAY (dbus_g_type_get_collection ("GPtrArray", G_TYPE_STRV))
#define XESAM_TYPE_VALUE_ARRAY_ARRAY (dbus_g_type_get_collection ("GPtrArray", dbus_g_type_get_collection("GPtrArray", G_TYPE_VALUE)))
#define XESAM_TYPE_UINT_ARRAY (DBUS_TYPE_G_UINT_ARRAY)

/*
 * UTILS
 */

static RequestCtx*
request_ctx_new (XesamGSearcher *searcher,
				 gpointer 		user_data,
				 const gchar 	*method_name,
				 GCallback 		callback)
{
	RequestCtx *ctx;
	
	ctx = g_slice_new0 (RequestCtx);
	ctx->searcher = searcher;
	ctx->user_data = user_data;
	ctx->callback = callback;
	ctx->method_name = method_name;
	
	g_object_ref (searcher);
	
	return ctx;
}

static void
request_ctx_free (RequestCtx *ctx)
{	
	/* Don't free property_name that will be the responsibility of the
	 * one receiving the XesamGSearcherGotProperty */
	
	g_object_unref (ctx->searcher);		
	g_slice_free (RequestCtx, ctx);
}

/*
 * MARSHALLERS
 */

static void
marshal_handle_response (DBusGProxy *proxy, 
						 DBusGProxyCall *call, 
						 gpointer data)
{
	RequestCtx	*ctx;
	GError		*error;
	gchar		*id;
	XesamGSearcherGotHandle response;
	
	ctx = (RequestCtx*) data;
	response = (XesamGSearcherGotHandle) ctx->callback;
	
	error = NULL;
	dbus_g_proxy_end_call (proxy, call, &error, 
						   G_TYPE_STRING, &id,
						   G_TYPE_INVALID);
	
	if (error) {
		g_critical ("Error reading %s response: %s\n",
					ctx->method_name, error->message);
		response (ctx->searcher, NULL, ctx->user_data, error);
		return;
	}
	
	if (!response) {
		xesam_g_debug_object (ctx->searcher,
							  "NULL callback supplied to %s request. Skipping",
							  ctx->method_name);
		g_free (id);
		return;
	}
		
	response (ctx->searcher, id, ctx->user_data, NULL);
}

static void
marshal_property_response (DBusGProxy		*proxy, 
						   DBusGProxyCall	*call, 
						   gpointer			data)
{
	RequestCtx	*ctx;
	GError		*error;
	GValue		*value;
	XesamGSearcherGotProperty response;
	
	ctx = (RequestCtx*) data;
	response = (XesamGSearcherGotProperty) ctx->callback;
	
	value = g_new0 (GValue, 1);
	
	error = NULL;
	dbus_g_proxy_end_call (proxy, call, &error, 
						   G_TYPE_VALUE, value,
						   G_TYPE_INVALID);
	
	if (error) {
		g_critical ("Error reading %s response: %s\n",
					ctx->method_name, error->message);
		g_free (value);
		response (ctx->searcher, ctx->prop_name, NULL, ctx->user_data, error);
		return;
	}
	
	if (!response) {
		xesam_g_debug_object (ctx->searcher,
							  "NULL callback supplied to %s request. Skipping",
							  ctx->method_name);
		g_value_unset (value);
		g_free (value);
		return;
	}
	
	response (ctx->searcher, ctx->prop_name, value, ctx->user_data, NULL);
}

static void
marshal_void_response (DBusGProxy		*proxy, 
					   DBusGProxyCall	*call, 
					   gpointer			data)
{
	RequestCtx	*ctx;
	GError		*error;
	XesamGSearcherVoidResponse response;
	
	ctx = (RequestCtx*) data;
	response = (XesamGSearcherVoidResponse) ctx->callback;
	
	error = NULL;
	dbus_g_proxy_end_call (proxy, call, &error,
						   G_TYPE_INVALID);
	
	if (error) {
		g_critical ("Error reading %s response: %s\n",
					ctx->method_name, error->message);
		response (ctx->searcher, ctx->user_data, error);
		return;
	}
	
	if (!response) {
		xesam_g_debug_object (ctx->searcher, "Skipping NULL callback for "
							                 "%s request", ctx->method_name);
	} else {
		response (ctx->searcher, ctx->user_data, NULL);
	}
}

static void
marshal_hits_response (DBusGProxy		*proxy, 
					   DBusGProxyCall	*call, 
					   gpointer			data)
{
	RequestCtx	*ctx;
	GError		*error;
	XesamGSearcherGotHits response;
	GPtrArray	*hit_data;
	
	ctx = (RequestCtx*) data;
	response = (XesamGSearcherGotHits) ctx->callback;
	
	error = NULL;
	dbus_g_proxy_end_call (proxy, call, &error,
						   XESAM_TYPE_VALUE_ARRAY_ARRAY, &hit_data,
						   G_TYPE_INVALID);
	
	if (error) {
		g_critical ("Error reading %s response: %s\n",
					ctx->method_name, error->message);
		response (ctx->searcher, NULL, ctx->user_data, error);
		return;
	}
	
	if (!response) {
		xesam_g_debug_object (ctx->searcher,
							  "NULL callback supplied to %s request. Skipping",
							  ctx->method_name);
		g_ptr_array_foreach (hit_data, (GFunc)free_ptr_array_of_values, NULL);
		return;
	}
	
	response (ctx->searcher, hit_data, ctx->user_data, NULL);
}

static void
marshal_hit_count_response (DBusGProxy 		*proxy, 
							DBusGProxyCall	*call, 
							gpointer		data)
{
	RequestCtx	*ctx;
	GError		*error;
	guint		count;
	XesamGSearcherGotHitCount response;
	
	ctx = (RequestCtx*) data;
	response = (XesamGSearcherGotHitCount) ctx->callback;
	
	error = NULL;
	dbus_g_proxy_end_call (proxy, call, &error, 
						   G_TYPE_UINT, &count,
						   G_TYPE_INVALID);
	
	if (error) {
		g_critical ("Error reading %s response: %s\n",
					ctx->method_name, error->message);
		response (ctx->searcher, 0, ctx->user_data, error);
	}
	
	if (!response) {
		xesam_g_debug_object (ctx->searcher,
							  "NULL callback supplied to %s request. Skipping",
							  ctx->method_name);
		return;
	}
	
	response (ctx->searcher, count, ctx->user_data, NULL);
}

static void
marshal_state_response (DBusGProxy 		*proxy, 
						DBusGProxyCall	*call, 
						gpointer		data)
{
	RequestCtx	*ctx;
	GError		*error;
	GStrv		state_info;
	XesamGSearcherGotState response;
	
	ctx = (RequestCtx*) data;
	response = (XesamGSearcherGotState) ctx->callback;
	
	if (!response) {
		g_critical ("NULL callback supplied to %s request", ctx->method_name);
		return;
	}
	
	error = NULL;
	dbus_g_proxy_end_call (proxy, call, &error, 
						   G_TYPE_STRV, &state_info,
						   G_TYPE_INVALID);
	
	if (error) {
		g_critical ("Error reading %s response: %s\n",
					ctx->method_name, error->message);
		response (ctx->searcher, NULL, ctx->user_data, error);
		return;
	}
	
	response (ctx->searcher, state_info, ctx->user_data, NULL);
}

/*
 * XESAM_G_SEARCHER METHOD IMPLEMENTATIONS
 */

static void
xesam_g_dbus_searcher_real_new_session (XesamGSearcher	*base,
										XesamGSearcherGotHandle callback,
										gpointer 		user_data)
{
	XesamGDBusSearcher  *self;
	RequestCtx			*ctx;
	
	self = XESAM_G_DBUS_SEARCHER (base);
	ctx = request_ctx_new (base, user_data, XESAM_G_METHOD_NewSession,
						   G_CALLBACK(callback));
	
	dbus_g_proxy_begin_call (self->priv->proxy, XESAM_G_METHOD_NewSession, 
							 marshal_handle_response,
							 ctx, (GDestroyNotify)request_ctx_free,
							 G_TYPE_INVALID);
}

static void
xesam_g_dbus_searcher_real_close_session (XesamGSearcher	*base,
										  const gchar		*session_handle,
										  XesamGSearcherVoidResponse callback,
										  gpointer			user_data)
{
	XesamGDBusSearcher		*self;
	RequestCtx				*ctx;
	
	self = XESAM_G_DBUS_SEARCHER (base);
	g_return_if_fail (session_handle != NULL);
	
	ctx = request_ctx_new (base, user_data, XESAM_G_METHOD_CloseSession,
						   G_CALLBACK(callback));
	
	xesam_g_debug_object (base, "Closing session '%s'", session_handle);
	
	dbus_g_proxy_begin_call (self->priv->proxy, XESAM_G_METHOD_CloseSession, 
							 marshal_void_response,
							 ctx, (GDestroyNotify)request_ctx_free,
							 G_TYPE_STRING, session_handle,
							 G_TYPE_INVALID);
}


static void
xesam_g_dbus_searcher_real_get_property (XesamGSearcher	*base,
										 const gchar	*session_handle,
										 const gchar	*prop_name,
										 XesamGSearcherGotProperty callback,
										 gpointer		user_data)
{
	XesamGDBusSearcher	*self;
	RequestCtx			*ctx;
	
	g_return_if_fail (session_handle != NULL);
	g_return_if_fail (prop_name != NULL);
	g_return_if_fail (callback != NULL);
	
	self = XESAM_G_DBUS_SEARCHER (base);
	ctx = request_ctx_new (base, user_data, XESAM_G_METHOD_GetProperty,
						   G_CALLBACK(callback));
	
	ctx->prop_name = g_strdup (prop_name);
	
	dbus_g_proxy_begin_call (self->priv->proxy, XESAM_G_METHOD_GetProperty, 
							 marshal_property_response,
							 ctx, (GDestroyNotify)request_ctx_free,
							 G_TYPE_STRING, session_handle,
							 G_TYPE_STRING, prop_name,
							 G_TYPE_INVALID);
	
}


static void
xesam_g_dbus_searcher_real_set_property (XesamGSearcher	*base,
										 const gchar	*session_handle,
										 const gchar	*prop_name,
										 const GValue	*value,
										 XesamGSearcherGotProperty callback,
										 gpointer		user_data)
{
	XesamGDBusSearcher	*self;
	RequestCtx			*ctx;
		
	g_return_if_fail (session_handle != NULL);
	g_return_if_fail (prop_name != NULL);
	g_return_if_fail (value != NULL);
	
	self = XESAM_G_DBUS_SEARCHER (base);
	ctx = request_ctx_new (base, user_data, XESAM_G_METHOD_SetProperty,
						   G_CALLBACK(callback));
	
	ctx->prop_name = g_strdup (prop_name);
	
	dbus_g_proxy_begin_call (self->priv->proxy, XESAM_G_METHOD_SetProperty, 
							 marshal_property_response,
							 ctx, (GDestroyNotify)request_ctx_free,
							 G_TYPE_STRING, session_handle,
							 G_TYPE_STRING, prop_name,
							 G_TYPE_VALUE, value,
							 G_TYPE_INVALID);
}

static void
xesam_g_dbus_searcher_real_new_search (XesamGSearcher	*base,
									   const gchar		*session_handle,
									   const gchar		*query,
									   XesamGSearcherGotHandle callback,
									   gpointer			user_data)
{
	XesamGDBusSearcher		*self;
	RequestCtx				*ctx;
	
	self = XESAM_G_DBUS_SEARCHER (base);
	g_return_if_fail (session_handle != NULL);
	g_return_if_fail (query != NULL);
	g_return_if_fail (callback != NULL);
	
	ctx = request_ctx_new (base, user_data, XESAM_G_METHOD_NewSearch,
						   G_CALLBACK(callback));
	
	dbus_g_proxy_begin_call (self->priv->proxy, XESAM_G_METHOD_NewSearch, 
							 marshal_handle_response,
							 ctx, (GDestroyNotify)request_ctx_free,
							 G_TYPE_STRING, session_handle,
							 G_TYPE_STRING, query,
							 G_TYPE_INVALID);
}


static void
xesam_g_dbus_searcher_real_start_search (XesamGSearcher	*base,
										 const gchar		*search_handle,
										 XesamGSearcherVoidResponse callback,
										 gpointer		user_data)
{
	XesamGDBusSearcher	*self;
	RequestCtx			*ctx;
	
	self = XESAM_G_DBUS_SEARCHER (base);
	g_return_if_fail (search_handle != NULL);
	
	ctx = request_ctx_new (base, user_data, XESAM_G_METHOD_StartSearch,
						   G_CALLBACK(callback));
	
	dbus_g_proxy_begin_call (self->priv->proxy, XESAM_G_METHOD_StartSearch, 
							 marshal_void_response,
							 ctx, (GDestroyNotify)request_ctx_free,
							 G_TYPE_STRING, search_handle,
							 G_TYPE_INVALID);
}


static void
xesam_g_dbus_searcher_real_close_search (XesamGSearcher	*base,
										 const gchar		*search_handle,
										 XesamGSearcherVoidResponse callback,
										 gpointer user_data)
{
	XesamGDBusSearcher * self;
	RequestCtx			*ctx;
	
	self = XESAM_G_DBUS_SEARCHER (base);
	g_return_if_fail (search_handle != NULL);
	
	ctx = request_ctx_new (base, user_data, XESAM_G_METHOD_CloseSearch,
						   G_CALLBACK(callback));
	
	dbus_g_proxy_begin_call (self->priv->proxy, XESAM_G_METHOD_CloseSearch, 
							 marshal_void_response,
							 ctx, (GDestroyNotify)request_ctx_free,
							 G_TYPE_STRING, search_handle,
							 G_TYPE_INVALID);
}


static void
xesam_g_dbus_searcher_real_get_hits (XesamGSearcher	*base,
									 const gchar	*search_handle,
									 guint			count,
									 XesamGSearcherGotHits callback,
									 gpointer		user_data)
{
	XesamGDBusSearcher	*self;
	RequestCtx			*ctx;
	
	g_return_if_fail (search_handle != NULL);
	g_return_if_fail (callback != NULL);
	
	self = XESAM_G_DBUS_SEARCHER (base);
	ctx = request_ctx_new (base, user_data, XESAM_G_METHOD_GetHits,
						   G_CALLBACK(callback));
	
	dbus_g_proxy_begin_call (self->priv->proxy, XESAM_G_METHOD_GetHits, 
							 marshal_hits_response,
							 ctx, (GDestroyNotify)request_ctx_free,
							 G_TYPE_STRING, search_handle,
							 G_TYPE_UINT, count,
							 G_TYPE_INVALID);
	
}


static void
xesam_g_dbus_searcher_real_get_hit_data (XesamGSearcher	*base,
										 const gchar	*search_handle,
										 GArray			*hit_ids,
										 GStrv			field_names,										 
										 XesamGSearcherGotHits callback,
										 gpointer 		user_data)
{
	XesamGDBusSearcher	*self;
	RequestCtx			*ctx;
	
	g_return_if_fail (search_handle != NULL);
	g_return_if_fail (field_names != NULL);
	g_return_if_fail (hit_ids != NULL);
	g_return_if_fail (callback != NULL);
	
	self = XESAM_G_DBUS_SEARCHER (base);
	ctx = request_ctx_new (base, user_data, XESAM_G_METHOD_GetHitData,
						   G_CALLBACK(callback));
	
	dbus_g_proxy_begin_call (self->priv->proxy, XESAM_G_METHOD_GetHitData, 
							 marshal_hits_response,
							 ctx, (GDestroyNotify)request_ctx_free,
							 G_TYPE_STRING, search_handle,
							 XESAM_TYPE_UINT_ARRAY, hit_ids,
							 G_TYPE_STRV, field_names,
							 G_TYPE_INVALID);
}


static void
xesam_g_dbus_searcher_real_get_hit_count (XesamGSearcher	*base,
										  const gchar		*search_handle,
										  XesamGSearcherGotHitCount callback,
										  gpointer			user_data)
{
	XesamGDBusSearcher * self;
	RequestCtx			*ctx;
	
	g_return_if_fail (search_handle != NULL);
	g_return_if_fail (callback != NULL);
	
	self = XESAM_G_DBUS_SEARCHER (base);
	ctx = request_ctx_new (base, user_data, XESAM_G_METHOD_GetHitCount,
						   G_CALLBACK(callback));
	
	dbus_g_proxy_begin_call (self->priv->proxy, XESAM_G_METHOD_GetHitCount, 
							 marshal_hit_count_response,
							 ctx, (GDestroyNotify)request_ctx_free,
							 G_TYPE_STRING, search_handle,
							 G_TYPE_INVALID);
}


static void
xesam_g_dbus_searcher_real_get_state (XesamGSearcher	*base,
									  XesamGSearcherGotState callback,
									  gpointer			user_data)
{
	XesamGDBusSearcher * self;
	RequestCtx			*ctx;
	
	g_return_if_fail (XESAM_IS_G_SEARCHER (base));
	g_return_if_fail (callback != NULL);
	
	self = XESAM_G_DBUS_SEARCHER (base);
	ctx = request_ctx_new (base, user_data, XESAM_G_METHOD_GetState,
						   G_CALLBACK(callback));
	
	dbus_g_proxy_begin_call (self->priv->proxy, XESAM_G_METHOD_GetState, 
							 marshal_state_response,
							 ctx, (GDestroyNotify)request_ctx_free,
							 G_TYPE_INVALID);
}

/**
 * xesam_g_dbus_searcher_new_default
 * @returns: A new searcher or %NULL if the connection to the bus fails
 *
 * Create a new #XesamGSearcher talking over the standard Xesam DBus Search
 * interface as defined by the Xesam 1.0 specification.
 */
XesamGDBusSearcher*
xesam_g_dbus_searcher_new_default (void)
{
	XesamGDBusSearcher			*self;
	
	self = xesam_g_dbus_searcher_new (XESAM_G_SERVER_DBUS_NAME,
									  XESAM_G_SERVER_DBUS_OBJECT);
	
	return self;
}

/**
 * xesam_g_dbus_searcher_new
 * @bus_name: The bus name to connect to
 * @object_path: Path of DBus object to look up
 * @returns: A new searcher or %NULL if the connection fails
 *
 * Create a new #XesamGSearcher talking to an object at a given path
 * under a given name. The object at the designated location should expose
 * a standard Xesam 1.0 Search interface.
 */
XesamGDBusSearcher*
xesam_g_dbus_searcher_new (const gchar	*bus_name,
						   const gchar	*object_path)
{
	XesamGDBusSearcher			*self;
	XesamGDBusSearcherPrivate   *priv;
	GError						*error;
	static gboolean				init_dbus_types = TRUE;
	
	/* Ugly hack to initialize dbus-glib specialized type system.
	 * See https://bugs.freedesktop.org/show_bug.cgi?id=13908 */
	if (init_dbus_types) {
		init_dbus_types = FALSE;
		xesam_g_debug ("XesamGDBusSearcher initilizing specialized dbus-glib types");
		DBusGConnection *connection;
		GError 			*error;
		error = NULL;
		connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
		if (!error)
			dbus_g_connection_unref (connection);
		else {
			g_critical ("Unable to connect to session bus: %s", error->message);
			return NULL;
		}
	}
	
	g_return_val_if_fail (bus_name != NULL, NULL);
	g_return_val_if_fail (object_path != NULL, NULL);
	
	self = g_object_newv (XESAM_TYPE_G_DBUS_SEARCHER, 0, NULL);
	priv = self->priv;
	
	xesam_g_debug_object (self, "Creating searcher for bus name:"
						  " '%s' on path '%s'", bus_name, object_path);
	
	/* Create a new connection */
	error = NULL;
	priv->connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (error) {
		g_critical ("Failed to open connection to bus: %s\n", error->message);
		g_error_free (error);
		g_object_unref (self);
		return NULL;
	}
	
	/* Set up the proxy */
	error = NULL;
	priv->proxy = dbus_g_proxy_new_for_name_owner (priv->connection,
												   bus_name,
												   object_path,
												   XESAM_G_SERVER_DBUS_INTERFACE,
												   &error);
	
	if (error) {
		g_critical ("Failed to connect to Xesam Searcher: %s\n", error->message);
		g_error_free (error);
		g_object_unref (self);
		return NULL;
	}
	
	
	dbus_g_proxy_add_signal (priv->proxy, XESAM_G_SIGNAL_HitsAdded, 
							 G_TYPE_STRING, G_TYPE_UINT, G_TYPE_INVALID);
	
	dbus_g_proxy_add_signal (priv->proxy, XESAM_G_SIGNAL_HitsRemoved, 
							 G_TYPE_STRING, DBUS_TYPE_G_UINT_ARRAY, G_TYPE_INVALID);
	
	dbus_g_proxy_add_signal (priv->proxy, XESAM_G_SIGNAL_HitsModified, 
							 G_TYPE_STRING, DBUS_TYPE_G_UINT_ARRAY, G_TYPE_INVALID);
	
	dbus_g_proxy_add_signal (priv->proxy, XESAM_G_SIGNAL_SearchDone, 
							 G_TYPE_STRING, G_TYPE_INVALID);
	
	dbus_g_proxy_add_signal (priv->proxy, XESAM_G_SIGNAL_StateChanged, 
							 G_TYPE_STRV, G_TYPE_INVALID);
	
	dbus_g_proxy_connect_signal (priv->proxy,
								 XESAM_G_SIGNAL_HitsAdded,
								 G_CALLBACK(_dispatch_hits_added),
								 self, NULL);
	
	dbus_g_proxy_connect_signal (priv->proxy,
								 XESAM_G_SIGNAL_HitsRemoved,
								 G_CALLBACK(_dispatch_hits_removed),
								 self, NULL);
	
	dbus_g_proxy_connect_signal (priv->proxy,
								 XESAM_G_SIGNAL_HitsModified,
								 G_CALLBACK(_dispatch_hits_modified),
								 self, NULL);
	
	dbus_g_proxy_connect_signal (priv->proxy,
								 XESAM_G_SIGNAL_SearchDone,
								 G_CALLBACK(_dispatch_search_done),
								 self, NULL);
	
	dbus_g_proxy_connect_signal (priv->proxy,
								 XESAM_G_SIGNAL_StateChanged,
								 G_CALLBACK(_dispatch_state_changed),
								 self, NULL);
	
	return self;
}

GQuark
xesam_g_dbus_error_quark (void)
{
  return g_quark_from_static_string ("xesam-g-dbus-error-quark");
}

static void
xesam_g_dbus_searcher_finalize (GObject *obj)
{
	XesamGDBusSearcher			*self;
	XesamGDBusSearcherPrivate	*priv;
	
	self = XESAM_G_DBUS_SEARCHER (obj);
	priv = self->priv;
	
	xesam_g_debug_object (self, "Finalizing");
	
	/* Free private members */
	g_object_unref (priv->proxy);
	dbus_g_connection_unref (priv->connection);
	
	if (priv->id)
		g_free (priv->id);
	
	/* Free properties */
	
	/* Free private struct */
	g_free (priv);
}

static void
xesam_g_dbus_searcher_class_init (XesamGDBusSearcherClass * klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);
	
	xesam_g_dbus_searcher_parent_class = g_type_class_peek_parent (klass);
	
	object_class->finalize = xesam_g_dbus_searcher_finalize;
	
	_register_dbus_signal_marshallers ();
}


static void
xesam_g_dbus_searcher_xesam_g_searcher_interface_init (XesamGSearcherIface * iface)
{
	xesam_g_dbus_searcher_xesam_g_searcher_parent_iface = 
											g_type_interface_peek_parent (iface);
	iface->new_session	 	= xesam_g_dbus_searcher_real_new_session;
	iface->close_session	= xesam_g_dbus_searcher_real_close_session;
	iface->get_property		= xesam_g_dbus_searcher_real_get_property;
	iface->set_property		= xesam_g_dbus_searcher_real_set_property;
	iface->new_search		= xesam_g_dbus_searcher_real_new_search;
	iface->start_search		= xesam_g_dbus_searcher_real_start_search;
	iface->close_search		= xesam_g_dbus_searcher_real_close_search;
	iface->get_hits			= xesam_g_dbus_searcher_real_get_hits;
	iface->get_hit_data		= xesam_g_dbus_searcher_real_get_hit_data;
	iface->get_hit_count	= xesam_g_dbus_searcher_real_get_hit_count;
	iface->get_state		= xesam_g_dbus_searcher_real_get_state;
}


static void
xesam_g_dbus_searcher_init (XesamGDBusSearcher * self)
{
	XesamGDBusSearcherPrivate	*priv;
	
	priv = g_new0 (XesamGDBusSearcherPrivate, 1);
	self->priv = priv;
}


GType
xesam_g_dbus_searcher_get_type (void)
{
	static GType xesam_g_dbus_searcher_type_id = 0;
	
	if (G_UNLIKELY (xesam_g_dbus_searcher_type_id == 0)) {
		
		static const GTypeInfo g_define_type_info = {
			sizeof (XesamGDBusSearcherClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) xesam_g_dbus_searcher_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (XesamGDBusSearcher),
			0,
			(GInstanceInitFunc) xesam_g_dbus_searcher_init
		};
		
		static const GInterfaceInfo xesam_g_searcher_info = {
			(GInterfaceInitFunc) xesam_g_dbus_searcher_xesam_g_searcher_interface_init,
			(GInterfaceFinalizeFunc) NULL,
			NULL};
		
		xesam_g_dbus_searcher_type_id = g_type_register_static (G_TYPE_OBJECT,
																"XesamGDBusSearcher",
																&g_define_type_info,
																0);
		
		g_type_add_interface_static (xesam_g_dbus_searcher_type_id,
									 XESAM_TYPE_G_SEARCHER,
									 &xesam_g_searcher_info);
	}
	
	return xesam_g_dbus_searcher_type_id;
}

static void
_register_dbus_signal_marshallers (void)
{
	/* Register the Hits{Modified,Removed} signal marshaller */
    dbus_g_object_register_marshaller 
							 (xesam_marshal_VOID__STRING_POINTER,
							  G_TYPE_NONE, G_TYPE_STRING, DBUS_TYPE_G_UINT_ARRAY,
							  G_TYPE_INVALID);
	
	/* Register the HitsAdded signal marshaller */
	dbus_g_object_register_marshaller 
							 (xesam_marshal_VOID__STRING_UINT,
							  G_TYPE_NONE, G_TYPE_STRING, G_TYPE_UINT,
							  G_TYPE_INVALID);
}

static void
_dispatch_hits_added (DBusGProxy 			*proxy,
					  const gchar			*search_id,
					  const guint 			count,
                      XesamGDBusSearcher	*self)
{	
	xesam_g_searcher_hits_added (XESAM_G_SEARCHER(self), search_id, count);
}

static void
_dispatch_hits_removed (DBusGProxy			*proxy,
						const gchar			*search_id,
						const GArray		*ids,
						XesamGDBusSearcher	*self)
{	
	xesam_g_searcher_hits_removed (XESAM_G_SEARCHER(self), search_id, ids);
}

static void
_dispatch_hits_modified (DBusGProxy				*proxy,
						 const gchar			*search_id,
						 const GArray			*ids,
						 XesamGDBusSearcher 	*self)
{	
	xesam_g_searcher_hits_modified (XESAM_G_SEARCHER(self), search_id, ids);
}

static void
_dispatch_search_done (DBusGProxy			*proxy,
					   const gchar			*search_id,
					   XesamGDBusSearcher	*self)
{	
	xesam_g_searcher_search_done (XESAM_G_SEARCHER(self), search_id);
}

static void
_dispatch_state_changed (DBusGProxy				*proxy,
						 const GStrv			state_info,
						 XesamGDBusSearcher 	*self)
{	
	xesam_g_searcher_state_changed (XESAM_G_SEARCHER(self), state_info);
}
