/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2007 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include "xesam-g-hit.h"
#include "xesam-g-search.h"
#include "xesam-g-globals-private.h"
#include "xesam-g-debug-private.h"
#include "xesam-g-utils.h"

static guint			xesam_g_hit_get_field_id_real 	(XesamGHit		*hit,
														 const gchar	*field_name);

static const GValue*	xesam_g_hit_get_field_real		(XesamGHit 		*hit,
														 const gchar	*field_name);

static const gchar*	 	xesam_g_hit_get_string_real		(XesamGHit		*hit,
														 const gchar	*field_name);

static const GValue*	xesam_g_hit_get_field_by_id_real(XesamGHit		*hit,
														 guint			field_id);

static guint			xesam_g_hit_get_id_real			(XesamGHit		*hit);

static gchar**			xesam_g_hit_get_field_names_real(XesamGHit		*hit);

static guint			 xesam_g_hit_num_fields_real	(XesamGHit		*hit);

/**
 * SECTION:xesam-g-hit
 * @short_description: A single hit mapping the metadata requested from the search engine
 *
 * The #XesamGHit class conveniently maps the fields specified in the originating
 * session's #XesamGSession::hit-fields property.
 *
 * #XesamGHit is available since 1.0
 */

G_DEFINE_TYPE (XesamGHit, xesam_g_hit, G_TYPE_OBJECT);

struct _XesamGHitPrivate {
	GPtrArray		*fields;
	GHashTable		*field_map;
	guint			id;
};

static void
xesam_g_hit_init (XesamGHit *hit)
{
	XesamGHitPrivate *priv;
	
	priv = g_new0 (XesamGHitPrivate, 1);
	
	hit->priv = priv;
}

static void
xesam_g_hit_finalize (GObject *object)
{	
	
	XesamGHit 			*hit;
	XesamGHitPrivate 	*priv;
	
	hit = XESAM_G_HIT (object);
	priv = hit->priv;
	
	xesam_g_debug_object (hit, "Finalizing");
	
	/* Free field data which we have assumed responsibility for
	 * from the parent Hits object */
	free_ptr_array_of_values (priv->fields);
	
	g_hash_table_unref (priv->field_map);
	
	g_free (priv);

	G_OBJECT_CLASS (xesam_g_hit_parent_class)->finalize (object);
}

static void
xesam_g_hit_class_init (XesamGHitClass *klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = xesam_g_hit_finalize;
	
	klass->get_field_id = xesam_g_hit_get_field_id_real;
	klass->get_field = xesam_g_hit_get_field_real;
	klass->get_string = xesam_g_hit_get_string_real;
	klass->get_field_by_id = xesam_g_hit_get_field_by_id_real;
	klass->get_id = xesam_g_hit_get_id_real;
	klass->get_field_names = xesam_g_hit_get_field_names_real;
	klass->num_fields = xesam_g_hit_num_fields_real;
}

static guint
xesam_g_hit_get_field_id_real (XesamGHit *hit, const gchar *field_name)
{
	return GPOINTER_TO_UINT (g_hash_table_lookup (hit->priv->field_map, field_name));
}

static const GValue*
xesam_g_hit_get_field_real (XesamGHit *hit, const gchar *field_name)
{
	/* Type checks in dispatcher */
	
	guint field_id;
	
	field_id = xesam_g_hit_get_field_id (hit, field_name);
	
	return xesam_g_hit_get_field_by_id (hit, field_id);
}

static const GValue*
xesam_g_hit_get_field_by_id_real (XesamGHit *hit, guint field_id)
{
	/* Type checks in dispatcher */
	
	g_return_val_if_fail (field_id < hit->priv->fields->len, NULL);
	
	return g_ptr_array_index (hit->priv->fields, field_id);
}

static const gchar*
xesam_g_hit_get_string_real (XesamGHit		*hit,
							 const gchar	*field_name)
{
	const GValue	*value;
	
	/* Type checks in dispatcher */
	
	value = xesam_g_hit_get_field (hit, field_name);
	
	if (value == NULL)
		return NULL;
	
	if (G_VALUE_HOLDS (value, G_TYPE_STRING))
		return g_value_get_string (value);
	
	g_warning ("XesamGHit does not contain a string value as requested");
	return NULL;
}

static guint
xesam_g_hit_get_id_real (XesamGHit *hit)
{
	/* Type checks in dispatcher */
	
	return hit->priv->id;
}

/**
 * xesam_g_hit_new
 * @id: The id of the hit to create
 * @field_map: A map from field_name to array offset for the @field_data
 * @field_data: An array of hit data. The array elements map to the fields
 *              specified in the #XesamGSession:hit-fields property via the
 *              @field_map argument to this method.
 * @returns: A newly constructed hit object
 *
 * Package private method declared in xesam-g-globals-private.h
 *
 * The passed %GPtrArray, @field_data, is henceforth considered owned
 * by the newly created Hit object.
 */
XesamGHit*
xesam_g_hit_new (guint		id, 
				 GHashTable *field_map, 
				 GPtrArray *field_data)
{
	XesamGHit 			*hit;
	XesamGHitPrivate 	*priv;
	
	g_return_val_if_fail (field_map != NULL, NULL);
	g_return_val_if_fail (field_data != NULL, NULL);
	
	/* This check is important since the data comes directly from a
	 * XesamGSearcher which is an untrusted source */
	g_return_val_if_fail (g_hash_table_size(field_map) == field_data->len, NULL);
	
	hit = XESAM_G_HIT (g_object_new(XESAM_TYPE_G_HIT, NULL));
	priv = hit->priv;
	
	priv->id = id;
	priv->fields = field_data;
	priv->field_map = g_hash_table_ref (field_map);
	
	return hit;
}

static void
_append_name (gchar *key, guint value, gchar** data)
{
	// FIXME: We need to check array index bounds!
	data[value] = g_strdup (key);
}

static gchar**
xesam_g_hit_get_field_names_real (XesamGHit	*hit)
{
	guint	size;
	gchar	**field_names;
	
	g_return_val_if_fail (XESAM_IS_G_HIT (hit), NULL);
	
	size = g_hash_table_size (hit->priv->field_map);
	field_names = g_new0 (gchar*, size+1);
	
	g_hash_table_foreach (hit->priv->field_map, (GHFunc)_append_name, field_names);
	
	return field_names;
}

static guint
xesam_g_hit_num_fields_real	(XesamGHit *hit)
{
	return g_hash_table_size (hit->priv->field_map);
}

/**
 * xesam_g_hit_get_field_id
 * @hit: The hit object on which to look up a field id
 * @field_name: The name of the field to look up the id for
 * @returns: An integer id that can be used for fast field lookups with
 *           xesam_g_hit_get_field_by_id()
 *
 * Used to look up a handle to read a given field by. This can be used to speed
 * up field lookups.
 *
 * The implementation guarantees that all #XesamGHit objects read from the same
 * #XesamGHits object must use the same field ids. Thus this handle can be
 * re-used on several hits from the same parent #XesamGHits object.
 * 
 * You can always access field by name with xesam_g_hit_get_field().
 */
guint
xesam_g_hit_get_field_id (XesamGHit	*hit, const gchar *field_name)
{
	g_return_val_if_fail (XESAM_IS_G_HIT (hit), 0);
	g_return_val_if_fail (field_name != NULL, 0);
	
	return XESAM_G_HIT_GET_CLASS (hit)->get_field_id (hit, field_name);
}

/**
 * xesam_g_hit_get_field
 * @hit: The hit object on which to look up a field value
 * @field_name: The fully qualified name of the field as set through the session 
 *              property #XesamGSession:hit-fields.
 * @returns: The value of a named field. If the field is not known
 *           %NULL is returned. See below for details on the returned #GValue.
 *
 * Get the value of a field by name look up. This method has the overhead
 * of a hash table look up. For even faster access use
 * xesam_g_hit_get_field_by_id()  if you know the
 * array index of the field in the session's #XesamGSession:hit-fields property.
 *
 * If you know the data type of the field beforehand you can use the convenience
 * method xesam_g_hit_get_string() to access the field value.
 *
 * <sect3>
 * The returned %GValue will hold a data type as described below. The field data
 * type as defined by the relevant ontology is on the left:
 * <title>Field Value Types</title>
 * <variablelist>
 *   <varlistentry>
 *     <term><emphasis>string</emphasis>&nbsp;</term>
 *     <listitem>%G_TYPE_STRING</listitem>
 *   </varlistentry>
 *   <varlistentry>
 *     <term><emphasis>integer</emphasis>&nbsp;</term>
 *     <listitem>%G_TYPE_INT</listitem>
 *   </varlistentry>
 *   <varlistentry>
 *     <term><emphasis>float</emphasis>&nbsp;</term>
 *     <listitem>%G_TYPE_FLOAT</listitem>
 *   </varlistentry>
 *   <varlistentry>
 *     <term><emphasis>boolean</emphasis>&nbsp;</term>
 *     <listitem>%G_TYPE_BOOLEAN</listitem>
 *   </varlistentry>
 *   <varlistentry>
 *     <term><emphasis>date</emphasis>&nbsp;</term>
 *     <listitem>An ISO-8601 encoded date in a %G_TYPE_STRING</listitem>
 *   </varlistentry>
 *   <varlistentry>
 *     <term><emphasis>uri</emphasis>&nbsp;</term>
 *     <listitem>An URI encoded in a %G_TYPE_STRING</listitem>
 *   </varlistentry>
  *  <varlistentry>
 *     <term><emphasis>list of strings</emphasis>&nbsp;</term>
 *     <listitem>%G_TYPE_STRV</listitem>
 *   </varlistentry>
 *   <varlistentry>
 *     <term><emphasis>list of integers</emphasis>&nbsp;</term>
 *     <listitem>a %GArray holding %gint<!-- -->s</listitem>
 *   </varlistentry>
 *   <varlistentry>
 *     <term><emphasis>list of floats</emphasis>&nbsp;</term>
 *     <listitem>A %GArray holding %gfloat<!-- -->s</listitem>
 *   </varlistentry>
 *   <varlistentry>
 *     <term><emphasis>list of booleans</emphasis>&nbsp;</term>
 *     <listitem>a %GArray holding %gboolean<!-- -->s</listitem>
 *   </varlistentry>
 *   <varlistentry>
 *     <term><emphasis>list of dates</emphasis>&nbsp;</term>
 *     <listitem>A %G_TYPE_STRV with dates formatted as described under the date
 *               entry
 *     </listitem>
 *   </varlistentry>
 *   <varlistentry>
 *     <term><emphasis>list of uris</emphasis>&nbsp;</term>
 *     <listitem>A %G_TYPE_STRV holding the URI strings</listitem>
 *   </varlistentry>
 * </variablelist>
 * </sect3>
 */
const GValue*
xesam_g_hit_get_field (XesamGHit *hit, const gchar *field_name)
{
	g_return_val_if_fail (XESAM_IS_G_HIT (hit), NULL);
	g_return_val_if_fail (field_name != NULL, 0);
	
	return XESAM_G_HIT_GET_CLASS (hit)->get_field (hit, field_name);
}

/**
 * xesam_g_hit_get_field_by_id
 * @hit: The hit object on which to look up a field value
 * @field_id: The offset into the array of field names
 *            provided when requesting the hits. Normally when receiving hits
 *            via the search' #XesamGSearch::hits-ready signal this will be
 *            equal to the offset of the field name into the session's
 *            #XesamGSession:hit-fields property, but if you request hits
 *            by other means (fx. by calling
 *            xesam_g_search_request_extended_data()) this may not be true.
 *            Consult the documentation of the relevant methods in that case.
 * @returns: The value of a field given by its offset in the
 *           session's hit-fields property. See xesam_g_hit_get_field() for
 *           details on the return value
 *
 * Look up a field value based on the offset of the field name in the 
 * sessions #XesamGSession:hit-fields property.
 *
 * To look up a field by name use xesam_g_hit_get_field().
 */
const GValue*
xesam_g_hit_get_field_by_id (XesamGHit *hit, guint field_id)
{
	g_return_val_if_fail (XESAM_IS_G_HIT (hit), NULL);
	
	return XESAM_G_HIT_GET_CLASS (hit)->get_field_by_id (hit, field_id);
}

/**
 * xesam_g_hit_get_string
 * @hit: The hit to look up a string valued field on
 * @returns: The string value of the field. The memory allocated by the string
 *           is private to the hit and should not be freed.
 *
 * Look up the value of a field containing a string. For generic field value
 * look ups use xesam_g_hit_get_field ().
 */
const gchar*
xesam_g_hit_get_string (XesamGHit	*hit,
						const gchar	*field_name)
{	
	g_return_val_if_fail (XESAM_IS_G_HIT (hit), NULL);
	g_return_val_if_fail (field_name != NULL, NULL);
	
	return XESAM_G_HIT_GET_CLASS (hit)->get_string (hit, field_name);
}

/**
 * xesam_g_hit_get_id
 * @hit: The hit to get the id for
 * @returns: The hit id. This is defined as the offset of the hit relative to
 *           the total number of hits retrieved from the server.
 *
 * The Hit id is the offset of the hit within the scope of all results
 * returned by a #XesamGSearch.
 */
guint
xesam_g_hit_get_id (XesamGHit *hit)
{
	g_return_val_if_fail (XESAM_IS_G_HIT(hit), 0);
	
	return XESAM_G_HIT_GET_CLASS (hit)->get_id (hit);
}

/**
 * xesam_g_hit_get_field_names:
 * @hit: The hit object for which to list the field names
 * @returns: A %NULL terminated array of strings 
 *
 * Return a NULL terminated array listing the field names
 * used in this Hit. You must free the array with g_strfreev().
 */
gchar**
xesam_g_hit_get_field_names (XesamGHit	*hit)
{	
	g_return_val_if_fail (XESAM_IS_G_HIT (hit), NULL);
	
	return XESAM_G_HIT_GET_CLASS (hit)->get_field_names (hit);
}

/**
 * xesam_g_hit_num_fields:
 * @hit: The hit object for which to count the number of fields
 * @returns: The number of fields present in the hit
 */
guint
xesam_g_hit_num_fields (XesamGHit *hit)
{
	g_return_val_if_fail (XESAM_IS_G_HIT (hit), NULL);
	
	return XESAM_G_HIT_GET_CLASS (hit)->num_fields (hit);
}
