/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2007 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include "xesam-g-hits.h"
#include "xesam-g-search.h"
#include "xesam-g-globals-private.h"
#include "xesam-g-utils.h"
#include "xesam-g-debug-private.h"

/**
 * SECTION:xesam-g-hits
 * @short_description: Container used to encapsulate a batch of #XesamGHit objects
 *
 * Container for conveniently handling and managing a batch of #XesamGHit objects.
 *
 * #XesamGHits is available since 1.0
 */

struct _XesamGHitsPrivate {
	gboolean		notify_continue;
	XesamGSearch	*search;
	GHashTable		*field_map;
	GPtrArray		*data;
	XesamGHit		**hit_instances;
	gpointer		user_data;
	GDestroyNotify	user_data_destroy;
};

/* Property ids */
enum
{
	PROP_0,

	PROP_FIELD_MAP,
	PROP_HITS_DATA,
	PROP_SEARCH,
	PROP_USER_DATA,
	PROP_USER_DATA_DESTROY,
	PROP_NOTIFY_SEARCH,
	
	LAST_PROPERTY
};

/* Private methods */
static void 		_ensure_continue_search 	(XesamGHits	*hits);

static guint		xesam_g_hits_get_count_real	(XesamGHits* hits);

static XesamGHit*	xesam_g_hits_get_real 		(XesamGHits *hits,
												 guint 		offset);

static guint		xesam_g_hits_get_hit_id_real(XesamGHits *hits,
												 guint		offset);

static XesamGHit*	xesam_g_hits_next_real	 	(XesamGHits *hits,
												 XesamGHit	*hit);

G_DEFINE_ABSTRACT_TYPE (XesamGHits, xesam_g_hits, G_TYPE_OBJECT);

static void
xesam_g_hits_init (XesamGHits *hits)
{
	XesamGHitsPrivate	*priv;
	
	priv = g_new0(XesamGHitsPrivate, 1);
	priv->notify_continue = FALSE;
	priv->field_map = NULL;
	priv->data = NULL;
	priv->user_data = NULL;
	priv->user_data_destroy = NULL;
	
	hits->priv = priv;
}

static void
xesam_g_hits_finalize (GObject *object)
{
	XesamGHits			*hits;
	XesamGHitsPrivate	*priv;
	XesamGHit			*hit;
	gint				idx;

	hits = XESAM_G_HITS (object);
	priv = hits->priv;
	
	xesam_g_debug_object (hits, "Finalizing");
	
	g_object_unref (priv->search);
	
	/* Free cached hit objects and free unused hit data structures.
	 * Recall that cached Hit objects assume responsibility for the
	 * sub-GPtrArrays of priv->data. */
	for (idx = 0; idx < xesam_g_hits_get_count (hits); idx++) {
		hit = priv->hit_instances[idx];
		if (hit) {
			if (!G_IS_OBJECT (hit)) {
				g_critical ("While finalizing XesamGHits, child hit object with"
							" id '%d' invalid. It has probably been "
							"unreferenced illegally elsewhere",
							xesam_g_hits_get_hit_id (hits, idx));
			} else {
				g_object_unref (hit);
			}
		} else {			
			free_ptr_array_of_values (g_ptr_array_index(priv->data,idx));
		}
			
	}
	
	/* Now free the structures holding the actual hits */
	g_free (priv->hit_instances);
	g_ptr_array_free (priv->data, FALSE);
	
	/* Free user data */
	if (priv->user_data_destroy) {
		if (priv->user_data != NULL)
			priv->user_data_destroy (priv->user_data);
		else
			g_warning ("When finalizing XesamGHits. Set to destroy user data, "
					   "but user_data is NULL");
	}
	
	g_free (priv);
	
	G_OBJECT_CLASS (xesam_g_hits_parent_class)->finalize (object);
}

static void
xesam_g_hits_set_property  (GObject 		*object, guint 		prop_id, 
							const GValue 	*value,  GParamSpec *pspec)
{	
	XesamGHits	*hits;
	gpointer	val;
	
	g_return_if_fail (XESAM_IS_G_HITS (object));
	
	hits = XESAM_G_HITS (object);
	
	switch (prop_id)
	{
	case PROP_FIELD_MAP:
		g_return_if_fail (hits->priv->field_map == NULL);
		hits->priv->field_map = (GHashTable*) g_value_dup_boxed (value);
		break;
	case PROP_HITS_DATA:
		g_return_if_fail (hits->priv->data == NULL);
		hits->priv->data = (GPtrArray*) g_value_get_pointer (value);
		hits->priv->hit_instances = g_new0(XesamGHit*, hits->priv->data->len);
		break;
	case PROP_SEARCH:
		g_return_if_fail (hits->priv->search == NULL);
		val = g_value_dup_object (value);
		g_return_if_fail (XESAM_IS_G_SEARCH (val));
		hits->priv->search = XESAM_G_SEARCH (val);
		
		/* Detect whether or not to auto-update the result set */
		g_object_get (hits->priv->search, "auto-continue",
					  &hits->priv->notify_continue, NULL);
		break;
	case PROP_USER_DATA:
		g_return_if_fail (hits->priv->user_data == NULL);
		hits->priv->user_data = g_value_get_pointer (value);
		break;
	case PROP_USER_DATA_DESTROY:
		g_return_if_fail (hits->priv->user_data_destroy == NULL);
		hits->priv->user_data_destroy = g_value_get_pointer (value);
		break;
	case PROP_NOTIFY_SEARCH:
		hits->priv->notify_continue = g_value_get_boolean (value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		return;
		break;
	}
}

static void
xesam_g_hits_get_property (GObject	*object, guint		prop_id, 
						   GValue	*value,  GParamSpec	*pspec)
{
	XesamGHits	*hits;
	
	g_return_if_fail (XESAM_IS_G_HITS (object));
	
	hits = XESAM_G_HITS (object);
	
	switch (prop_id)
	{
	case PROP_FIELD_MAP:
		g_value_set_boxed (value, hits->priv->field_map);
		break;
	case PROP_HITS_DATA:
		g_value_set_pointer (value, hits->priv->data);
		break;
	case PROP_SEARCH:
		g_value_set_object (value, hits->priv->search);
		break;
	case PROP_USER_DATA:
		g_value_set_pointer (value, hits->priv->user_data);
		break;
	case PROP_USER_DATA_DESTROY:
		g_value_set_pointer (value, hits->priv->user_data_destroy);
		break;
	case PROP_NOTIFY_SEARCH:
		g_value_set_boolean (value, hits->priv->notify_continue);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		return;
		break;
	}
}

static void
xesam_g_hits_class_init (XesamGHitsClass *klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = xesam_g_hits_finalize;
	object_class->set_property = xesam_g_hits_set_property;
	object_class->get_property = xesam_g_hits_get_property;
	
	klass->get_count = xesam_g_hits_get_count_real;
	klass->get = xesam_g_hits_get_real;
	klass->get_hit_id = xesam_g_hits_get_hit_id_real;
	klass->next = xesam_g_hits_next_real;
	
	/**
	 * XesamGHits:field-map
	 *
	 * A %GHashTable mapping field names to offsets into the %GPtrArray<!-- -->s
	 * contained in #XesamGHits:hit-data.
	 */
	g_object_class_install_property (object_class,
	                                 PROP_FIELD_MAP,
									 g_param_spec_boxed ("field-map",
														 "Field map",
														 "Field name to offset map",
														 G_TYPE_HASH_TABLE,
														 G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE));
	
	/**
	 * XesamGHits:hits-data
	 *
	 * The raw hit data encapsulated as a %GPtrArray of %GPtrArray<!-- -->s.
	 * Each child-%GPtrArray contains %GValue<!-- -->s ordered arcording
	 * to the session property #XesamGSession:hit-fields.
	 */
	g_object_class_install_property (object_class,
	                                 PROP_HITS_DATA,
									 g_param_spec_pointer ("hits-data",
														   "Hits data",
														   "Raw hit data",
														   G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE));
	
	/**
	 * XesamGHits:search
	 *
	 * The #XesamGSearch that spawned this #XesamGHits object
	 */
	g_object_class_install_property (object_class,
	                                 PROP_SEARCH,
									 g_param_spec_object ("search",
														  "Parent search",
														  "Parent search that spawned these hits",
														  XESAM_TYPE_G_SEARCH,
														  G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE));
	
	/**
	 * XesamGHits:user-data:
	 *
	 * User data supplied by the parent #XesamGSearch object. This can either
	 * be set on the search with xesam_g_search_set_user_data() or be
	 * provided via calls like xesam_g_search_request_extended_data().
	 */
	g_object_class_install_property (object_class,
	                                 PROP_USER_DATA,
									 g_param_spec_pointer ("user-data",
														   "User data",
														   "User supplied data",
														   G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE));
	
	/**
	 * XesamGHits:user-data-destroy:
	 *
	 * A %GDestroyNotify used to free any user_data associated with this
	 * hits object.
	 */
	g_object_class_install_property (object_class,
	                                 PROP_USER_DATA,
									 g_param_spec_pointer ("user-data-destroy",
														   "User data destroy notification",
														   "Function used to free any associated user data",
														   G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE));
	
	/**
	 * XesamGHits:notify-search
	 *
	 * Whether to call xesam_g_search_continue() when calls to either
	 * xesam_g_hits_get() or xesam_g_hits_next() are detected.
	 */
	g_object_class_install_property (object_class,
	                                 PROP_NOTIFY_SEARCH,
									 g_param_spec_boolean ("notify-search",
														   "Notify search",
														   "Whether to make the search continue hit retrieval on access",
														   FALSE,
														   G_PARAM_READWRITE));
}

static guint
xesam_g_hits_get_count_real (XesamGHits* hits)
{	
	/* Type checking in dispatcher */
	
	g_return_val_if_fail (hits->priv->data != NULL, 0);
	
	return hits->priv->data->len;
}


static XesamGHit*
xesam_g_hits_get_real (XesamGHits *hits, guint offset)
{
	XesamGHit 			*hit;
	XesamGHitsPrivate 	*priv;
	guint				hit_id;
	GPtrArray			*field_data;
	
	priv = hits->priv;
	
	g_return_val_if_fail (priv->field_map != NULL, NULL);
	g_return_val_if_fail (priv->data != NULL, NULL);
	
	if (offset >= xesam_g_hits_get_count (hits)) {
		g_critical ("Off set %u out of bounds, "
					"XesamGHits object only contain %u hits",
					offset, xesam_g_hits_get_count (hits));
		return NULL;
	}
	
	/* Check cache */
	hit = priv->hit_instances[offset];
	if (hit) {
		return hit;
	}
	
	/* Not cached, spawn Hit object on the fly */
	
	/* Allow subclasses to calculate the hit id */
	hit_id = xesam_g_hits_get_hit_id (hits, offset);
	
	field_data = g_ptr_array_index (priv->data, offset);		
	hit = xesam_g_hit_new (hit_id, priv->field_map, field_data);
	
	/* Cache the Hit object */
	hits->priv->hit_instances[offset] = hit;
	
	return hit;
}

static guint
xesam_g_hits_get_hit_id_real (XesamGHits	*hits,
							  guint			offset)
{
	/* Type checks assumed in method dispatch */
	
	g_error ("Not implemented");
	return 0;
}

static XesamGHit*
xesam_g_hits_next_real (XesamGHits	*hits,
						XesamGHit	*hit)
{
	g_error ("Not implemented");
	return 0;
}

/* Make sure that we call xesam_g_search_continue one and only
 * one time per Hits object.
 * All public methods invoke this to make sure we trigger a
 * load of the next batch of pending hits. */
static void
_ensure_continue_search (XesamGHits *hits)
{
	if (hits->priv->notify_continue) {
		xesam_g_debug_object (hits, "Detected interaction. Requesting Search@%p"
							  " to continue hit retrieval", hits->priv->search);
		hits->priv->notify_continue = FALSE;
		xesam_g_search_continue (hits->priv->search);
	}
}

/**
 * xesam_g_hit_get_user_data
 * @hits: The hits object to look up user data on
 *
 * Retrieve any user data associated with this #XesamGHits object.
 *
 * Calling this method will not cause hit retrieval to continue automatically.
 */
gpointer
xesam_g_hit_get_user_data (XesamGHits	*hits) {
	
	g_return_val_if_fail (XESAM_IS_G_HITS (hits), NULL);
	
	return hits->priv->user_data;
}

/**
 * xesam_g_hits_get_count
 * @hits: The hits object to get the hit count on
 * @returns: The number of hits in the #XesamGHits object
 *
 * Calling this method will not cause hit retrieval to continue automatically.
 */
guint
xesam_g_hits_get_count (XesamGHits* hits)
{
	g_return_val_if_fail (XESAM_IS_G_HITS (hits), 0);
	
	return XESAM_G_HITS_GET_CLASS(hits)->get_count (hits);
}

/**
 * xesam_g_hits_get
 * @hits: The #XesamGHits object to retrieve a hit from
 * @offset: Offset in the #XesamGHits of the hit to retrieve
 * @returns: The hit at the specified offset or %NULL if the offset is out of
 *           bounds
 *
 * Get a handle to the n'th #XesamGHit in a #XesamGHits object.
 *
 * #XesamGHit objects are generated lazily and cached. The hit is cached in 
 * the #XesamGHits object for the entire lifetime of the parent hits.
 *
 * If you need a reference to the hit past the lifetime of the #XesamGHits 
 * object's lifetime you need to reference the #XesamGHit object.
 *
 * Calling this method will cause the parent #XesamGSearch object to start
 * reading the next batch of hits from the server, if it hasn't already done so.
 * You can disable this feature by setting the property
 * #XesamGSearch:auto-continue to %FALSE on the parent #XesamGSearch object.
 */
XesamGHit*
xesam_g_hits_get (XesamGHits	*hits,
				  guint			offset)
{
	g_return_val_if_fail (XESAM_IS_G_HITS (hits), NULL);
	
	_ensure_continue_search (hits);
	
	return XESAM_G_HITS_GET_CLASS(hits)->get (hits, offset);
}

/**
 * xesam_g_hits_get_hit_id
 * @hits: The #XesamGHits to inspect
 * @offset: The offset of the hit in the #XesamGHits to inspect the id of
 * @returns: The hit id of the hit at the specified @offset. In case @offset
 *           is out of bounds 0 will be returned and a critical error logged.
 *           Note that checking for 0 is not a reliable error condition as
 *           the very first hit in a search will have id 0.
 *
 * Get the hit id of a hit.
 *
 * Recall that the id of a hit correspond to the sequence number in which
 * it was read from the server.
 *
 * Calling this method will not cause hit retrieval to continue automatically.
 */
guint
xesam_g_hits_get_hit_id (XesamGHits	*hits,
						 guint		offset)
{
	g_return_val_if_fail (XESAM_IS_G_HITS (hits), 0);
	
	return XESAM_G_HITS_GET_CLASS(hits)->get_hit_id (hits, offset);
}

/**
 * xesam_g_hits_next
 * @hits: The #XesamGHits object to read the next hit from
 * @hit: The hit after which the following should be returned. If @hit
 *       is %NULL then the first hit will be returned
 * @returns: The hit immediately following @hit in @hits, or %NULL if @hit
 *           is the last hit in @hits.
 *           Note that for some hit subclasses the next hit may not be the
 *           one with hit id incremented by one.
 *
 * Use this method to conveniently iterate through a #XesamGHits object.
 *
 * Calling this method will cause the parent #XesamGSearch object to start
 * reading the next batch of hits from the server, if it hasn't already done so.
 * You can disable this feature by setting the property
 * #XesamGSearch:auto-continue to %FALSE on the parent #XesamGSearch object.
 *
 * Here is a simple example on how to employ xesam_g_hits_next()
 *
 * |[
 *   static void
 *   print_hit_urls (XesamGHits *hits)
 *   {
 *     XesamGHit	*hit;
 *	   
 *     hit = NULL;
 *     while ((hit = xesam_g_hits_next (hits, hit))) {
 *       g_printf ("%d: %s\n",
 *                 xesam_g_hit_get_id (hit),
 *                 xesam_g_hit_get_string (hit, "xesam:url"));
 *     }
 *   }
 * ]|
 *
 */
XesamGHit*
xesam_g_hits_next (XesamGHits	*hits,
				   XesamGHit	*hit)
{
	g_return_val_if_fail (XESAM_IS_G_HITS (hits), NULL);
	
	_ensure_continue_search (hits);
	
	if (hit != NULL)
		g_return_val_if_fail (XESAM_IS_G_HIT (hit), NULL);
	
	return XESAM_G_HITS_GET_CLASS(hits)->next (hits, hit);
}

/**
 * xesam_g_hits_get_user_data
 * @hits: The #XesamGHits to look up user data on
 *
 * Convenience method to access the #XesamGHits:user-data property
 *
 * Returns: The user data associated with these hits
 */
gpointer
xesam_g_hits_get_user_data	(XesamGHits	*hits)
{
	g_return_val_if_fail (XESAM_IS_G_HITS (hits), NULL);
	
	return hits->priv->user_data;
}
