/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2008 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */


#include "xesam-g-query-builder.h"

/**
 * SECTION:xesam-g-query-builder
 * @short_description: Used to compile a query from the output of a
 *                     #XesamGQueryParser or #XesamGUserSearchParser, or to
 *                     programmatically construct a query
 *
 * #XesamGQueryBuilder is a callback interface invoked from either a
 * #XesamGQueryParser or #XesamGUserSearchParser. Users familiar with
 * a SAX parser should get the idea. Builders are registered on a parser
 * via the method xesam_g_query_parser_add_builder() or
 * xesam_g_user_search_parser_add_builder().
 *
 * As the parser scans through the input it will invoke the relevant callbacks
 * on the builder. The callbacks are guaranteed to be invoked in such a way as
 * to produce a relevant query.
 *
 * It is important to note that this interface does not define any way
 * for consumers to retrieve the compiled query from the builder when the
 * parser is done. Implementors of this interface should add that if needed.
 *
 * #XesamGQueryBuilder is available since 1.0
 */

GQuark
xesam_g_query_builder_error_quark (void)
{
	return g_quark_from_static_string ("xesam_g_query_builder_error-quark");
}

/**
 * xesam_g_query_builder_start_query
 * @self: The builder receiving the event
 * @token: The token defining the query type. This is either
 *         #XESAM_G_QUERY_TOKEN_QUERY or #XESAM_G_QUERY_TOKEN_USER_QUERY
 * @content_cat: The Content category to query or %NULL if unset
 * @source_cat: The Source category to query or %NULL if unset
 * @error: Place to store a errors or %NULL to not receive notification of errors
 *
 * A parser invokes this method when it starts parsing a new query.
 *
 * The builder will reset its internal state when this method is invoked.
 * In other words it is guaranteed that this call will always start building
 * a new query from scratch and not clobber up with another half-finished
 * query inside the builder.
 *
 * Returns: %FALSE on error, %TRUE otherwise
 */
gboolean
xesam_g_query_builder_start_query	(XesamGQueryBuilder		*self,
									 XesamGQueryToken		token,
									 const gchar			*content_cat,
									 const gchar			*source_cat,
									 GError					**error)
{
	return XESAM_G_QUERY_BUILDER_GET_INTERFACE (self)->start_query (self,
																	token,
																	content_cat,
																	source_cat,
																	error);
}

/**
 * xesam_g_query_builder_add_clause
 * @self: The builder receiving the event
 * @token: The token defining the clause type. See
 *         xesam_g_query_token_is_clause_type() for a list of clause type tokens
 * @boost: Boost factor for the clause. This string is parseable as a float.
 *         This paramter may be %NULL in which case a default boost of 1.0
 *         should be assumed
 * @negate: Whether to negate the matching of this clause. %NULL specifies
 *          that this parameter is undefined in which case the default value
 *          %FALSE should be assumed
 * @attr_names: A %NULL-terminated array of attribute names
 * @attr_vals: A %NULL-terminated array of attribute values with indices
 *             corresponding to those of @attr_names
 * @error: Place to store errors or %NULL to not receive notification of errors
 *
 * A parser invokes this method when it encounters a new clause type token.
 * Following this method call fields and values will be added to the clause
 * via xesam_g_query_builder_add_field() and xesam_g_query_builder_add_value().
 * Finally xesam_g_query_builder_close_clause() will be invoked to signal
 * that all info relevant for the clause has been transfered.
 *
 * Returns: %FALSE on error, %TRUE otherwise
 */
gboolean
xesam_g_query_builder_add_clause	(XesamGQueryBuilder			*self,
									 XesamGQueryToken			token,
									 const gchar				*boost,
									 const gboolean				*negate,
									 const gchar				**attr_names,
									 const gchar				**attr_vals,
									 GError						**error)
{
	return XESAM_G_QUERY_BUILDER_GET_INTERFACE (self)->add_clause (self,
																   token,
																   boost,
																   negate,
																   attr_names,
																   attr_vals,
																   error);
}

/**
 * xesam_g_query_builder_add_field
 * @self: The builder receiving the event
 * @token: The token defining the field type. This is either
 *         #XESAM_G_QUERY_TOKEN_FIELD or #XESAM_G_QUERY_TOKEN_FULL_TEXT_FIELDS
 * @name: The name of the field. For example 'xesam:title'. In case
 *        @token is #XESAM_G_QUERY_TOKEN_FULL_TEXT_FIELDS @name will be %NULL
 * @error: Place to store errors or %NULL to not receive notification of errors
 *
 * A parser invokes this method when it encounters a field type token. This
 * will normally happen following a xesam_g_query_builder_add_clause().
 *
 * Following this call either more fields will be added by calling
 * xesam_g_query_builder_add_field() or value specifications will start
 * with invocations to xesam_g_query_builder_add_value().
 *
 * Returns: %FALSE on error, %TRUE otherwise
 */
gboolean
xesam_g_query_builder_add_field	(XesamGQueryBuilder		*self,
								 XesamGQueryToken		token,
								 const char				*name,
								 GError					**error)
{
	return XESAM_G_QUERY_BUILDER_GET_INTERFACE (self)->add_field (self,
																  token,
																  name,
																  error);
}

/**
 * xesam_g_query_builder_add_value
 * @self: The builder receiving the event
 * @token: The token defining the value type. A full list of value types
 *         can be found in xesam_g_query_token_is_value_type()
 * @value: A string containing the specific value to match
 * @error: Place to store errors or %NULL to not receive notification of errors
 *
 * A parser invokes this method when it encounters a value type token. This
 * will normally happen following a xesam_g_query_builder_add_field().
 *
 * Returns: %FALSE on error, %TRUE otherwise
 */
gboolean
xesam_g_query_builder_add_value	(XesamGQueryBuilder		*self,
								 XesamGQueryToken		token,
								 const gchar			*value,
								 const gchar			**attr_names,
								 const gchar			**attr_vals,
								 GError					**error)
{
	return XESAM_G_QUERY_BUILDER_GET_INTERFACE (self)->add_value (self,
																  token,
																  value,
																  attr_names,
																  attr_vals,
																  error);
}

/**
 * xesam_g_query_builder_close_clause
 * @self: The builder receiving the event
 * @error: Place to store errors or %NULL to not receive notification of errors
 *
 * A parser invokes this method when the end of a clause has been reached.
 * Following this either a new clause will be added with
 * xesam_g_query_builder_add_clause() or the query will be closed with
 * xesam_g_query_builder_close_query().
 *
 * Returns: %FALSE on error, %TRUE otherwise
 */
gboolean
xesam_g_query_builder_close_clause	(XesamGQueryBuilder	*self,
									 GError				**error)
{
	return XESAM_G_QUERY_BUILDER_GET_INTERFACE (self)->close_clause (self,
																	 error);
}

/**
 * xesam_g_query_builder_close_query
 * @self: The builder receiving the event
 * @error: Place to store errors or %NULL to not receive notification of errors
 *
 * A parser invokes this method when the current query is fully parsed and
 * no more tokens will be emitted on it.
 *
 * After this invocation the builder should be ready to deliver any feedback
 * (such as a compiled query or other) if it has any output.
 *
 * Following this method the builder will either be finalized or restarted
 * with an invocation of xesam_g_query_builder_start_query().
 *
 * Returns: %FALSE on error, %TRUE otherwise
 */
gboolean
xesam_g_query_builder_close_query	(XesamGQueryBuilder	*self,
									 GError				**error)
{
	return XESAM_G_QUERY_BUILDER_GET_INTERFACE (self)->close_query (self,
																	error);
}

static void
xesam_g_query_builder_base_init (XesamGQueryBuilderIface * iface)
{
	static gboolean initialized = FALSE;
	
	if (!initialized) {
		initialized = TRUE;
	}
}


GType
xesam_g_query_builder_get_type (void)
{
	static GType xesam_g_query_builder_type_id = 0;
	
	if (G_UNLIKELY (xesam_g_query_builder_type_id == 0)) {
		static const GTypeInfo g_define_type_info = {
			sizeof (XesamGQueryBuilderIface),
			(GBaseInitFunc) xesam_g_query_builder_base_init,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) NULL,
			(GClassFinalizeFunc) NULL,
			NULL,
			0,
			0,
			(GInstanceInitFunc) NULL
		};
		
		xesam_g_query_builder_type_id = g_type_register_static (G_TYPE_INTERFACE,
																"XesamGQueryBuilder",
																&g_define_type_info,
																0);
		
		g_type_interface_add_prerequisite (xesam_g_query_builder_type_id,
										   G_TYPE_OBJECT);
	}
	
	return xesam_g_query_builder_type_id;
}
