/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2008 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include "xesam-g-query-parser.h"
#include "xesam-g-query-builder.h"
#include "xesam-g-debug-private.h"

/* Structure used to collect the data related to a value element */
typedef struct {
	XesamGQueryToken		token;
	gchar					**attribute_names;
	gchar					**attribute_values;
	GString					*text;
} ValueContext;

typedef enum {
	NONE_ELEMENT_ID,
	REQUEST_ELEMENT_ID,
	QUERY_ELEMENT_ID,
	CLAUSE_ELEMENT_ID,
	FIELD_ELEMENT_ID,
	VALUE_ELEMENT_ID
} ElementTypeId;

static void	reset_parser			(XesamGQueryParser	*parser);

static ValueContext*
			value_context_new		(XesamGQueryToken		token,
									 const gchar        	**attribute_names,
									 const gchar        	**attribute_values);

static void value_context_destroy	(ValueContext		*ctx,
									 gboolean			free_text);

static void start_xml_element		(GMarkupParseContext *context,
									 const gchar         *element_name,
									 const gchar        **attribute_names,
									 const gchar        **attribute_values,
									 gpointer             user_data,
									 GError             **error);

static void end_xml_element			(GMarkupParseContext *context,
									 const gchar         *element_name,
									 gpointer             user_data,
									 GError             **error);

static void text_xml_element		(GMarkupParseContext *context,
									 const gchar         *text,
									 gsize                text_len,  
									 gpointer             user_data,
									 GError             **error);

static void passthrough_xml_element	(GMarkupParseContext *context,
									 const gchar         *passthrough_text,
									 gsize                text_len,  
									 gpointer             user_data,
									 GError             **error);

static void xml_error				(GMarkupParseContext *context,
									 GError              *error,
									 gpointer             user_data);

enum  {
	XESAM_G_QUERY_PARSER_DUMMY_PROPERTY
};

struct _XesamGQueryParserPrivate {
	GSList					*builders;
	GMarkupParseContext		*parse_ctx;
	gboolean				found_request_element;
	ValueContext			*value_ctx;
};

static GMarkupParser QUERY_PARSER_STRUCT = {
	start_xml_element,
	end_xml_element,
	text_xml_element,
	passthrough_xml_element,
	xml_error
};

#define	QUERY_PARSER &QUERY_PARSER_STRUCT

static gpointer xesam_g_query_parser_parent_class = NULL;

/**
 * SECTION:xesam-g-query-parser
 * @short_description: A general purpose parser for the Xesam Query Language
 *
 * #XesamGQueryParser is a general purpose parser for the Xesam Query Language
 * version 1.0. To actually build a query that is machine understandable a
 * #XesamGQueryBuilder should be implemented by the consumer and added to the
 * parser via xesam_g_query_parser_add_builder().
 *
 * #XesamGQueryParser is available since 1.0
 */

/**
 * xesam_g_query_parser_add_builder
 * @self: The query parser to add a builder to
 * @builder: The query builder to add
 *
 * Add a #XesamGQueryBuilder to be notified when parsing a query.
 * It is possible to add more than one builder to a query parser.
 *
 * As the query parser scans through a query the methods of the
 * builder will be invoked appropriately. Applications should provide
 * their own implementations of #XesamGQueryBuilder<!-- -->s adding some way
 * to retrieve the build query.
 */
void
xesam_g_query_parser_add_builder (XesamGQueryParser		*self,
								  XesamGQueryBuilder	*builder)
{
	g_return_if_fail (XESAM_IS_G_QUERY_PARSER (self));
	g_return_if_fail (XESAM_IS_G_QUERY_BUILDER (builder));
	
	self->priv->builders = g_slist_prepend (self->priv->builders, builder);
	g_object_ref (builder);
}

/**
 * xesam_g_query_parser_parse_string
 * @self: The parser to use
 * @query: A %NULL-terminated string in the Xesam Query Language
 * @error: An optional place to store a #GError or %NULL to ignore errors
 *
 * Parse a %NULL-terminated string. Returns %TRUE if and only if the parsing
 * complete without errors.
 *
 * When this method returns succesfully the query can be retrieved from all
 * #XesamGQueryBuilder<!-- -->s registered on this parser with
 * xesam_g_query_parser_add_builder(). Applications need to add a way to
 * retrieve the parsed query from the builder(s).
 *
 * Returns: %TRUE if and only if the parsing completed without errors
 */
gboolean
xesam_g_query_parser_parse_string (XesamGQueryParser		*self,
								   const char				*query,
								   GError					**error)
{
	XesamGQueryParserPrivate	*priv;
	GError						*local_error;
	gssize						query_size;
	
	g_return_val_if_fail (XESAM_IS_G_QUERY_PARSER (self), FALSE);
	g_return_val_if_fail (query != NULL, FALSE);
	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
	
	xesam_g_debug_object (self, "Parsing string:\n%s", query);
	
	priv = self->priv;
	
	if (priv->parse_ctx != NULL) {
		if (error == NULL) {
			g_critical ("Concurrent parsing detected. Only one query can be "
						"parsed at a time");
			return FALSE;
		}
		g_set_error (error, XESAM_G_QUERY_PARSER_ERROR,
					 XESAM_G_QUERY_PARSER_ERROR_CONCURRENT_PARSE,
					 "Concurrent parsing detected. Only one query can be "
					 "parsed at a time");
		return FALSE;
	}
	
	priv->parse_ctx = g_markup_parse_context_new (QUERY_PARSER,
												  0,
												  self,
												  NULL);
	
	/* Find byte size of the NULL terminated query string */
	query_size = 0;
	while (query[query_size] != '\0')
		query_size++;
	
	local_error = NULL;
	g_markup_parse_context_parse (priv->parse_ctx,
								  query,
								  query_size,
								  &local_error);
	if (local_error) {
		g_propagate_error (error, local_error);
		return FALSE;
	}
	
	g_markup_parse_context_end_parse (priv->parse_ctx,
									  &local_error);
	if (local_error) {
		g_propagate_error (error, local_error);
		return FALSE;
	}
	
	reset_parser (self);
	
	return TRUE;
}


GQuark
xesam_g_query_parser_error_quark (void)
{
	return g_quark_from_static_string ("xesam_g_query_parser_error-quark");
}


XesamGQueryParser*
xesam_g_query_parser_new (void)
{
	XesamGQueryParser * self;
	self = g_object_newv (XESAM_TYPE_G_QUERY_PARSER, 0, NULL);
	return self;
}

static void
xesam_g_query_parser_init (XesamGQueryParser * self)
{
	XesamGQueryParserPrivate *priv;
	
	priv = g_slice_new (XesamGQueryParserPrivate);
	
	priv->parse_ctx = NULL;
	priv->builders = NULL;
	priv->value_ctx = NULL;
	priv->found_request_element = FALSE;
		
	self->priv = priv;
}

static void
xesam_g_query_parser_finalize (GObject *self)
{
	XesamGQueryParserPrivate	*priv;
	XesamGQueryBuilder			*builder;
	GSList						*iter;
	
	xesam_g_debug_object (self, "Finalizing");
	
	priv = XESAM_G_QUERY_PARSER(self)->priv;
	
	for (iter = priv->builders; iter; iter = iter->next) {
		builder = XESAM_G_QUERY_BUILDER (iter->data); /* We don't need this, but it's here for easier debugging */
		g_object_unref (builder);
	}
	
	if (priv->value_ctx) {
		g_debug ("ValueContext not freed on finalization. "
				 "This is a bug in xesam-glib");
		value_context_destroy (priv->value_ctx, TRUE);
	}
	
	if (priv->parse_ctx) {
		/* This can happen on errors */
		g_markup_parse_context_free (priv->parse_ctx);
	}
	
	G_OBJECT_CLASS(xesam_g_query_parser_parent_class)->finalize (self);
}

static void
xesam_g_query_parser_class_init (XesamGQueryParserClass * klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);
	
	xesam_g_query_parser_parent_class = g_type_class_peek_parent (klass);
	
	object_class->finalize = xesam_g_query_parser_finalize;
}

GType
xesam_g_query_parser_get_type (void)
{
	static GType xesam_g_query_parser_type_id = 0;
	
	if (G_UNLIKELY (xesam_g_query_parser_type_id == 0)) {
		static const GTypeInfo g_define_type_info = { 
			sizeof (XesamGQueryParserClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) xesam_g_query_parser_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (XesamGQueryParser),
			0,
			(GInstanceInitFunc) xesam_g_query_parser_init
		};
		
		xesam_g_query_parser_type_id = g_type_register_static (G_TYPE_OBJECT,
															   "XesamGQueryParser",
															   &g_define_type_info,
															   0);
	}
	
	return xesam_g_query_parser_type_id;
}

static void
reset_parser (XesamGQueryParser	*self)
{
	g_return_if_fail (XESAM_IS_G_QUERY_PARSER(self));
	
	self->priv->found_request_element = FALSE;
	
	if (self->priv->parse_ctx != NULL) {
		g_markup_parse_context_free (self->priv->parse_ctx);
		self->priv->parse_ctx = NULL;
	}
	
	if (self->priv->value_ctx) {
		value_context_destroy (self->priv->value_ctx, TRUE);
		self->priv->value_ctx = NULL;
	}
}

static ValueContext*
value_context_new (XesamGQueryToken			token,
				   const gchar 		     	**attribute_names,
				   const gchar				**attribute_values)
{
	ValueContext *ctx;
	
	g_return_val_if_fail (attribute_names != NULL, NULL);
	g_return_val_if_fail (attribute_values != NULL, NULL);
	
	ctx = g_slice_new (ValueContext);
	ctx->token = token;
	
	/* Why is g_strdupv not declared to take a const gchar** ? */
	ctx->attribute_names = g_strdupv ((gchar**)attribute_names);
	ctx->attribute_values = g_strdupv ((gchar**)attribute_values);
	
	ctx->text = g_string_sized_new (512);
	
	//g_debug ("New vctx@%p", ctx);
	
	return ctx;
}

static void
value_context_destroy (ValueContext *ctx, gboolean free_text)
{
	g_return_if_fail (ctx != NULL);
	
	g_debug ("FREE CTX: %p", ctx);
	
	g_strfreev (ctx->attribute_names);
	g_strfreev (ctx->attribute_values);
	g_string_free (ctx->text, free_text);
	g_slice_free (ValueContext, ctx);
}

static void
start_xml_element (GMarkupParseContext	*context,
				   const gchar			*element_name,
				   const gchar			**attribute_names,
				   const gchar			**attribute_values,
				   gpointer				user_data,
				   GError				**error)
{
	XesamGQueryParser	*self;
	XesamGQueryParserPrivate	*priv;
	XesamGQueryToken	tok;
	gint				line_num, col_num;
	gboolean			negate_attr;
	gchar				*attr1 = NULL;
	gchar				*attr2 = NULL;
	
	g_return_if_fail (XESAM_IS_G_QUERY_PARSER (user_data));
	self = XESAM_G_QUERY_PARSER (user_data);
	priv = self->priv;
	
	tok = xesam_g_query_token_parse (element_name);
	
	if (tok == XESAM_G_QUERY_TOKEN_REQUEST)
		self->priv->found_request_element = TRUE;
	
	/* Assert that we've encountered a <request> element in the past or just now */
	if (!self->priv->found_request_element) {
		g_set_error (error, XESAM_G_QUERY_PARSER_ERROR,
					 XESAM_G_QUERY_PARSER_ERROR_MISSING_REQUEST,
					 "Query not enclosed in 'request' element. Found: '%s'",
					 element_name);
		return;
	}
	
	/* If we did not find a known element raise an error */
	if (tok == XESAM_G_QUERY_TOKEN_NONE) {
		g_markup_parse_context_get_position (context, &line_num, &col_num);
		g_set_error (error, XESAM_G_QUERY_PARSER_ERROR,
					 XESAM_G_QUERY_PARSER_ERROR_UNKNOWN_ELEMENT,
					 "Unexpected element '%s' at line %d column %d of query\n",
					 element_name, line_num, col_num);
		return;
	}
	
	/* Invoke callback corresponding to element_id on all builders */
	GSList	*iter;
	XesamGQueryBuilder *builder;
	if (xesam_g_query_token_is_clause_type (tok)) {
		for (iter = self->priv->builders; iter; iter = iter->next) {
			builder = XESAM_G_QUERY_BUILDER (iter->data);
			
			g_markup_collect_attributes (element_name,
										 attribute_names,
										 attribute_values,
										 error,
										 G_MARKUP_COLLECT_STRING | G_MARKUP_COLLECT_OPTIONAL, "boost", &attr1,
										 G_MARKUP_COLLECT_BOOLEAN | G_MARKUP_COLLECT_OPTIONAL, "negate", &negate_attr,
										 G_MARKUP_COLLECT_INVALID);
			
			if (error && *error)
				return;
			
			xesam_g_query_builder_add_clause (builder,
											  tok,
											  attr1,
											  &negate_attr,
											  attribute_names,
											  attribute_values,
											  error);
			if (error && *error)
				return;
		}
	} else if (xesam_g_query_token_is_query_type (tok)) {
		for (iter = self->priv->builders; iter; iter = iter->next) {
			builder = XESAM_G_QUERY_BUILDER (iter->data);
			
			g_markup_collect_attributes (element_name,
										 attribute_names,
										 attribute_values,
										 error,
										 G_MARKUP_COLLECT_STRING | G_MARKUP_COLLECT_OPTIONAL, "content", &attr1,
										 G_MARKUP_COLLECT_STRING | G_MARKUP_COLLECT_OPTIONAL, "source", &attr2,
										 G_MARKUP_COLLECT_INVALID);
			
			if (error && *error)
				return;
			
			xesam_g_query_builder_start_query (builder,
											 tok,
											 attr1,
											 attr2,
											 error);
			if (error && *error)
				return;
		}
	} else if (xesam_g_query_token_is_value_type (tok)) {
		/* The actual value node can not be emitted before we have collected 
		 * all text inside it. Thus builders are not notified here but
		 * in end_xml_element() */
		
		if (self->priv->value_ctx)
			g_warning ("Leak detected. ValueContext of XesamGQueryParser "
					   "not freed");
		
		/* It is important that this is the only place we set the
		 * priv->value_ctx. This is used to detect whether we are
		 * reading a value element */
		self->priv->value_ctx = value_context_new (tok,
												   attribute_names,
												   attribute_values);
	} else if (xesam_g_query_token_is_field_type (tok)) {
		for (iter = self->priv->builders; iter; iter = iter->next) {
			builder = XESAM_G_QUERY_BUILDER (iter->data);
			
			g_markup_collect_attributes (element_name,
										 attribute_names,
										 attribute_values,
										 error,
										 G_MARKUP_COLLECT_STRING, "name", &attr1,
										 G_MARKUP_COLLECT_INVALID);
			
			if (error && *error)
				return;
			
			xesam_g_query_builder_add_field (builder,
											 tok,
											 attr1,
											 error);
			if (error && *error)
				return;
		}
	} else if (tok == XESAM_G_QUERY_TOKEN_REQUEST) {
		/* Fine, do nothing */
	}
	else 
		g_critical ("Unhandled token type when parsing query. This is a bug "
					"in xesam-glib. Got token type '%s' for element '%s'", 
					xesam_g_query_token_get_name(tok), element_name);
}

static void
end_xml_element (GMarkupParseContext *context,
				 const gchar         *element_name,
				 gpointer            user_data,
				 GError              **error)
{
	XesamGQueryParser	*self;
	XesamGQueryBuilder	*builder;
	XesamGQueryToken	token;
	ValueContext		*ctx;
	GSList 				*iter;
	
	g_return_if_fail (XESAM_IS_G_QUERY_PARSER (user_data));
	
	self = XESAM_G_QUERY_PARSER (user_data);
	ctx = self->priv->value_ctx;
	
	if (self->priv->value_ctx) {
		for (iter = self->priv->builders; iter; iter = iter->next) {
				builder = XESAM_G_QUERY_BUILDER (iter->data);
				xesam_g_query_builder_add_value (builder,
												 ctx->token,
												 ctx->text->str,
												 (const gchar**)ctx->attribute_names,
												 (const gchar**)ctx->attribute_values,
												 error);
				if (error && *error) {
					value_context_destroy (ctx, TRUE);
					self->priv->value_ctx = NULL;
					return;
				}
			}
		
		value_context_destroy (ctx, TRUE);
		
		/* NULLing the value_ctx is important. It is used to detect if 
		 * we are parsing a value element */
		self->priv->value_ctx = NULL;
		return;
	}
	
	/* Check if we are closing a clause or query and notify builders if so */
	token = xesam_g_query_token_parse (element_name);
	if (xesam_g_query_token_is_clause_type (token)) {
		for (iter = self->priv->builders; iter; iter = iter->next) {
				builder = XESAM_G_QUERY_BUILDER (iter->data);
				xesam_g_query_builder_close_clause (builder,
													error);
				if (error && *error)
					return;
			}
	} else if (xesam_g_query_token_is_query_type (token)) {
		for (iter = self->priv->builders; iter; iter = iter->next) {
				builder = XESAM_G_QUERY_BUILDER (iter->data);
				xesam_g_query_builder_close_query (builder,
												   error);
				if (error && *error)
					return;
			}
	}
}

static void
text_xml_element (GMarkupParseContext *context,
				  const gchar         *text,
				  gsize                text_len,  
				  gpointer             user_data,
				  GError             **error)
{
	XesamGQueryParser	*self;
	ValueContext		*ctx;
	
	g_return_if_fail (XESAM_IS_G_QUERY_PARSER (user_data));
	self = XESAM_G_QUERY_PARSER (user_data);
	ctx = self->priv->value_ctx;
	
	if (ctx) {
		//g_debug ("add text to vctx@%p", ctx);
		g_string_append_len (ctx->text, text, text_len);
	}
}

static void
passthrough_xml_element	(GMarkupParseContext *context,
						 const gchar         *passthrough_text,
						 gsize                text_len,  
						 gpointer             user_data,
						 GError               **error)
{
	GString *s = g_string_new_len (passthrough_text, text_len);
	g_printf ("PASSTHROUGH: %s", s->str);
	g_string_free (s, TRUE);
}

static void
xml_error (GMarkupParseContext *context,
		   GError              *error,
		   gpointer             user_data)
{
	XesamGQueryParserPrivate *priv;
	
	xesam_g_debug_object (user_data, "Error parsing query: %s", error->message);
	
	g_return_if_fail (XESAM_IS_G_QUERY_PARSER(user_data));
	
	priv = XESAM_G_QUERY_PARSER(user_data)->priv;
	
	if (priv->value_ctx) {
		value_context_destroy (priv->value_ctx, TRUE);
		priv->value_ctx = NULL;
	}
	
}
