/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2008 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <glib.h>
#include "xesam-g-query-token.h"

/**
 * SECTION:xesam-g-query-token
 * @short_description: Enum and utilities for modelling a Xesam query
 *
 * #XesamGQueryToken is an enumeration that can be used to model the elements
 * of a Xesam query.
 *
 * This enumeration has an embedded type hierarchy, such that strings, intergers,
 * booleans, and floats all can be regarded as subtypes of a value type.
 * Likewise the individual clauses defined by the Xesam Query Language
 * are also grouped into collectors and selectors.
 */

/**
 * xesam_g_query_token_get_name
 * @tok: The token to look up the name for
 *
 * Returns: The Xml element name used in the Xesam Query Language version 1.0
 *          for the token. The returned string should not be freed or
 *          manipulatied
 *
 */
const gchar*
xesam_g_query_token_get_name (XesamGQueryToken tok)
{
	switch (tok) {
		case XESAM_G_QUERY_TOKEN_NONE             : return "NONE";
		case XESAM_G_QUERY_TOKEN_REQUEST          : return "request";
		case XESAM_G_QUERY_TOKEN_QUERY            : return "query";
		case XESAM_G_QUERY_TOKEN_USER_QUERY       : return "userQuery";
		case XESAM_G_QUERY_TOKEN_FIELD            : return "field";
		case XESAM_G_QUERY_TOKEN_FULL_TEXT_FIELDS : return "fullTextFields";
		case XESAM_G_QUERY_TOKEN_STRING           : return "string";
		case XESAM_G_QUERY_TOKEN_INTEGER          : return "integer";
		case XESAM_G_QUERY_TOKEN_BOOLEAN          : return "boolean";
		case XESAM_G_QUERY_TOKEN_FLOAT            : return "float";
		case XESAM_G_QUERY_TOKEN_DATE             : return "date";
		case XESAM_G_QUERY_TOKEN_AND              : return "and";
		case XESAM_G_QUERY_TOKEN_OR               : return "or";
		case XESAM_G_QUERY_TOKEN_CONTAINS         : return "contains";
		case XESAM_G_QUERY_TOKEN_EQUALS           : return "equals";
		case XESAM_G_QUERY_TOKEN_LESS_THAN        : return "lessThan";
		case XESAM_G_QUERY_TOKEN_LESS_THAN_EQUALS : return "lessThanEquals";
		case XESAM_G_QUERY_TOKEN_GREATER_THAN     : return "greaterThan";
		case XESAM_G_QUERY_TOKEN_GREATER_THAN_EQUALS : return "greaterThanEquals";
		case XESAM_G_QUERY_TOKEN_STARTS_WITH      : return "startsWith";
		case XESAM_G_QUERY_TOKEN_IN_SET           : return "inSet";
		case XESAM_G_QUERY_TOKEN_FULL_TEXT        : return "fullText";
		case XESAM_G_QUERY_TOKEN_PROXIMITY        : return "proximity";
		case XESAM_G_QUERY_TOKEN_REG_EXP          : return "regExp";
		case XESAM_G_QUERY_TOKEN_CATEGORY         : return "category";
		default:
			g_critical ("Unhandled query_token type '%d' in QUeryToken.get_name()",
						tok);
			return XESAM_G_QUERY_TOKEN_NONE;
	}
}

/**
 * xesam_g_query_token_parse
 * @elt: Name of element from the Xesam Query Language version 1.0
 *
 * Returns: The #XesamGQueryToken corresponding to the element. If @elt
 *          does not correspond to any element in the Xesam Query Language
 *          #XESAM_G_QUERY_TOKEN_NONE is returned
 *
 */
XesamGQueryToken
xesam_g_query_token_parse (const gchar *elt)
{
	switch (elt[0])
	{
		case 'a':
			if (g_str_equal (elt, "and"))
				return XESAM_G_QUERY_TOKEN_AND;
			break;
		case 'b':
			if (g_str_equal (elt, "boolean"))
				return XESAM_G_QUERY_TOKEN_BOOLEAN;
			break;
		case 'c':
			if (g_str_equal (elt, "contains"))
				return XESAM_G_QUERY_TOKEN_CONTAINS;
			else if (g_str_equal (elt, "category"))
				return XESAM_G_QUERY_TOKEN_CATEGORY;
			break;
		case 'd':
			if (g_str_equal (elt, "date"))
				return XESAM_G_QUERY_TOKEN_DATE;
			break;
		case 'e':
			if (g_str_equal (elt, "equals"))
				return XESAM_G_QUERY_TOKEN_EQUALS;
			break;
		case 'f':
			switch (elt[1]) {
				case 'u':
					if (g_str_equal (elt, "fullText"))
						return XESAM_G_QUERY_TOKEN_FULL_TEXT;
					else if (g_str_equal (elt, "fullTextFields"))
						return XESAM_G_QUERY_TOKEN_FULL_TEXT_FIELDS;
					break;
				case 'i' :
					if (g_str_equal (elt, "field"))
						return XESAM_G_QUERY_TOKEN_FIELD;
					break;
				case 'l':
					if (g_str_equal (elt, "float"))
						return XESAM_G_QUERY_TOKEN_FLOAT;
					break;
				default:
					break;
			}
			break;
		case 'g':
			if (g_str_equal (elt, "greaterThanEquals"))
				return XESAM_G_QUERY_TOKEN_GREATER_THAN_EQUALS;
			else if (g_str_equal (elt, "greaterThan"))
				return XESAM_G_QUERY_TOKEN_GREATER_THAN;
			break;
		case 'i':
			if (g_str_equal (elt, "inSet"))
				return XESAM_G_QUERY_TOKEN_IN_SET;
			else if (g_str_equal (elt, "integer"))
				return XESAM_G_QUERY_TOKEN_INTEGER;
			break;
		case 'l':
			if (g_str_equal (elt, "lessThanEquals"))
				return XESAM_G_QUERY_TOKEN_LESS_THAN_EQUALS;
			else if (g_str_equal (elt, "lessThan"))
				return XESAM_G_QUERY_TOKEN_LESS_THAN;
			break;
		case 'o':
			if (g_str_equal (elt, "or"))
				return XESAM_G_QUERY_TOKEN_OR;
			break;
		case 'p':
			if (g_str_equal (elt, "proximity"))
				return XESAM_G_QUERY_TOKEN_PROXIMITY;
			break;
		case 'q':
			if (g_str_equal (elt, "query"))
				return XESAM_G_QUERY_TOKEN_QUERY;
			break;
		case 'r':
			if (g_str_equal (elt, "request"))
				return XESAM_G_QUERY_TOKEN_REQUEST;
			else if (g_str_equal (elt, "regExp"))
				return XESAM_G_QUERY_TOKEN_REG_EXP;
			break;
		case 's':
			if (g_str_equal (elt, "string"))
				return XESAM_G_QUERY_TOKEN_STRING;
			else if (g_str_equal (elt, "startsWith"))
				return XESAM_G_QUERY_TOKEN_STARTS_WITH;
			break;
		case 'u':
			if (g_str_equal (elt, "userQuery"))
				return XESAM_G_QUERY_TOKEN_USER_QUERY;
			break;
		default:
			return XESAM_G_QUERY_TOKEN_NONE;
	}
	
	return XESAM_G_QUERY_TOKEN_NONE;
}
				
/**
 * xesam_g_query_token_is_query_type
 * @tok: The token to inspect
 *
 * Returns: %TRUE if and only if @tok is a #XESAM_G_QUERY_TOKEN_QUERY or
 *          #XESAM_G_QUERY_TOKEN_USER_QUERY
 *
 */
gboolean
xesam_g_query_token_is_query_type (XesamGQueryToken	tok)
{
	return (tok > XESAM_G_QUERY_TOKEN_QUERY_FIRST && tok < XESAM_G_QUERY_TOKEN_QUERY_LAST);
}

/**
 * xesam_g_query_token_is_field_type
 * @tok: The token to inspect
 *
 * Returns: %TRUE if and only if @tok is a #XESAM_G_QUERY_TOKEN_FIELD or
 *          #XESAM_G_QUERY_TOKEN_FULL_TEXT_FIELDS
 *
 */
gboolean
xesam_g_query_token_is_field_type (XesamGQueryToken	tok)
{
	return (tok > XESAM_G_QUERY_TOKEN_FIELD_FIRST && tok < XESAM_G_QUERY_TOKEN_FIELD_LAST);
}

/**
 * xesam_g_query_token_is_value_type
 * @tok: The token to inspect
 *
 * Returns: %TRUE if and only if @tok is a #XESAM_G_QUERY_TOKEN_STRING,
 *          #XESAM_G_QUERY_TOKEN_INTEGER, #XESAM_G_QUERY_TOKEN_BOOLEAN, or
 *          #XESAM_G_QUERY_TOKEN_FLOAT. Note that date and URLs from the
 *          Xesam ontology is treated as strings
 *
 */
gboolean
xesam_g_query_token_is_value_type (XesamGQueryToken	tok)
{
	return (tok > XESAM_G_QUERY_TOKEN_VALUE_FIRST && tok < XESAM_G_QUERY_TOKEN_VALUE_LAST);
}

/**
 * xesam_g_query_token_is_clause_type
 * @tok: The token to inspect
 *
 * Returns: %TRUE if and only if @tok is one of
 * #XESAM_G_QUERY_TOKEN_AND,
 * #XESAM_G_QUERY_TOKEN_OR,
 * #XESAM_G_QUERY_TOKEN_EQUALS,
 * #XESAM_G_QUERY_TOKEN_CONTAINS,
 * #XESAM_G_QUERY_TOKEN_LESS_THAN,
 * #XESAM_G_QUERY_TOKEN_LESS_THAN_EQUALS,
 * #XESAM_G_QUERY_TOKEN_GREATER_THAN,
 * #XESAM_G_QUERY_TOKEN_GREATER_THAN_EQUALS,
 * #XESAM_G_QUERY_TOKEN_STARTS_WITH,
 * #XESAM_G_QUERY_TOKEN_IN_SET,
 * #XESAM_G_QUERY_TOKEN_FULL_TEXT,
 * #XESAM_G_QUERY_TOKEN_PROXIMITY,
 * #XESAM_G_QUERY_TOKEN_REG_EXP, or
 * #XESAM_G_QUERY_TOKEN_CATEGORY,
 *
 */
gboolean
xesam_g_query_token_is_clause_type (XesamGQueryToken	tok)
{
	return (tok > XESAM_G_QUERY_TOKEN_CLAUSE_FIRST && tok < XESAM_G_QUERY_TOKEN_CLAUSE_LAST);
}

/**
 * xesam_g_query_token_is_collector_type
 * @tok: The token to inspect
 *
 * Returns: %TRUE if and only if @tok is one of
 * #XESAM_G_QUERY_TOKEN_AND or
 * #XESAM_G_QUERY_TOKEN_OR
 */
gboolean
xesam_g_query_token_is_collector_type (XesamGQueryToken	tok)
{
	return (tok > XESAM_G_QUERY_TOKEN_COLLECTOR_FIRST && tok < XESAM_G_QUERY_TOKEN_COLLECTOR_LAST);
}

/**
 * xesam_g_query_token_is_selector_type
 * @tok: the token to inspect
 *
 * Returns: %TRUE if and only if @tok is one of
 * #XESAM_G_QUERY_TOKEN_EQUALS,
 * #XESAM_G_QUERY_TOKEN_CONTAINS,
 * #XESAM_G_QUERY_TOKEN_LESS_THAN,
 * #XESAM_G_QUERY_TOKEN_LESS_THAN_EQUALS,
 * #XESAM_G_QUERY_TOKEN_GREATER_THAN,
 * #XESAM_G_QUERY_TOKEN_GREATER_THAN_EQUALS,
 * #XESAM_G_QUERY_TOKEN_STARTS_WITH,
 * #XESAM_G_QUERY_TOKEN_IN_SET,
 * #XESAM_G_QUERY_TOKEN_FULL_TEXT,
 * #XESAM_G_QUERY_TOKEN_PROXIMITY,
 * #XESAM_G_QUERY_TOKEN_REG_EXP, or
 * #XESAM_G_QUERY_TOKEN_CATEGORY,
 *
 */
gboolean
xesam_g_query_token_is_selector_type (XesamGQueryToken	tok)
{
	return (tok > XESAM_G_QUERY_TOKEN_SELECTOR_FIRST && tok < XESAM_G_QUERY_TOKEN_SELECTOR_LAST);
}

/**
 * xesam_g_query_token_is_extension_type
 * @tok: The token to inspect
 *
 * Note that this method will only detect query language extensions that are
 * expressed as clause types. Other extensions to the Xesam Query Language
 * exist as attributes on the string element.
 *
 * You can inspect which query language extensions the search engine supports
 * by reading the session property #XesamGSession:vendor-extensions.
 *
 * Returns: %TRUE if and only if @tok is one of
 * #XESAM_G_QUERY_TOKEN_PROXIMITY,
 * #XESAM_G_QUERY_TOKEN_REG_EXP, or
 * #XESAM_G_QUERY_TOKEN_CATEGORY,
 *
 */
gboolean
xesam_g_query_token_is_extension_type (XesamGQueryToken	tok)
{
	return (tok > XESAM_G_QUERY_TOKEN_EXTENSION_FIRST && tok < XESAM_G_QUERY_TOKEN_EXTENSION_LAST);
}
