/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2007 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <glib.h>
#include <glib/gprintf.h>

#include "xesam-g-query.h"

/**
 * SECTION:xesam-g-query
 * @short_description: Structured representation of a query
 *
 * In Xesam terminology a 'query' is formatted string prepared for the search
 * engine. A 'search' is the opaque representation of a query that has been
 * accepted and is executing on the search engine.
 *
 * #XesamGQuery is available since 1.0
 */

G_DEFINE_TYPE (XesamGQuery, xesam_g_query, G_TYPE_INITIALLY_UNOWNED);

struct _XesamGQueryPrivate {
	gchar 		*xml;
	gboolean 	is_changed; // For future query building api
};

const gchar* TEXT_QUERY_TEMPLATE = \
"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\
<request>\n\
  <userQuery>\n\
    %s\n\
  </userQuery>\n\
</request>\n";

static void
xesam_g_query_init (XesamGQuery *query)
{
	XesamGQueryPrivate *priv;
	
	priv = g_new0(XesamGQueryPrivate, 1);
	priv->xml = NULL;
	priv->is_changed = FALSE;
	
	query->priv = priv;
}

static void
xesam_g_query_finalize (GObject *object)
{
	XesamGQuery 		*query;
	XesamGQueryPrivate 	*priv;

	query = XESAM_G_QUERY (object);
	priv = query->priv;
	
	if (priv->xml)
		g_free (priv->xml);
	
	g_free (priv);
	
	G_OBJECT_CLASS (xesam_g_query_parent_class)->finalize (object);
}

static void
xesam_g_query_class_init (XesamGQueryClass *klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);
	GObjectClass* parent_class = G_OBJECT_CLASS (klass);

	object_class->finalize = xesam_g_query_finalize;
}

/**
 * xesam_g_query_get_xml
 * @query: Query object to get XML representation for
 * @returns: A string encoding the query in the XESAM Query Language
 *
 * Get an XML representation of a query. The XML will be cached until
 * the query is changed. The query string should not be freed or modified.
 */
const gchar*
xesam_g_query_get_xml (XesamGQuery *query)
{
	if (query->priv->xml && !query->priv->is_changed)
		return query->priv->xml;
	
	g_return_val_if_reached (NULL);
}

/* private helper to format user queries - we need this extra method call
 * to get varargs playing correctly */
static void
_write_user_query (gchar **dest, const gchar *format, ...){
	va_list args;
	
	va_start (args, format);
	g_vasprintf (dest, format, args);
	va_end (args);
	
}

/**
 * xesam_g_query_new_from_textf
 * @format: A string in the XESAM User Search Language. The string may contain
 *          printf-like escapes.
 * @Varargs: Values for the format string
 * @returns: A newly allocated query object
 *
 * Create a new query based on a printf-style argument. The formatted
 * result must conform to the XESAM User Search Language Spec.
 *
 * WARNING: Be cautious if you use this method on user supplied input.
 * Any formatting characters should be escaped.
 */
XesamGQuery*
xesam_g_query_new_from_textf (const gchar *format, ...)
{
	XesamGQuery *query;
	gchar		*escaped_search_text, *unescaped_search_text;
	va_list 	args;
	
	query = XESAM_G_QUERY(g_object_new (XESAM_TYPE_G_QUERY, NULL));
	va_start (args,  format);
	g_vasprintf (&(unescaped_search_text), format, args);
	va_end (args);
	
	escaped_search_text = g_markup_escape_text (unescaped_search_text, -1);
	
	_write_user_query (&(query->priv->xml), 
					   TEXT_QUERY_TEMPLATE, escaped_search_text);
	
	g_free (escaped_search_text);
	g_free (unescaped_search_text);
	
	return query;
}

/**
 * xesam_g_query_new_from_text
 * @search_text: A string in the XESAM User Search Langauge
 * @returns: A newly allocated query object
 *
 * Create a new query based on a user supplied string conforming to
 * the XESAM User Search Language Spec.
 *
 * In contrast to #xesam_g_query_new_from_textf this method is safe to use
 * on user supplied input.
 */
XesamGQuery*
xesam_g_query_new_from_text (const gchar *search_text)
{
	XesamGQuery *query;
	gchar		*escaped_text;
	
	query = XESAM_G_QUERY(g_object_new (XESAM_TYPE_G_QUERY, NULL));
	
	escaped_text = g_markup_escape_text (search_text, -1);
	_write_user_query (&(query->priv->xml), TEXT_QUERY_TEMPLATE, escaped_text);
	
	g_free (escaped_text);
	
	return query;
}

/**
 * xesam_g_query_new_from_xml
 * @query_xml: A string encoded in the XESAM Query Languge
 * @returns: A newly allocated query object
 *
 * Create a new query containing the raw xml supplied in the constructor.
 * The xml should conform to the XESAM Query Language specification, although
 * this will not be checked on construction time.
 */
XesamGQuery*
xesam_g_query_new_from_xml (const gchar *query_xml)
{
	return xesam_g_query_new_steal_xml (g_strdup (query_xml));
}

/**
 * xesam_g_query_new_steal_xml
 * @query_xml: A string encoded in the XESAM Query Languge
 * @returns: A newly allocated query object
 *
 * Create a new query containing the raw xml supplied in the constructor.
 * The xml should conform to the XESAM Query Language specification, although
 * this will not be checked on construction time.
 *
 * The query object will assume ownership of @query_xml.
 *
 * This method is especially handy in conjunction with a #XesamGXmlQueryBuilder
 * to avoid the extra string duplication of the XML query returned by
 * xesam_g_xml_query_builder_get_xml(). If not using
 * xesam_g_xml_query_builder_get_query() directly.
 */
XesamGQuery*
xesam_g_query_new_steal_xml (gchar *query_xml)
{
	XesamGQuery *query;
	
	g_return_val_if_fail (query_xml != NULL, NULL);
	
	query = XESAM_G_QUERY(g_object_new (XESAM_TYPE_G_QUERY, NULL));
	query->priv->xml = query_xml;
	
	return query;
}
