/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2007 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include "xesam-g-search.h"
#include "xesam-g-session.h"
#include "xesam-g-query.h"
#include "xesam-g-hits.h"
#include "xesam-g-sparse-hits.h"
#include "xesam-g-sequential-hits.h"
#include "xesam-glib-globals.h"
#include "xesam-g-globals-private.h"
#include "xesam-g-utils.h"
#include "xesam-g-debug-private.h"
#include "xesam-marshal.h"

/**
 * SECTION:xesam-g-search
 * @short_description: Client side representation of a search
 * @see_also: #XesamGSession
 *
 * Searches are spawned from one of the #XesamGSession methods
 * xesam_g_session_new_search() or xesam_g_session_new_search_from_text().
 *
 * To actually launch the search you must call xesam_g_search_start(), but before
 * doing so make sure you connect to any relevant signals on the search. Most
 * importantly #XesamGSearch::hits-ready.
 *
 * For an example on how to employ a #XesamGSearch please consult the example on
 * <link linkend="simple-search">how to do a simple search</link>.
 *
 * #XesamGSearch is available since 1.0
 */

/* IMPLEMENTATION NOTES
 *
 * = Hit Count Invariant =
 * The implementation of the hit retrieval and hit id subscription is based
 * on a handy invariant callen the Hit Count Invariant. A hit can be in one
 * of three states
 *
 * 1) Registered on the server, but not requested yet
 * 2) Requested, but not received (requested meaning that GetHits has been called on its id)
 * 3) Read (ie received from server)
 *
 * We have four variables related to hit counting
 *
 * num_found - Total number registered on the server
 * pending_hits - Number of hits on the server that we have not yet requested
 * in_transit - Number of hits requested, but not received yet (this has no variable assigned to it)
 * num_read - Number of hits read from the server
 *
 * The Hit Count Invariant is now
 *
 * num_found = num_read + in_transit + pending_hits
 *
 *
 * = Hit Id Assignment =
 * The id of a hit is just the serial number in which it was read from the server.
 * We batch hits together in XesamGHits objects. To assign the hit ids to a bunch
 * of asynchronously retrieved hits, we must now the offset of the hit count
 * from when they where requested. This is:
 *
 * offset = num_read + in_transit
 *        = num_found - pending_hits (by HCI)
 *
 *
 *
 *
 */

/* Private structs */
typedef struct _XesamGSearchContext XesamGSearchContext;
typedef struct _XesamGHitRetrievalContext XesamGHitRetrievalContext;

struct _XesamGSearchPrivate {
	XesamGSession 	*session;
	GHashTable		*field_map;
	GHashTable		*extended_field_map;
	gpointer		user_data;
	GDestroyNotify	user_data_destroy;
	gchar			*id;
	GError			*last_error;
	gboolean		started;
	gboolean		closed;
	gboolean		continue_read;
	gboolean		auto_continue;
	gboolean		done;
	guint			num_read;
	guint			num_found;
	guint			max_batch_size;
	guint			pending_hits;
	gulong			on_session_ready_handler;
	gulong			on_ready_close_handler;
	gulong			on_ready_start_handler;
	gulong			on_start_get_extended_handler;
};

struct _XesamGHitRetrievalContext {
	XesamGSearch	*search;
	guint			num_read;
	guint			batch_offset;
	GArray			*hit_ids;
	GHashTable		*field_map;
	gpointer		user_data;
	GDestroyNotify	user_data_destroy;
};

struct _XesamGSearchContext {
	XesamGSearch	*search;
	XesamGQuery		*query;
};

enum
{
	HITS_FOUND,
	HITS_READY,
	HITS_REMOVED,
	HITS_MODIFIED,
	EXTENDED_DATA_READY,
	DONE,
	READY,
	CLOSED,
	STARTED,
	ERROR,

	LAST_SIGNAL
};

/* Property ids */
enum
{
	PROP_0,

	PROP_MAX_BATCH_SIZE,
	PROP_AUTO_CONTINUE,
	
	LAST_PROPERTY
};

static guint g_search_signals[LAST_SIGNAL] = { 0 };

/* Private methods */

static void _get_search_handle			(XesamGSearchContext *search_ctx);

static void _search_handle_received_cb	(XesamGSearcher *searcher,
										 gchar			*search_handle,
										 gpointer		data,
										 GError			*error);

static void	_closed_cb					(XesamGSearcher *searcher,
										 gpointer		data,
										 GError			*error);

static void	_started_cb					(XesamGSearcher *searcher,
										 gpointer		data,
										 GError			*error);

static XesamGSearchContext* 
			_search_ctx_new				(XesamGSearch 	*search,
										 XesamGQuery	*query);

static void _search_ctx_destroy			(XesamGSearchContext *search_ctx);

static XesamGHitRetrievalContext*
			_hit_retrieval_ctx_new		(XesamGSearch	*search,
										 guint			num_read,
										 GArray			*hit_ids,
										 GHashTable		*field_map,
										 gpointer		user_data,
										 GDestroyNotify	user_data_destroy);

static void _hit_retrieval_ctx_destroy	(XesamGHitRetrievalContext *ctx);

static void	dispatch_hit_data_request	(XesamGHitRetrievalContext *ctx);

static gboolean
			_retrieve_hits 				(XesamGSearch 	*search);

static void _retrieve_hits_cb 			(XesamGSearcher *searcher,
										 GPtrArray		*hit_data,
										 gpointer		data,
										 GError			*error);

static void _parent_session_closed		(XesamGSearch	*search);

static void _dispatch_hits_added		(XesamGSearcher *searcher,
										 const gchar	*search_handle,
										 guint			count,
										 gpointer		data);

static void _dispatch_hits_modified		(XesamGSearcher *searcher,
										 const gchar	*search_handle,
										 const GArray   *hit_ids,
										 gpointer		data);

static void _dispatch_hits_removed		(XesamGSearcher *searcher,
										 const gchar	*search_handle,
										 const GArray   *hit_ids,
										 gpointer		data);

static void _dispatch_search_done		(XesamGSearcher *searcher,
										 const gchar	*search_handle,
										 gpointer		data);

static void _set_last_error				(XesamGSearch   *search,
										 GError			*error);

static void _set_new_error				(XesamGSearch   *search,
										 gint			code,
										 const gchar	*format,
										 ...);

G_DEFINE_TYPE (XesamGSearch, xesam_g_search, G_TYPE_OBJECT);

static void
xesam_g_search_init (XesamGSearch *search)
{
	XesamGSearchPrivate *priv;
	
	priv = g_new0 (XesamGSearchPrivate, 1);
	priv->user_data = NULL;
	priv->user_data_destroy = NULL;
	priv->closed = FALSE;
	priv->started = FALSE;
	priv->continue_read = TRUE;
	priv->auto_continue = TRUE;
	priv->done = FALSE;
	priv->max_batch_size = 20;
	priv->field_map = NULL;
	priv->extended_field_map = NULL;
	
	search->priv = priv;
	
	xesam_g_debug_object (search, "Created");
}

static void
xesam_g_search_finalize (GObject *object)
{
	XesamGSearch 		*search;
	XesamGSearchPrivate *priv;
	
	search = XESAM_G_SEARCH(object);
	priv = search->priv;
	
	xesam_g_debug_object (search, "Finalizing");
	
	/* Disconnect signal handlers on session and searcher */
	g_signal_handlers_disconnect_matched (priv->session,
										  G_SIGNAL_MATCH_DATA,
										  0, 0, NULL,  NULL,
										  search);
	
	g_signal_handlers_disconnect_matched (xesam_g_session_get_searcher(priv->session),
										  G_SIGNAL_MATCH_DATA,
										  0, 0, NULL,  NULL,
										  search);
	
	/* Close the search if it isn't already closed */
	if (!priv->closed) {
		xesam_g_debug_object (search,
							  "Search not closed in finalization. Closing now.");
		
		/* Just fire a close request on the searcher, ignoring the response */
		xesam_g_searcher_close_search (xesam_g_session_get_searcher (priv->session),
									   priv->id, NULL, NULL);
	}
	
	if (priv->id)
		g_free (priv->id);
	
	if (priv->last_error)
		g_error_free (priv->last_error);
	
	if (priv->field_map)
		g_hash_table_unref (priv->field_map);
	
	if (priv->extended_field_map)
		g_hash_table_unref (priv->extended_field_map);
	
	/* Release the user data */
	if (priv->user_data_destroy != NULL) {
		if (priv->user_data != NULL)
			priv->user_data_destroy (priv->user_data);
		else
			g_warning ("When finalizing XesamGSearch: Destroy notification for "
					   " user data skipped because user_data is NULL");
	}
	
	g_object_unref (priv->session);
	
	g_free (search->priv);

	G_OBJECT_CLASS (xesam_g_search_parent_class)->finalize (object);
}

static void
xesam_g_search_set_property  (GObject 		*object, guint 		prop_id, 
							  const GValue 	*value,  GParamSpec *pspec)
{	
	XesamGSearch	*search;
	
	g_return_if_fail (XESAM_IS_G_SEARCH (object));
	
	search = XESAM_G_SEARCH (object);
	
	switch (prop_id)
	{
	case PROP_MAX_BATCH_SIZE:
		search->priv->max_batch_size = g_value_get_uint (value);
		break;	
	case PROP_AUTO_CONTINUE:
		search->priv->auto_continue = g_value_get_boolean (value);
		search->priv->continue_read = search->priv->auto_continue;
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		return;
		break;
	}
}

static void
xesam_g_search_get_property (GObject *object, guint prop_id, 
							 GValue *value, GParamSpec *pspec)
{
	XesamGSearch	*search;
	
	g_return_if_fail (XESAM_IS_G_SEARCH (object));
	
	search = XESAM_G_SEARCH (object);
	
	switch (prop_id)
	{
	case PROP_MAX_BATCH_SIZE:
		g_value_set_uint (value, search->priv->max_batch_size);
		break;
	case PROP_AUTO_CONTINUE:
		g_value_set_boolean (value, search->priv->auto_continue);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		return;
		break;
	}
}

static void
xesam_g_search_class_init (XesamGSearchClass *klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = xesam_g_search_finalize;
	object_class->set_property = xesam_g_search_set_property;
	object_class->get_property = xesam_g_search_get_property;	

	/**
	 * XesamGSearch:max-batch-size
	 *
	 * Maximum number of hits per #XesamGHits object, 
	 * batches can still contain less.
	 */
	g_object_class_install_property (object_class,
	                                 PROP_MAX_BATCH_SIZE,
	                                 g_param_spec_uint ("max-batch-size",
														"Max batch size for hits",
														"Maximum number of hits per Hits object, "
														"batches can still contain less",
														0, G_MAXUINT, 20,
														G_PARAM_READWRITE));

	/**
	 * XesamGSearch:auto-continue
	 *
	 * Whether or not to automatically read the next batch of hits when
	 * clients call xesam_g_hits_get() or xesam_g_hits_next() on one of the
	 * spawned #XesamGHits objects.
	 *
	 * Note that multiple calls to said methods will always only trigger the
	 * retrieval of the next hits one single time.
	 */
	g_object_class_install_property (object_class,
	                                 PROP_AUTO_CONTINUE,
	                                 g_param_spec_boolean ("auto-continue",
														   "Auto continue",
														   "Whether or not to automatically start reading "
														   "the next hits",
														   TRUE,
														   G_PARAM_READWRITE));
	
	/**
	 * XesamGSearch::hits-found
	 * @search: The object on which this signal is emitted
	 * @num_found: The number of hits found by the search engine since the last
	 *             time this signal was emitted.
	 *
	 * Emitted when the search engine has found new hits. This signal only
	 * notifies how many it has found. When the actual hit data is ready
	 * #XesamGSearch::hits-ready is emitted.
	 *
	 * To get the total number of found hits call xesam_g_search_get_num_found().
	 */
	g_search_signals[HITS_FOUND] =
		g_signal_new ("hits-found",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__UINT,
		              G_TYPE_NONE, 1,
		              G_TYPE_UINT);
	
	/**
	 * XesamGSearch::hits-ready
	 * @search: The object on which this signal is emitted
	 * @hits: A #XesamGHits object holding all the retrieved #XesamGHit objects
	 *
	 * Emitted when new hits are ready. The #XesamGHits object is fully loaded
	 * with the new #XesamGHit objects stored in memory.
	 *
	 * To limit the number of hits fetched per hits-ready signal use the
	 * #XesamGSearch:max-batch-size property.
	 */
	g_search_signals[HITS_READY] =
		g_signal_new ("hits-ready",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__OBJECT,
		              G_TYPE_NONE, 1,
		              XESAM_TYPE_G_HITS);
	/**
	 * XesamGSearch::hits-removed
	 * @search: The object on which this signal is emitted
	 * @hit_ids: A #GArray holding the %guint ids of the hits that no longer
	 *           match the query.
	 *
	 * Emitted when hits no longer match the given search. 
	 *
	 * On some search engines moving a file may simply emit a
	 * #XesamGSearch::hits-modified,
	 * since this really only changes the xesam:url field. Other search engines
	 * might remove the hit before re-adding it.
	 *
	 * This signal is not used unless the #XesamGSession:search-live property
	 * is %TRUE.
	 */
	g_search_signals[HITS_REMOVED] =
		g_signal_new ("hits-removed",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__POINTER,
		              G_TYPE_NONE, 1,
		              G_TYPE_POINTER);

	/**
	 * XesamGSearch::hits-modified
	 * @search: the object on which this signal is emitted
	 * @hits: A #GArray of guint ids of the hits with modified metadata.
	 *
	 * Emitted when metadata on the hits in question have been modified.
	 *
	 * On some search engines moving a file may simply emit a hits-modified,
	 * since this really only changes the xesam:url field. Other search engines
	 * might remove the hit before re-adding it.
	 *
	 * This signal is not used unless the #XesamGSession:search-live property
	 * is %TRUE.
	 */
	g_search_signals[HITS_MODIFIED] =
		g_signal_new ("hits-modified",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__POINTER,
		              G_TYPE_NONE, 1,
		              G_TYPE_POINTER);
	
	/**
	 * XesamGSearch::done
	 * @search: The object on which this signal is emitted
	 *
	 * Emitted when the search engine has finished searching its index.
	 *
	 * If #XesamGSession:search-live is %TRUE new hits can still be found if new
	 * items are created or changed to match the query.
	 *
	 * Note that xesam_g_search_get_num_found() will give you the total number
	 * of found hits on the search, but xesam_g_search_get_num_read() may not.
	 */
	g_search_signals[DONE] =
		g_signal_new ("done",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__VOID,
		              G_TYPE_NONE, 0);
	
	/**
	 * XesamGSearch::ready
	 * @search: The object on which this signal is emitted
	 *
	 * Emitted when the search has received an id from the server.
	 */
	g_search_signals[READY] =
		g_signal_new ("ready",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__VOID,
		              G_TYPE_NONE, 0);
	
	/**
	 * XesamGSearch::closed
	 * @search: The object on which this signal is emitted
	 *
	 * Emitted when someone calls #xesam_g_search_close. Note that this can
	 * also happen if someone closes the parent #XesamGSession.
	 *
	 * The search will also be closed on object finalization if it isn't already
	 * closed. No signal will be emitted in this case because the object will
	 * be under finalization.
	 */
	g_search_signals[CLOSED] =
		g_signal_new ("closed",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__VOID,
		              G_TYPE_NONE, 0);
	
	/**
	 * XesamGSearch::started
	 *
	 * Emitted when someone calls #xesam_g_search_start
	 */
	g_search_signals[STARTED] =
		g_signal_new ("started",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__VOID,
		              G_TYPE_NONE, 0);
	
	/**
	 * XesamGSearch::error
	 *
	 * Emitted if there is a communication error with the search engine.
	 * The actual error can be looked up with xesam_g_search_get_last_error().
	 */
	g_search_signals[ERROR] =
		g_signal_new ("error",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__VOID,
		              G_TYPE_NONE, 0);
}

/**
 * xesam_g_search_new
 * @session: The session to be parent of the new search
 * @query: The query the search should run
 * @returns: A newly allocated #XesamGSearch 
 *
 * Package private constructor. Use xesam_g_session_new_search 
 * as the public api.
 */
XesamGSearch*
xesam_g_search_new (XesamGSession *session,
					XesamGQuery *query)
{
	XesamGSearch 		*search;
	XesamGSearchContext	*search_ctx;
	XesamGSearcher		*searcher;
	
	g_return_val_if_fail (XESAM_IS_G_SESSION (session), NULL);
	g_return_val_if_fail (XESAM_IS_G_QUERY (query), NULL);
	
	search = XESAM_G_SEARCH (g_object_new (XESAM_TYPE_G_SEARCH, NULL));
	search->priv->session = g_object_ref (session);
	
	searcher = xesam_g_session_get_searcher (session);
	
	/* We need to react if the parent session is closed as we 
	 * are invalid in that case */
	g_signal_connect_swapped (session, "closed", 
							  G_CALLBACK(_parent_session_closed), search);
	
	g_signal_connect (searcher, "hits-added", G_CALLBACK(_dispatch_hits_added), search); 
	g_signal_connect (searcher, "hits-modified", G_CALLBACK(_dispatch_hits_modified), search);
	g_signal_connect (searcher, "hits-removed", G_CALLBACK(_dispatch_hits_removed), search);
	g_signal_connect (searcher, "search-done", G_CALLBACK(_dispatch_search_done), search);
	
	/* search_ctx is freed in _get_search_handle() */
	search_ctx = _search_ctx_new (search, query);
	_get_search_handle (search_ctx);
	
	return search;
}

/**
 * xesam_g_search_new_from_text
 * @session: The parent session of the search
 * @search_text: A string in the XESAM User Search Language
 * @returns: A newly allocated search object
 *
 * Package private constructor. Searches should be created via
 * the session's #xesam_g_session_new_search_from_text.
 */
XesamGSearch*
xesam_g_search_new_from_text (XesamGSession	*session,
							  gchar			*search_text)
{
	XesamGQuery 	*query;
	XesamGSearch 	*search;
	
	query = xesam_g_query_new_from_text (search_text);
	search = xesam_g_search_new (session, query);
	
	/* The search context grabs the floating ref to the query object */
	
	return search;
}

/**
 * xesam_g_search_get_session
 * @search: The search to llok up parent session for
 * @returns: The parent session for the given search
 *
 * Look up the parent session for a search.
 */
XesamGSession*
xesam_g_search_get_session (XesamGSearch *search)
{
	return search->priv->session;
}

/**
 * xesam_g_search_is_ready
 * @search: The search to check for readiness
 *
 * Returns: %TRUE if and only if the search has received a search handle
 * from the server.
 */
gboolean
xesam_g_search_is_ready (XesamGSearch 	*search)
{
	g_return_val_if_fail (XESAM_IS_G_SEARCH (search), FALSE);
	
	return search->priv->id != NULL;
}

/**
 * xesam_g_search_start
 * @search: The search to start
 * @user_data: User data to attach to this search. Any #XesamGHits objects
 * spawned via this search will have this user data available via
 * xesam_g_hits_get_user_data().
 *
 * Instruct the server to start a given search. No hits
 * will be returned before this method has been invoked.
 */
void
xesam_g_search_start (XesamGSearch		*search)
{
	XesamGSearchPrivate *priv;
	gboolean			should_start;
	
	g_return_if_fail (XESAM_IS_G_SEARCH (search));
	
	priv = search->priv;
	
	should_start = !priv->started;
	priv->started = TRUE;
	
	/* If we don't have an id yet, delay until we do */
	if (!xesam_g_search_is_ready (search)) {
		if (priv->on_ready_start_handler != 0) {
			xesam_g_debug_object (search, "Got start request, but there's already"
								  " a start request pending, waiting for search "
								  "handle");
			return;
		}
		
		xesam_g_debug_object (search, "On start_search(), search object not ready."
							          " Defering start until the search is ready");
		/* it is asserted in _search_handle_received_cb that id != NULL, 
		 * so no infinite recursion */
		priv->on_ready_start_handler = g_signal_connect (search, "ready", 
														 G_CALLBACK(xesam_g_search_start),
														 NULL);
		g_object_ref (search);
		return;
	}
	
	/* This is a queued call to start() that was waiting for the search handle.
	 * Disconnect the ready-handler on the search */
	if (priv->on_ready_start_handler != 0) {
		g_signal_handler_disconnect (search, priv->on_ready_start_handler);
		priv->on_ready_start_handler = 0;
		g_object_unref (search);
		should_start = TRUE;
	}
	
	if (!should_start) {
		xesam_g_debug_object (search, "Got start() request but is already "
							  "running");
		return;
	}
	
	xesam_g_debug_object (search, "Sending start request");
	g_object_ref (search);
	xesam_g_searcher_start_search (xesam_g_session_get_searcher (priv->session),
								   priv->id,
								   _started_cb,
								   search);
}

/**
 * xesam_g_search_is_running
 * @search: The search to inspect
 *
 * Check whether xesam_g_search_start() has been called on the #XesamGSearch
 * object.
 *
 * Returns: %TRUE if and only if xesam_g_search_start(), but not
 *          xesam_g_search_close() has been invoked on the search object
 */
gboolean
xesam_g_search_is_running (XesamGSearch *search)
{
	g_return_val_if_fail (XESAM_IS_G_SEARCH (search), FALSE);
	
	return search->priv->started && !search->priv->closed;
}

/**
 * xesam_g_search_close
 * @search: The search to close
 *
 * Close a search. The server is allowed to free all resources
 * related to the search when this method is invoked.
 *
 * A closed search is considered invalid for further use.
 * 
 * There are two ways in which a search can automatically be closed. If the
 * last reference to the search is dropped (the object is finalized), or if
 * the parent session is closed.
 *
 * A search will be marked closed as soon as xesam_g_search_close() is
 * invoked. The #XesamGSearch::closed signal will be emitted when 
 * the close request has been acknowledged by the underlying #XesamGSearcher.
 */
void
xesam_g_search_close (XesamGSearch *search)
{	
	XesamGSearchPrivate		*priv;
	gboolean				should_close;
	
	g_return_if_fail (XESAM_IS_G_SEARCH(search));
	
	priv = search->priv;
	
	should_close = !priv->closed;
	priv->closed = TRUE;
	
	/* If the search is not ready yet, retry when it is */
	if (!xesam_g_search_is_ready(search)) {
		if (priv->on_ready_close_handler) {
			xesam_g_debug_object (search, "Trying to close search '%s', but it "
								  "already has a close request pending",
								  priv->id);
			return;
		}
		
		xesam_g_debug_object (search, "Trying to close non-ready search."
							          "Defering close.");
		/* FIXME: Clean up "ready" handler */
		priv->on_ready_close_handler = g_signal_connect (search, "ready",
														 (GCallback) xesam_g_search_close,
														 NULL);
		g_object_ref (search);
		return;
	}
	
	/* Disconnect ready-handler if this is a delayed call */
	if (priv->on_ready_close_handler != 0) {
		g_signal_handler_disconnect (search, priv->on_ready_close_handler);
		priv->on_ready_close_handler = 0;
		g_object_unref (search);
		should_close = TRUE;
	}
	
	if (!should_close) {
		xesam_g_debug_object (search, "Got close() request, but is already closed");
		return;
	}
	
	xesam_g_debug_object (search, "Closing search %s", search->priv->id);
	g_object_ref (search);
	xesam_g_searcher_close_search (xesam_g_session_get_searcher (search->priv->session),
								   search->priv->id,
								   _closed_cb,
								   search);
}

/**
 * xesam_g_search_is_closed
 * @search: The search to inspect the state of
 *
 * Inspect whether xesam_g_search_close() has been called on a search.
 * A search will be marked closed as soon as xesam_g_search_close() is
 * invoked. The #XesamGSearch::closed signal will be emitted when 
 * the close request has been acknowledged by the underlying #XesamGSearcher.
 *
 * Returns: %TRUE if xesam_g_search_close() has been called on @search.
 */
gboolean
xesam_g_search_is_closed (XesamGSearch 	*search)
{
	g_return_val_if_fail (XESAM_IS_G_SEARCH (search), TRUE);
	
	return search->priv->closed;
}

/**
 * xesam_g_search_is_done
 * @search: The search to inspect the state of
 *
 * Inspect whether #XesamGSearch::done has been emitted on a search.
 * Note that even though the search is done hits might still be added in case
 * the parent session is a live session - ie has the property
 * #XesamGSession:search-live set to %TRUE.
 *
 * Returns: %TRUE if and only id #XesamGSearch::done has been emitted on @search.
 */
gboolean
xesam_g_search_is_done (XesamGSearch 	*search)
{
	g_return_val_if_fail (XESAM_IS_G_SEARCH (search), TRUE);
	
	return search->priv->done;
}

/**
 * xesam_g_search_continue
 * @search: The search to continue fetching hits for
 *
 * Tell a #XesamGSearch that it should continue to fetch hits after a 
 * #XesamGSearch::hits-found signal has been emitted.
 * If the #XesamGSearch:auto-continue property is %TRUE this method is
 * automatically called if you call any method on a #XesamGHits object 
 * passed via the #XesamGSearch::hits-ready signal.
 *
 * If you choose to ignore a #XesamGHits object emitted with a
 * #XesamGSearch::hits-ready you must manually call this method before any more
 * hits are retrieved from the server.
 *
 * Under normal operations it is not necessary to invoke this method manually.
 */
void
xesam_g_search_continue	(XesamGSearch *search)
{
	search->priv->continue_read = TRUE;
	_retrieve_hits (search);
}

/**
 * xesam_g_search_get_num_found
 * @search: Search object for which to get the total number of found hits
 * @returns: The total accumulated number of hits found by this search
 *
 * Look up how many hits have been found by this search in total. This does not
 * necessarily equal the number of hits retrieved, but just the number
 * registered on the server.
 *
 * If you want to get the total number of hits matching a given search you
 * should wait for the #XesamGSearch::done signal before calling this method.
 *
 * See also xesam_g_search_get_num_read().
 */
guint
xesam_g_search_get_num_found (XesamGSearch *search)
{
	return search->priv->num_found;
}

/**
 * xesam_g_search_get_num_read
 * @search: Search object for which to get the number of hits retrieved from the
 *          the server
 * @returns: The number of hits retrieved from the server
 *
 * Look up how many hits have been read from the server. This is not necessarily
 * equal to the actual number of hits found by the server. The number of
 * hits read equals the cumulative number of hits emitted via the
 * #XesamGSearch::hits-ready signal.
 *
 * Note that this method does not necessarily give the total number of found hits
 * when #XesamGSearch::done is emitted (xesam_g_search_get_num_found() will do that).
 * The reason for this is that the read hits are only counted when the hits have
 * been retrieved from the server, and the hits may arrive after the 'done' signal
 * has been emitted.
 *
 * See also xesam_g_search_get_num_found().
 */
guint
xesam_g_search_get_num_read	(XesamGSearch *search)
{
	return search->priv->num_read;
}

/**
 * xesam_g_search_request_extended_data
 * @search: The search to which the hit id's are relative.
 * @hit_ids: An array of hit ids as returned from #xesam_g_hit_get_id.
 * @user_data: Arbitrary user data to attach to the #XesamGHits object when it
 *             is returned. This can be used to distinguis the hits from
 *             ordinary hits if relevant. If this argument is %NULL the
 *             any user data set with xesam_g_search_set_user_data() will be
 *             added to the hits instead
 * @user_data_destroy: A function called to free @user_data when the resulting
 *                     #XesamGHits object is finalized. If @user_data is %NULL
 *                     this argument will be ignored
 *
 * Read additional metadata fields from the search engine as specified in
 * the session property #XesamGSession:hit-fields-extended. The field ids
 * of the final #XesamGHit objects will be the offsets of the field names into
 * the #XesamGSession:hit-fields-extended array.
 *
 * Emits a #XesamGSearch::hits-ready with a new #XesamGHits object,
 * when results are ready.
 *
 * If @user_data is %NULL any user data supplied to
 * xesam_g_search_set_user_data() will be included in the emitted hits.
 * If @user_data is non-%NULL the emitted
 * hits will use this user data and destroy notification instead. In the last
 * case @destroy_notify will be invoked when the emitted #XesamGHits is
 * finalized.
 */
void
xesam_g_search_request_extended_data (XesamGSearch		*search,
									  GArray			*hit_ids,
									  gpointer			user_data,
									  GDestroyNotify	user_data_destroy)
{
	XesamGHitRetrievalContext	*ctx;
	XesamGSearchPrivate			*priv;
	XesamGSession				*session;
	
	g_return_if_fail (XESAM_IS_G_SEARCH (search));
	g_return_if_fail (hit_ids != NULL);
	
	session = xesam_g_search_get_session (search);
	priv = search->priv;
	
	/* Create a context. If no user data is supplied, use that of the search */
	if (user_data == NULL) {
		user_data = search->priv->user_data;
		user_data_destroy = NULL;
	}
	ctx = _hit_retrieval_ctx_new (search, hit_ids->len, hit_ids,
								  search->priv->extended_field_map,
								  user_data, user_data_destroy);
	
	dispatch_hit_data_request (ctx);
}

static void
dispatch_hit_data_request (XesamGHitRetrievalContext *ctx)
{
	XesamGSession		*session;
	XesamGSearch		*search;
	XesamGSearchPrivate	*priv;
	GStrv				extended_fields;
	
	g_return_if_fail (ctx != NULL);
	
	search = ctx->search;
	priv = search->priv;
	session = xesam_g_search_get_session (search);
	
	
	/* Check if we are awaiting search start - meaning this is a rescheduled
	 * call */
	if (priv->on_start_get_extended_handler != 0) {
		if (!xesam_g_search_is_running (search)) {
			g_critical ("In xesam_g_search_request_extended_data: Concurrent "
						"request for extended data while waiting for search start");
			_hit_retrieval_ctx_destroy (ctx);
			return;
		}
		
		xesam_g_debug_object (search, "Handling defered request to retrieve hit"
							  " data");
		
		g_signal_handler_disconnect (search,
									 priv->on_start_get_extended_handler);
		priv->on_start_get_extended_handler = 0;
		g_object_unref (search);
	}
	
	/* If we are not ready defer the request */
	if (!xesam_g_search_is_running (search) ||
		!xesam_g_search_is_ready(search)) {
		xesam_g_debug_object (search, "Not ready to retrieve extended data. "
							  "Defering request");
		search->priv->on_start_get_extended_handler =
				g_signal_connect_swapped (search, "started",
										  (GCallback) dispatch_hit_data_request,
										  ctx);
		g_object_ref (search);
		return;
	}
	
	/* The search being ready should guarantee that session:hit-fields-extended
	 * is synced */
	g_return_if_fail (xesam_g_session_test_property (session,
													 "hit-fields-extended"));
	
	g_object_get (session, "hit-fields-extended", &extended_fields, NULL);
	
	/* If the context does not have a field_map it is the extended one
	 * we want */
	if (ctx->field_map == NULL) {
		if (search->priv->extended_field_map == NULL)
			search->priv->extended_field_map = calc_field_map (extended_fields);
		
		ctx->field_map = g_hash_table_ref (search->priv->extended_field_map);
	}
	
	
	xesam_g_searcher_get_hit_data (xesam_g_session_get_searcher (session),
								   ctx->search->priv->id,
								   ctx->hit_ids,
								   extended_fields,
								   _retrieve_hits_cb,
								   ctx);								   
}

/**
 * xesam_g_search_get_id
 * @search: The search to get the id for.
 * @returns: The id. Do not modify this pointer. If the search has not yet
 *           emitted #XesamGSearch::ready %NULL will be returned
 *
 * Get the unique id for a given search.
 */
const gchar*
xesam_g_search_get_id (XesamGSearch *search)
{
	return search->priv->id;
}

/**
 * xesam_g_search_get_max_batch_size
 * @search: The search to set the max batch size for
 * @returns: The maximum number of hits included in each #XesamGHits
 *
 * Convenience method to get the value of the 
 * #XesamGSearch:max-batch-size property.
 */
guint
xesam_g_search_get_max_batch_size (XesamGSearch	*search)
{
	return search->priv->max_batch_size;
}

/**
 * xesam_g_search_set_max_batch_size
 * @search: The search to set the max batch size for
 * @max_size: Maximum number of hits per #XesamGHits object emitted
 *
 * Convenience method to set the value of the 
 * #XesamGSearch:max-batch-size property.
 */
void
xesam_g_search_set_max_batch_size	(XesamGSearch *search, guint max_size)
{
	g_object_set (G_OBJECT(search), "max-batch-size", max_size, NULL);
}

/**
 * xesam_g_search_get_last_error
 * @search:  The search to inspect for errors
 * @returns: The last error encountered from the search engine. If no error(s)
 *           has happened %NULL is returned
 *
 * See also the signal XesamGSearch::error
 */
const GError*
xesam_g_search_get_last_error (XesamGSearch	*search)
{
	g_return_val_if_fail (XESAM_IS_G_SEARCH (search), NULL);
					  
	return search->priv->last_error;
}

/**
 * xesam_g_search_set_user_data
 * @search: The search to get the user data from
 * @user_data: A pointer to some user supplied data
 * @user_data_destroy: A function to clean up @user_data when this search is
 *                     finalized
 *
 * Set a pointer to a chunk of user data that will be available on all
 * #XesamGHits spawned by this search.
 *
 * When @search is finalized @user_data_destroy will be invoked if it is
 * non-NULL. If you override the user data with later calls to
 * this method @user_data_destroy will also be called. 
 *
 * You are allowed to change the user data during the lifetime of @search.
 *
 * Returns: @user_data
 */
gpointer
xesam_g_search_set_user_data (XesamGSearch		*search,
							  gpointer			user_data,
							  GDestroyNotify	user_data_destroy)
{
	g_return_val_if_fail (XESAM_IS_G_SEARCH (search), user_data);
	
	/* Destroy the user data if we have some already */
	if (search->priv->user_data_destroy != NULL &&
		search->priv->user_data != NULL)
		search->priv->user_data_destroy (search->priv->user_data);
	
	search->priv->user_data = user_data;
	search->priv->user_data_destroy = user_data_destroy;
	
	return user_data;
}

/**
 * xesam_g_search_get_user_data
 * @search: The search to get the user data from
 *
 * Look up the user data set by xesam_g_search_get_user_data().
 *
 * Returns: The user data provided on the invocation of
 *          xesam_g_search_set_user_data()
 */
gpointer
xesam_g_search_get_user_data (XesamGSearch	*search) {
	
	g_return_val_if_fail (XESAM_IS_G_SEARCH (search), NULL);
	
	return search->priv->user_data;
}

/* Retrieve hit data on the condition that there are some ready
 * and that we have been instructed to continue fetching hits. */
gboolean
_retrieve_hits (XesamGSearch *search)
{
	guint 						num_read;
	XesamGSession				*session;
	XesamGHitRetrievalContext 	*ctx;
	
	g_return_val_if_fail (XESAM_IS_G_SEARCH(search), FALSE);
	
	num_read = MIN (search->priv->pending_hits, search->priv->max_batch_size);
	session = xesam_g_search_get_session (search);
	
	if (num_read == 0) {
		xesam_g_debug_object (search, "Ignoring request to retrieve "
							  "hits. No hits pending.");
		return FALSE;
	}

	if (!search->priv->continue_read) {
		xesam_g_debug_object (search, "Ignoring request to retrieve "
							  "hits. Not set to continue");
		return FALSE;
	}
	
	/* With this user data configuration the Hits will not call user_data_destroy,
	 * we only want to do that when we finalize this Search object */
	ctx = _hit_retrieval_ctx_new (search, num_read, NULL,
								  search->priv->field_map,
								  search->priv->user_data, NULL);
	
	search->priv->pending_hits -= num_read;
	
	xesam_g_debug_object (search, "Retrieving %d hits", num_read);
	
	xesam_g_searcher_get_hits (xesam_g_session_get_searcher (session),
							   search->priv->id,
							   num_read,
							   _retrieve_hits_cb,
							   ctx);
	
	search->priv->continue_read = FALSE;
	
	return FALSE;
}

/* Called when hit data is ready. Emits hits-ready when Hits object is ready
 * IMPORTANT: This method is a callback from a XesamGSearcher which is an
 * untrusted source, we should be careful to check the args */
static void
_retrieve_hits_cb (XesamGSearcher  *searcher,
				   GPtrArray		*hit_data,
				   gpointer			data,
				   GError			*error)
{
	XesamGHitRetrievalContext 	*ctx;
	XesamGSearch				*search;
	XesamGHits					*hits;
	
	g_return_if_fail (data != NULL);
	
	ctx = (XesamGHitRetrievalContext*) (data);
	search = ctx->search;
	
	g_return_if_fail (XESAM_IS_G_SEARCH(search));
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	
	if (error) {
		g_warning ("Error reading hits for search '%s': %s",
					search->priv->id, error->message);
		_set_last_error (search, error);
		return;
	}
	
	if (hit_data == NULL) {
		g_warning ("Error reading hits for search '%s'. Server returned NULL",
					search->priv->id);
		_set_new_error (search, XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
						"Error reading hits for search '%s'. Server returned NULL",
						search->priv->id);
		return;
	}
	
	if (hit_data->len != ctx->num_read) {
		g_warning ("Error reading hits for search '%s'. Server returned %d "
				   "hits on request for %d hits",
					search->priv->id, hit_data->len, ctx->num_read);
		_set_new_error (search, XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
						"Error reading hits for search '%s'. Server returned %d "
						"hits on request for %d hits",
						search->priv->id, hit_data->len, ctx->num_read);
		return;
	}
	
	/* If ctx->hit_ids == NULL this is a normal sequential reading of hits.
	 * Otherwise it is a request for a collection of sparse hits.
	 *
	 * The length of the individual hit's GPtrArray should match the size of
	 * of the field map. This is asserted when we create the actual hit
	 * in xesam_g_hit_new()
	 */
	if (ctx->hit_ids == NULL) {
		if (ctx->user_data_destroy != NULL)
			g_warning ("Unexpected user_data_destroy. "
					   "This is a bug in xesam-glib");
		
		/* If the field map is missing we want the primary one */
		if (ctx->field_map == NULL) {
			if (search->priv->field_map == NULL)
				search->priv->field_map = calc_field_map_from_property (xesam_g_search_get_session (search),
																		"hit-fields");
			
			ctx->field_map = g_hash_table_ref (search->priv->field_map);
		}
		
		hits = xesam_g_sequential_hits_new (search,
											ctx->batch_offset,
											ctx->field_map,
											hit_data,
											search->priv->auto_continue,
											ctx->user_data,
											NULL);
		
		search->priv->num_read += hit_data->len;
		
	} else {
		/* This is a request for sparse collection of hits */
		
		/* If the field map is missing we want the extended one */
		if (ctx->field_map == NULL) {
			if (search->priv->field_map == NULL)
				search->priv->extended_field_map = calc_field_map_from_property (xesam_g_search_get_session (search),
																				 "hit-fields-extended");
			ctx->field_map = g_hash_table_ref (search->priv->extended_field_map);
		}
		
		hits = xesam_g_sparse_hits_new (search,
										ctx->hit_ids,
										ctx->field_map,
										hit_data,
										ctx->user_data,
										ctx->user_data_destroy);
	}
	
	xesam_g_debug_object (search, "%s@%p retrieved",
						  G_OBJECT_TYPE_NAME(hits), hits);
	
	g_signal_emit (search, g_search_signals[HITS_READY], 0, hits);
	
	/* Any signal handlers interested in the Hits object should reference it */
	g_object_unref (hits);
	
	_hit_retrieval_ctx_destroy (ctx);
}

/* Remember to unref the the search object as it is reffed before the async call */
static void
_search_handle_received_cb (XesamGSearcher  *searcher,
							gchar			*search_handle,
							gpointer		data,
							GError			*error)
{
	XesamGSearch 	*search;
	
	g_return_if_fail (XESAM_IS_G_SEARCH (data));
	g_return_if_fail (XESAM_IS_G_SEARCHER (searcher));
	
	search = XESAM_G_SEARCH (data);
	xesam_g_debug_object (search, "Got search handle '%s'", search_handle);
	
	/* id should not have been set previously */
	g_return_if_fail (search->priv->id == NULL);
	
	if (error) {
		g_warning ("Failed to obtain search handle: %s", error->message);
		_set_last_error (search, error);
		g_object_unref (search);
		return;
	}
	
	if (search_handle == NULL) {
		_set_new_error (search, XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
						"Searcher returned NULL search handle");
		g_object_unref (search);
		return;
	}
	
	search->priv->id = search_handle;
	
	g_signal_emit (search, g_search_signals[READY], 0);
	g_object_unref (search);
}

/* Request a search handle from the search engine. */
static void
_get_search_handle (XesamGSearchContext *search_ctx)
{
	XesamGSearch	*search;
	XesamGQuery		*query;
	XesamGSession	*session;
	gchar			*session_id;
	
	g_return_if_fail (search_ctx != NULL);
	
	search = search_ctx->search;
	query = search_ctx->query;
	session = search->priv->session;
	session_id = xesam_g_session_get_id (session);
	
	/* If session is not ready, retry when it is */
	if (!xesam_g_session_is_ready(session)) {
		search->priv->on_session_ready_handler = 
					g_signal_connect_swapped (session, "ready", 
											  G_CALLBACK (_get_search_handle),
											  search_ctx);
		return;
	}
	
	/* Disconnect the "ready" handler on the session if we have one */
	if (search->priv->on_session_ready_handler != 0) {
		g_signal_handler_disconnect (session,
									 search->priv->on_session_ready_handler);
		search->priv->on_session_ready_handler = 0;
	}
	
	/* Start fetching a search handle */
	g_object_ref (search);
	xesam_g_searcher_new_search (xesam_g_session_get_searcher(session),
								 xesam_g_session_get_id (session),
								 xesam_g_query_get_xml (query),
								 _search_handle_received_cb,
								 search);
	
	_search_ctx_destroy (search_ctx);
}

/* Remember to unref search, which was reffed when we dispatched the close-request */
static void
_closed_cb (XesamGSearcher  *searcher,
			gpointer		data,
			GError			*error)
{
	XesamGSearch   *search;
	
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	g_return_if_fail (XESAM_IS_G_SEARCH(data));
	
	search = XESAM_G_SEARCH (data);
	
	if (error) {
		_set_last_error (search, error);
		g_object_unref (search);
		return;
	}
	
	xesam_g_debug_object (search, "Closed");
	
	search->priv->closed = TRUE;
	g_signal_emit (search, g_search_signals[CLOSED], 0);
	
	g_object_unref (search);
}

/* Remember to unref search, which was reffed when we dispatched the start-request */
static void
_started_cb	(XesamGSearcher *searcher,
			 gpointer		data,
			 GError			*error)
{
	XesamGSearch   *search;
	
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	g_return_if_fail (XESAM_IS_G_SEARCH(data));
	
	search = XESAM_G_SEARCH (data);
	
	if (error) {
		_set_last_error (search, error);
		g_object_unref (search);
		return;
	}
	
	xesam_g_debug_object (search, "Started");
	
	search->priv->started = TRUE;
	g_signal_emit (search, g_search_signals[STARTED], 0);
	
	g_object_unref (search);
}

static XesamGSearchContext*
_search_ctx_new (XesamGSearch *search, XesamGQuery *query)
{
	XesamGSearchContext *ctx;
	
	g_return_val_if_fail (XESAM_IS_G_SEARCH (search), NULL);
	g_return_val_if_fail (XESAM_IS_G_QUERY (query), NULL);
	
	ctx = g_new (XesamGSearchContext, 1);
	ctx->search = search;
	ctx->query = query;
	
	g_object_ref (search);
	g_object_ref_sink (query);
	
	return ctx;
}

static void
_search_ctx_destroy	(XesamGSearchContext *search_ctx)
{
	g_return_if_fail (search_ctx != NULL);
	
	g_object_unref (search_ctx->search);
	g_object_unref (search_ctx->query);
	g_free (search_ctx);
}

/**
 * If @hit_ids is NULL this context represents a request for a sequential
 * batch of hits. If it is non-NULL this is a request for a collection of
 * sparse hits.
 */
static XesamGHitRetrievalContext*
_hit_retrieval_ctx_new (XesamGSearch	*search,
						guint			num_read,
						GArray			*hit_ids,
						GHashTable		*field_map,
						gpointer		user_data,
						GDestroyNotify	user_data_destroy)
{
	XesamGHitRetrievalContext *ctx;
	
	g_return_val_if_fail (XESAM_IS_G_SEARCH (search), NULL);
	
	ctx = g_new0 (XesamGHitRetrievalContext, 1);
	ctx->search = search;
	
	/* The next offset is the number of hits read + the number of hits that are
	 * in transit, ie. have been requested, but not returned yet.
	 * By the Hit Count Invariant (see top of file) we have the next offset as:
	 */
	ctx->batch_offset = search->priv->num_found - search->priv->pending_hits;
	ctx->num_read = num_read;
	ctx->hit_ids = hit_ids;
	ctx->field_map = field_map ? g_hash_table_ref (field_map) : NULL;
	ctx->user_data = user_data;
	ctx->user_data_destroy = user_data_destroy;
	
	g_object_ref (search);
	
	return ctx;
}

static void
_hit_retrieval_ctx_destroy (XesamGHitRetrievalContext *ctx)
{
	g_return_if_fail (ctx != NULL);
	
	g_object_unref (ctx->search);
	
	if (ctx->field_map)
		g_hash_table_unref (ctx->field_map);
	
	g_free (ctx);
}

/* Callback for the closed signal on the session */
static void
_parent_session_closed (XesamGSearch	*search)
{
	g_return_if_fail (XESAM_IS_G_SEARCH (search));
	
	if (!search->priv->closed) {
		xesam_g_debug_object (search, "Parent session closed. Closing");
		
		/* We should not call xesam_g_search_close() since that would
		 * call the searcher which should already have invalidated us
		 * because our session is dead */
		search->priv->closed = TRUE;
		g_signal_emit (search, g_search_signals[CLOSED], 0);
	}
}

static void
_dispatch_hits_added (XesamGSearcher	*searcher,
					  const gchar		*search_handle,
					  guint				count,
					  gpointer			data)
{
	XesamGSearch	*search;
	
	g_return_if_fail (XESAM_IS_G_SEARCH (data));
	g_return_if_fail (XESAM_IS_G_SEARCHER (searcher));
	
	search = XESAM_G_SEARCH (data);
	
	if (!g_str_equal (search_handle, search->priv->id)) {
		xesam_g_debug ("Search '%s' ignoring hits-added on wrong handle '%s'.",
					search->priv->id, search_handle);
		return;
	}
	
	xesam_g_debug_object (search, "Hits found %d", count);
	
	search->priv->num_found += count;
	search->priv->pending_hits += count;
	g_signal_emit (search, g_search_signals[HITS_FOUND], 0, count);
	
	/* Start fetching hits (if we do not have some queued already) */
	_retrieve_hits (search);
}

static void
_dispatch_hits_modified (XesamGSearcher *searcher,
						 const gchar	*search_handle,
						 const GArray   *hit_ids,
						 gpointer		data)
{
	XesamGSearch	*search;
	
	g_return_if_fail (XESAM_IS_G_SEARCH (data));
	g_return_if_fail (XESAM_IS_G_SEARCHER (searcher));
	
	search = XESAM_G_SEARCH (data);
	
	if (!g_str_equal (search_handle, search->priv->id)) {
		xesam_g_debug ("Search '%s' ignoring hits-modified on wrong handle '%s'.",
					search->priv->id, search_handle);
		return;
	}
	
	xesam_g_debug_object (search, "Hits modified %d", hit_ids->len);
	
	g_signal_emit (search, g_search_signals[HITS_MODIFIED], 0, hit_ids);
}

static void
_dispatch_hits_removed (XesamGSearcher  *searcher,
						const gchar		*search_handle,
						const GArray	*hit_ids,
						gpointer		data)
{
	XesamGSearch	*search;
	
	g_return_if_fail (XESAM_IS_G_SEARCH (data));
	g_return_if_fail (XESAM_IS_G_SEARCHER (searcher));
	
	search = XESAM_G_SEARCH (data);
	
	if (!g_str_equal (search_handle, search->priv->id)) {
		xesam_g_debug ("Search '%s' ignoring hits-removed on wrong handle '%s'.",
					search->priv->id, search_handle);
		return;
	}
	
	xesam_g_debug_object (search, "Hits removed %d", hit_ids->len);
	
	g_signal_emit (search, g_search_signals[HITS_REMOVED], 0, hit_ids);
}

static void
_dispatch_search_done (XesamGSearcher   *searcher,
					   const gchar		*search_handle,
					   gpointer			data)
{
	XesamGSearch	*search;
	
	g_return_if_fail (XESAM_IS_G_SEARCH (data));
	g_return_if_fail (XESAM_IS_G_SEARCHER (searcher));
	
	search = XESAM_G_SEARCH (data);
	
	if (!g_str_equal (search_handle, search->priv->id)) {
		xesam_g_debug ("Search '%s' ignoring search-done on wrong handle '%s'.",
					search->priv->id, search_handle);
		return;
	}
	
	if (search->priv->done) {
		g_warning ("Search '%s' notified as done more than once",
				   search->priv->id);
		return;
	}
	
	xesam_g_debug_object (search, "Done");
	
	search->priv->done = TRUE;
	g_signal_emit (search, g_search_signals[DONE], 0);
}

static void
_set_new_error (XesamGSearch	*search,
				gint			code,
				const gchar		*format,
				...)
{	
	GError	*error;
	va_list args;
	gchar   *message;
	
	g_return_if_fail (XESAM_IS_G_SEARCH(search));
	
	va_start (args, format);
	g_vasprintf (&message, format, args);
	va_end (args);
	
	error = g_error_new (XESAM_G_SEARCHER_ERROR, code, message);
	
	g_free (message);
	_set_last_error (search, error);
	
}

/* Update search's last_error, free the incoming one,
 * and emit an "error" signal */
static void 
_set_last_error	(XesamGSearch   *search,
				 GError			*error)
{
	XesamGSearchPrivate	*priv;
	
	g_return_if_fail (XESAM_IS_G_SEARCH(search));
	g_return_if_fail (error != NULL);
	
	priv = search->priv;
	
	if (priv->last_error)
		g_error_free (priv->last_error);
	priv->last_error = g_error_copy (error);
	g_error_free (error);
	
	g_signal_emit (search, g_search_signals[ERROR], 0);
}
