/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2007 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include "xesam-g-searcher.h"
#include "xesam-glib-globals.h"
#include "xesam-marshal.h"

/* FIXME: We need this because XESAM_TYPE_UINT_ARRAY is just a def of
 *        DBUS_TYPE_G_UINT_ARRAY, which we need in the signals.
 *        The dbus-glib dep here should be removed */
#include <dbus/dbus-glib.h>

/**
 * SECTION:xesam-g-searcher
 * @short_description: Low level interface to a search engine
 *
 * #XesamGSearcher is a low level interface abstracting access to a search
 * engine. In other words this interface represents the protocol used to talk
 * to a search engine.
 * 
 * Clients consuming xesam-glib should normally not use this
 * interface, but use #XesamGSession, #XesamGSearch, and friends instead. They
 * provide an object oriented API abstracting out the protocol.
 * 
 * The intent of this interface is to allow other implementations than the
 * DBus one defined by the Xesam standard - such as an in-process library,
 * or networked search engine. Because of this all methods on this interface
 * has asynchrnous callbacks instead of return values.
 *
 * Search engines not wishing to, or not capable of, exposing a DBus interface
 * should use this interface instead.
 *
 * Given an implementation of #XesamGSearcher it is easy to export it over DBus
 * by simply passing a searcher instance to xesam_g_dbus_searcher_stub_export().
 *
 * #XesamGSearcher is available since 1.0
 */

/* 
 * CALLBACK DOCS
 */

/**
 * XesamGSearcherGotHandle
 * @searcher: The searcher on which the initial request was made
 * @handle: Search- or session handle received
 * @user_data: Arbitrary pointer provided by the user in the function 
 *             triggering this callback.
 * @error: Set in case the #XesamGSearcher encountered an error. If @error is
 *         non-%NULL, @handle will be %NULL
 *
 * Used for responses to xesam_g_searcher_new_session() and 
 * xesam_g_searcher_new_search().
 *
 * As with all #XesamGSearcher callbacks, the receiver is responsible for
 * freeing all parameters, ie @handle, @user_data, and @error.
 */

/**
 * XesamGSearcherGotProperty
 * @searcher: The searcher on which the initial request was made
 * @prop_name: Name of the property received. This is the Xesam formatted name,
 *             ie separated with dots like "hit.snippet.length"
 * @value: The value of the property, or %NULL on error
 * @user_data: Arbitrary pointer provided by the user in the function triggering
 *             this callback.
 * @error: Set in case the #XesamGSearcher encountered an error. If @error is
 *         non-%NULL, @value will be %NULL
 *
 * Used for responses to xesam_g_searcher_set_property() and
 * xesam_g_searcher_get_property().
 *
 * As with all #XesamGSearcher callbacks, the receiver is responsible for
 * freeing all parameters, ie @prop_name, @value, @user_data, and @error.
 */


/**
 * XesamGSearcherGotHits
 * @searcher: The searcher on which the initial request was made
 * @hits: A #GPtrArray of #GPtrArray<!-- -->s of #GValue<!-- -->s. Each 
 *        #GPtrArray in the outermost array maps to a hit. The #GValue<!-- -->s
 *        in each hit map to the fields requested in the given context. This
 *        argument is %NULL on error
 * @user_data: Arbitrary pointer provided by the user in the function triggering
 *             this callback.
 * @error: Set in case the #XesamGSearcher encountered an error. If @error is
 *         non-%NULL, @hits will be %NULL
 *
 * Callback used for responses to xesam_g_searcher_get_hits() and 
 * xesam_g_searcher_get_hit_data().
 * 
 * As with all #XesamGSearcher callbacks, the receiver is responsible for
 * freeing all parameters, ie @hits, @user_data, and @error.
 *
 * In responses to xesam_g_searcher_get_hits() the hit ids can be calculated
 * as the sequence number in which the hit was read from the search engine.
 * For example the ten first hits to be read will have ids 0-9. Reading another
 * batch of 20 hits will give you the hits with id 10-29.
 *
 * For responses to xesam_g_searcher_get_hit_data() the hit ids will match those
 * passed in as the hit_ids parameter to said method.
 */

/**
 * XesamGSearcherGotHitCount
 * @searcher: The searcher on which the initial request was made
 * @count: The number of hits currently found by the server. The final number
 *         of hits can be read when #XesamGSearcher::search-done is emitted
 * @user_data: Arbitrary pointer provided by the user in the function triggering
 *             this callback.
 * @error: Set in case the #XesamGSearcher encountered an error. If @error is
 *         non-%NULL @count can be arbitrary
 *
 * Callback used for responses to xesam_g_searcher_get_hit_count().
 */

/**
 * XesamGSearcherGotState
 * @searcher: The searcher on which the initial request was made
 * @state_info: A #GStrv of length 2. The first entry is one of IDLE, UPDATE,
 *              or FULL_INDEX, the second entry is parsable as an integer between
 *              0 and 100 and denotes a percentage  of the running task's
 *              completion. The second entry is meaningless if the first entry
 *              is IDLE. This argument is %NULL on errors
 * @user_data: Arbitrary pointer provided by the user in the function triggering
 *             this callback.
 * @error: Set in case the #XesamGSearcher encountered an error. If @error is
 *         non-%NULL @state_info is %NULL
 *
 * Callback used for responses to xesam_g_searcher_get_state().
 *
 * As with all #XesamGSearcher callbacks, the receiver is responsible for
 * freeing all parameters, ie @state_info, @user_data, and @error.
 */

/* Signals */
enum
{
	HITS_ADDED,
	HITS_REMOVED,
	HITS_MODIFIED,
	SEARCH_DONE,
	STATE_CHANGED,
	PADDING_1,
	PADDING_2,
	PADDING_3,
	
	LAST_SIGNAL
};

guint xesam_g_searcher_signals[LAST_SIGNAL] = { 0 };


/**
 * xesam_g_searcher_new_session
 * @self: Searcher on which to create a new session
 * @callback: Called when the the session handle is ready
 * @user_data: Data to pass back with #callback
 *
 * Asynchronously create a new session. When the search engine is ready
 * the session handle will be passed back through @callback.
 */
void
xesam_g_searcher_new_session (XesamGSearcher *self,
						      XesamGSearcherGotHandle callback,
						      gpointer user_data)
{
	XESAM_G_SEARCHER_GET_INTERFACE (self)->new_session (self, callback,
														user_data);
}

/**
 * xesam_g_searcher_close_session
 * @self: Searcher on which to create a new session
 * @session_handle: Unique identifier supplied by the search engine, as obtained
 *                  from the callback of xesam_g_searcher_new_session()
 * @callback: Called when the server has acknowledged the session as closed.
 *            Optionally %NULL
 *            in which case nothing will be done when the method has completed.
 *            In this case you should also pass @user_data %NULL
 * @user_data: Data to pass back with #callback
 *
 * Notify the search engine that a session is no longer needed. Any searches
 * spawned on this session is also invalidated automatically by the server.
 */
void
xesam_g_searcher_close_session (XesamGSearcher *self,
						        const gchar *session_handle,
						        XesamGSearcherVoidResponse callback,
						        gpointer user_data)
{
	XESAM_G_SEARCHER_GET_INTERFACE (self)->close_session (self, session_handle,
														  callback, user_data);
}

/**
 * xesam_g_searcher_get_property
 * @self: The searcher to get the session property on
 * @session_handle: The session for which the property should be retrieved, as
 *                  obtained from the callback of xesam_g_searcher_new_session()
 * @prop_name: The name of the property to retrieve
 * @callback: Callback to invoke when the data is ready
 * @user_data: Any data to pass to @callback
 *
 * Look up a session property on a #XesamGSearcher. The value of the property
 * is passed back through @callback.
 */
void
xesam_g_searcher_get_property (XesamGSearcher *self,
						       const gchar *session_handle,
						       const gchar *prop_name,
						       XesamGSearcherGotProperty callback,
						       gpointer user_data)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER (self));
	
	XESAM_G_SEARCHER_GET_INTERFACE (self)->get_property (self, session_handle,
														 prop_name, callback,
														 user_data);
}

/**
 * xesam_g_searcher_set_property
 * @self: The searcher to get the session property on
 * @session_handle: The session for which the property should be retrieved, as
 *                  obtained from the callback of xesam_g_searcher_new_session()
 * @prop_name: The name of the property to retrieve
 * @value: The value to set
 * @callback: Callback to invoke when the data is ready. This may optionally be
 *            %NULL in which case @user_data should also be %NULL
 * @user_data: Any data to pass to @callback
 *
 * Set a session property on a #XesamGSearcher.
 *
 * It is important to notice that the search engine might not respect the
 * requested property value. The actual property value the search engine
 * will use is passed back through @callback.
 */
void
xesam_g_searcher_set_property (XesamGSearcher *self,
						       const gchar *session_handle,
						       const gchar *prop_name,
						       const GValue *value,
						       XesamGSearcherGotProperty callback,
						       gpointer user_data)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER (self));
	
	XESAM_G_SEARCHER_GET_INTERFACE (self)->set_property (self, session_handle,
														 prop_name, value,
														 callback, user_data);
}

/**
 * xesam_g_searcher_new_search
 * @self: The searcher on which to spawn a search
 * @session_handle: The session for which to spawn the search, as
 *                  obtained from the callback of xesam_g_searcher_new_session()
 * @query: The valid Xesam Query xml string
 * @callback: Callback to invoke when the search handle is ready
 * @user_data: Any data to pass to @callback
 *
 * Spawn a new search in a session. Before any hits will be reported via
 * the #XesamGSearcher::hits-added signal xesam_g_searcher_start_search()
 * must be invoked on the search handle obtained in the callback from this method.
 */
void
xesam_g_searcher_new_search (XesamGSearcher * self,
						     const gchar *session_handle,
						     const gchar *query,
						     XesamGSearcherGotHandle callback,
						     gpointer user_data)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER (self));
	
	XESAM_G_SEARCHER_GET_INTERFACE (self)->new_search (self, session_handle,
													   query, callback,
													   user_data);
}

/**
 * xesam_g_searcher_start_search
 * @self: The searcher to start the search on
 * @search_handle: The search handle obtained from the callback of
 *                 xesam_g_searcher_new_search()
 * @callback: Callback to invoke when the search has been started.
 *            This may optionally be
 *            %NULL in which case @user_data should also be %NULL
 * @user_data: Any data to pass to @callback
 *
 * Run a search registered with xesam_g_searcher_new_search().
 */
void
xesam_g_searcher_start_search (XesamGSearcher *self,
						       const gchar *search_handle,
						       XesamGSearcherVoidResponse callback,
						       gpointer user_data)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER (self));
	
	XESAM_G_SEARCHER_GET_INTERFACE (self)->start_search (self, search_handle,
														 callback, user_data);
}

/**
 * xesam_g_searcher_close_search
 * @self: The searcher to close the search on
 * @search_handle: The search handle obtained from the callback of
 *                 xesam_g_searcher_new_search()
 * @callback: Callback to invoke when the search has been closed.
 *            This may optionally be
 *            %NULL in which case @user_data should also be %NULL
 * @user_data: Any data to pass to @callback
 *
 * Notify the search engine that the search is not longer needed. @search_handle
 * will be invalid when @callback is invoked.
 */
void
xesam_g_searcher_close_search (XesamGSearcher *self,
							   const gchar *search_handle,
							   XesamGSearcherVoidResponse callback,
							   gpointer user_data)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER (self));
	
	XESAM_G_SEARCHER_GET_INTERFACE (self)->close_search (self, search_handle,
														 callback, user_data);
}

/**
 * xesam_g_searcher_get_hits
 * @self: The searcher to start the search on
 * @search_handle: The search handle obtained from the callback of
 *                 xesam_g_searcher_new_search()
 * @callback: Callback to invoke when the hit data is ready
 * @user_data: Any data to pass to @callback
 *
 * Read hit metadata from the search engine. The metadata fields of the hits
 * will match those set in the session property hit.fields.  When the server
 * returns the hit data, it is passed to the caller via @callback.
 */
void
xesam_g_searcher_get_hits (XesamGSearcher   *self,
						   const gchar		*search_handle,
						   guint			count,
						   XesamGSearcherGotHits callback,
						   gpointer			user_data)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER (self));
	
	XESAM_G_SEARCHER_GET_INTERFACE (self)->get_hits (self, search_handle, count,
													 callback, user_data);
}

/**
 * xesam_g_searcher_get_hit_data
 * @self: The searcher to start the search on
 * @search_handle: The search handle obtained from the callback of
 *                 xesam_g_searcher_new_search()
 * @hit_ids: A #GArray containing the guint ids of the hits for which to look
 *           up additional data
 * @A #GStrv containing the names of the fields to look up
 * @callback: Callback to invoke when the requested metadata is ready
 * @user_data: Any data to pass to @callback
 *
 * Get arbitrary hit metadata. When the hit data is ready it is passed back
 * to the caller via @callback.
 */
void
xesam_g_searcher_get_hit_data (XesamGSearcher   *self,
						       const gchar		*search_handle,
							   GArray			*hit_ids,
						       GStrv			field_names,						       
						       XesamGSearcherGotHits callback,
						       gpointer			user_data)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER (self));
	
	XESAM_G_SEARCHER_GET_INTERFACE (self)->get_hit_data (self, search_handle,
														 hit_ids, field_names, 
														 callback, user_data);
}

/**
 * xesam_g_searcher_get_hit_count
 * @self: The searcher to start the search on
 * @search_handle: The search handle obtained from the callback of
 *                 xesam_g_searcher_new_search()
 * @callback: Callback to invoke when the search has been started.
 *            This may optionally be
 *            %NULL in which case @user_data should also be %NULL
 * @user_data: Any data to pass to @callback
 *
 * Get the current number of found hits. This number is not final before
 * #XesamGSearcher::search-done has been emitted on @search_handle.
 *
 * The hit count is passed to @callback when the server replies.
 */
void
xesam_g_searcher_get_hit_count (XesamGSearcher  *self,
						        const gchar		*search_handle,
						        XesamGSearcherGotHitCount callback,
						        gpointer		user_data)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER (self));
	
	XESAM_G_SEARCHER_GET_INTERFACE (self)->get_hit_count (self, search_handle,
														  callback, user_data);
}

/**
 * xesam_g_searcher_get_state
 * @self: The searcher to inspect
 * @callback: Callback to invoke when the search engine state has been determined
 * @user_data: Any data to pass to @callback
 *
 * Inspect the current state of the search engine. The state is passed to
 * @callback when the server replies.
 */
void
xesam_g_searcher_get_state (XesamGSearcher * self,
						    XesamGSearcherGotState callback,
						    gpointer user_data)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER (self));
	
	XESAM_G_SEARCHER_GET_INTERFACE (self)->get_state (self, callback, user_data);
}


static void
xesam_g_searcher_base_init (XesamGSearcherIface * iface)
{
	static gboolean initialized = FALSE;
	
	if (!initialized) {
		
		initialized = TRUE;
		
		/* FIXME: We probably should initialize the dbus glib specialized types
		 * here, but bug https://bugs.freedesktop.org/show_bug.cgi?id=13908 blocks
		 * us from doing that. */
		
		/**
		 * XesamGSearcher::hits-added
		 * @searcher: The searcher emitting the signal
		 * @search_handle: The search on which the hits where found
		 * @count: The number of hits found on the server
		 *
		 * Emitted when the server finds new hits. To retrieve the actual hits
		 * call xesam_g_searcher_get_hits().
		 */
		xesam_g_searcher_signals[HITS_ADDED] =
			g_signal_new ("hits-added",
						  XESAM_TYPE_G_SEARCHER,
						  G_SIGNAL_RUN_LAST,
						  0, NULL, NULL,
						  xesam_marshal_VOID__STRING_UINT,
						  G_TYPE_NONE, 
						  2, G_TYPE_STRING, G_TYPE_UINT);
		
		/**
		 * XesamGSearcher::hits-removed
		 * @searcher: The searcher emitting the signal
		 * @search_handle: The search on which the hits where removed
		 * @hit_ids: A %GArray containing the the %guint hit ids of the
		 *           removed hits
		 *
		 * Emitted when the server discovers that an object represented by a hit
		 * has been removed or changed in a way such that it longer matches the
		 * query of @search_handle.
		 */
		xesam_g_searcher_signals[HITS_REMOVED] =
			g_signal_new ("hits-removed",
						  XESAM_TYPE_G_SEARCHER,
						  G_SIGNAL_RUN_LAST,
						  0, NULL, NULL,
						  xesam_marshal_VOID__STRING_POINTER,
						  G_TYPE_NONE,
						  2, G_TYPE_STRING, G_TYPE_POINTER);
		
		/**
		 * XesamGSearcher::hits-modified
		 * @searcher: The searcher emitting the signal
		 * @search_handle: The search on which the hits where modified
		 * @hit_ids: A %GArray containing the the %guint hit ids of the
		 *           modified hits
		 *
		 * Emitted when the server discovers that the object (think, a file)
		 * represented by a hit has changed. To retrieve the updated hit
		 * data call xesam_g_searcher_get_hit_data().
		 */
		xesam_g_searcher_signals[HITS_MODIFIED] =
			g_signal_new ("hits-modified",
						  XESAM_TYPE_G_SEARCHER,
						  G_SIGNAL_RUN_LAST,
						  0, NULL, NULL,
						  xesam_marshal_VOID__STRING_POINTER,
						  G_TYPE_NONE,
						  2, G_TYPE_STRING, G_TYPE_POINTER);
		
		/**
		 * XesamGSearcher::search-done
		 * @searcher: The searcher emitting the signal
		 * @search_handle: The search on which the hits where removed
		 *
		 * Emitted when the server is done scanning for hits on @search_handle.
		 * On most servers this will mean that the server is done scanning the
		 * index.
		 *
		 * It is important to note that further emissions of 
		 * #XesamGSearcher::hits-added, #XesamGSearcher::hits-removed, and
		 * #XesamGSearcher::hits-modified may still occur if the session property
		 * "search.live" has been set to %TRUE via
		 * xesam_g_searcher_set_property(). A live search is a search where
		 * the search engine will keep monitoring the result set for changes
		 * to the underlying objects and notify the consumer on such changes.
		 */
		xesam_g_searcher_signals[SEARCH_DONE] =	    	      
			g_signal_new ("search-done",
						  XESAM_TYPE_G_SEARCHER,
						  G_SIGNAL_RUN_LAST,
						  0, NULL, NULL,
						  g_cclosure_marshal_VOID__STRING,
						  G_TYPE_NONE,
						  1, G_TYPE_STRING);
		
		/**
		 * XesamGSearcher::state-changed
		 * @searcher: The searcher emitting the signal
		 * @state_info: A %GStrv of length 2. The string in position 0 is either
		 *              "IDLE", "UPDATE", or "FULL_INDEX" and the string at
		 *              position 1 is parseable as an integer in the range
		 *              0-100 indicating the completeness of either the
		 *              UPDATE or FULL_INDEX state. The second value should be
		 *             ignored if the state is IDLE
		 *
		 * Emitted when the server state changes. Normally meaning that it starts
		 * updating its indexes or stops doing so.
		 *
		 * You can manually retrieve the state information via
		 * xesam_g_searcher_get_state().
		 */
		xesam_g_searcher_signals[STATE_CHANGED] =
			g_signal_new ("state-changed",
						  XESAM_TYPE_G_SEARCHER,
						  G_SIGNAL_RUN_LAST,
						  0, NULL, NULL, 
						  g_cclosure_marshal_VOID__BOXED,
						  G_TYPE_NONE,
						  1, G_TYPE_STRV);
	}
}


GType
xesam_g_searcher_get_type (void)
{
	static GType xesam_g_searcher_type_id = 0;
	if (G_UNLIKELY (xesam_g_searcher_type_id == 0)) {
		
		static const GTypeInfo g_define_type_info = { 
			sizeof (XesamGSearcherIface),
			(GBaseInitFunc) xesam_g_searcher_base_init,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) NULL,
			(GClassFinalizeFunc) NULL,
			NULL,
			0,
			0,
			(GInstanceInitFunc) NULL
		};
		
		xesam_g_searcher_type_id =
			g_type_register_static (G_TYPE_INTERFACE, "XesamGSearcher",
						&g_define_type_info, 0);
					
		g_type_interface_add_prerequisite (xesam_g_searcher_type_id,
					 G_TYPE_OBJECT);
	}
	
	return xesam_g_searcher_type_id;
}

GQuark
xesam_g_searcher_error_quark (void)
{
  return g_quark_from_static_string ("xesam-g-searcher-error-quark");
}

/**
 * xesam_g_searcher_hits_added
 * @self: The searcher to emit the signal on
 * @search_handle: The search handle on which hits where updated
 * @count: The number of hits added
 *
 * Emit the %XesamGSearcher::hits-added signal on the searcher.
 */
void
xesam_g_searcher_hits_added		(XesamGSearcher		*self,
								 const gchar		*search_handle,
								 guint				count)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER(self));
	g_return_if_fail (search_handle != NULL);
	
	g_signal_emit (self, xesam_g_searcher_signals[HITS_ADDED], 0,
				   search_handle, count);
}

/**
 * xesam_g_searcher_hits_modified
 * @self: The searcher to emit the signal on
 * @search_handle: The search handle on which hits where updated
 * @hit_ids: Array of %guint<!-- -->s containing the ids of the changed hits
 *
 * Emit the %XesamGSearcher::hits-modified signal on the searcher.
 */
void
xesam_g_searcher_hits_modified	(XesamGSearcher		*self,
								 const gchar		*search_handle,
								 const GArray		*hit_ids)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER(self));
	g_return_if_fail (search_handle != NULL);
	g_return_if_fail (hit_ids != NULL);
	
	g_signal_emit (self, xesam_g_searcher_signals[HITS_MODIFIED], 0,
				   search_handle, hit_ids);
}

/**
 * xesam_g_searcher_hits_removed
 * @self: The searcher to emit the signal on
 * @search_handle: The search handle on which hits where updated
 * @hit_ids: Array of %guint<!-- -->s containing the ids of the removed hits
 *
 * Emit the %XesamGSearcher::hits-removed signal on the searcher.
 */
void
xesam_g_searcher_hits_removed	(XesamGSearcher		*self,
								 const gchar		*search_handle,
								 const GArray		*hit_ids)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER(self));
	g_return_if_fail (search_handle != NULL);
	g_return_if_fail (hit_ids != NULL);
	
	g_signal_emit (self, xesam_g_searcher_signals[HITS_REMOVED], 0,
				   search_handle, hit_ids);
}

/**
 * xesam_g_searcher_search_done
 * @self: The searcher to emit the signal on
 * @search_handle: The search handle for which the search has completed on the
 *                 server
 *
 * Emit the %XesamGSearcher::search-done signal on the searcher.
 */
void
xesam_g_searcher_search_done	(XesamGSearcher		*self,
								 const gchar		*search_handle)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER(self));
	g_return_if_fail (search_handle != NULL);
	
	g_signal_emit (self, xesam_g_searcher_signals[SEARCH_DONE], 0,
				   search_handle);
}

/**
 * xesam_g_searcher_state_changed
 * @self: The searcher to emit the signal on
 * @state_info: %NULL terminated string array of length 2. The first entry should
 *              be a valid state as defined by the Xesam Search 1.0 specification
 *              and the second entry should be parsable as an integer between
 *              0 and 100.
 *
 * Emit the %XesamGSearcher::state_changed signal on the searcher.
 */
void
xesam_g_searcher_state_changed	(XesamGSearcher		*self,
								 const GStrv		state_info)
{
	g_return_if_fail (XESAM_IS_G_SEARCHER(self));
	g_return_if_fail (state_info != NULL);
	
	g_signal_emit (self, xesam_g_searcher_signals[STATE_CHANGED], 0,
				   state_info);
}
