/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2008 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include "xesam-g-hits.h"
#include "xesam-g-sequential-hits.h"
#include "xesam-g-search.h"
#include "xesam-g-globals-private.h"
#include "xesam-g-utils.h"
#include "xesam-g-debug-private.h"

/**
 * SECTION:xesam-g-sequential-hits
 * @short_description: Package private container used to model a collection of sequential hits
 *
 * Container for conveniently handling a collection of sequential #XesamGHit
 * objects.
 *
 * #XesamGSequentialHits is available since 1.0
 */

struct _XesamGSequentialHitsPrivate {
	guint batch_offset;
};

/* Private methods */

static guint		xesam_g_sequential_hits_get_hit_id_real(XesamGHits *hits,
															guint		offset);

static XesamGHit*	xesam_g_sequential_hits_next_real	 	(XesamGHits *hits,
															 XesamGHit	*hit);

G_DEFINE_TYPE (XesamGSequentialHits, xesam_g_sequential_hits, XESAM_TYPE_G_HITS);

static void
xesam_g_sequential_hits_init (XesamGSequentialHits *hits)
{
	XesamGSequentialHitsPrivate	*priv;
	
	priv = g_new0(XesamGSequentialHitsPrivate, 1);
	priv->batch_offset = 0;
	
	hits->priv = priv;
}

static void
xesam_g_sequential_hits_finalize (GObject *object)
{
	XesamGSequentialHits		*hits;
	XesamGSequentialHitsPrivate	*priv;

	hits = XESAM_G_SEQUENTIAL_HITS (object);
	priv = hits->priv;
	
	g_free (priv);
	
	G_OBJECT_CLASS (xesam_g_sequential_hits_parent_class)->finalize (object);
}

static void
xesam_g_sequential_hits_class_init (XesamGSequentialHitsClass *klass)
{
	GObjectClass	*object_class = G_OBJECT_CLASS (klass);
	XesamGHitsClass	*hits_class = XESAM_G_HITS_CLASS (klass);

	object_class->finalize = xesam_g_sequential_hits_finalize;
	
	hits_class->get_hit_id = xesam_g_sequential_hits_get_hit_id_real;
	hits_class->next = xesam_g_sequential_hits_next_real;
}

static guint
xesam_g_sequential_hits_get_hit_id_real (XesamGHits	*base,
										 guint		offset)
{
	XesamGSequentialHits	*hits;
	
	g_return_val_if_fail (XESAM_IS_G_SEQUENTIAL_HITS (base), 0);
	
	hits = XESAM_G_SEQUENTIAL_HITS (base);
	
	return hits->priv->batch_offset + offset;
}

static XesamGHit*
xesam_g_sequential_hits_next_real (XesamGHits	*base,
								   XesamGHit	*hit)
{
	XesamGSequentialHits	*hits;
	guint					id, offset, count;
	
	g_return_val_if_fail (XESAM_IS_G_SEQUENTIAL_HITS(base), NULL);
	
	hits = XESAM_G_SEQUENTIAL_HITS (base);
	
	if (hit == NULL)
		return xesam_g_hits_get (XESAM_G_HITS(hits), 0);
	
	g_return_val_if_fail (XESAM_IS_G_HIT (hit), NULL);
	
	id = xesam_g_hit_get_id (hit);
	offset = id - hits->priv->batch_offset;
	count = xesam_g_hits_get_count (hits);
	
	if (offset < 0 || offset >= count ) {
		g_critical ("Hit with id '%d' passed to xesam_g_hits_next() does not "
					"belong in given XesamGHits object with offset %d and "
					"count %d",
					xesam_g_hit_get_id (hit),
					hits->priv->batch_offset,
					xesam_g_hits_get_count (base));
		return NULL;
	}
	
	++offset;
	if (offset == count)
		return NULL; /* This was the last hit */
	
	return xesam_g_hits_get (base, offset);
	
}

/**
 * xesam_g_sequential_hits_new
 * @search: Search to create a hits object for
 * @batch_offset: The offset into the total result set of the parent search
 *                from which these hits where read
 * @field_map: A hash map mapping field names to integer offsets into @hits_data
 * @hits_data: An array of %GPtrArray<!-- -->s with hit data as per
 *             #XesamGSession:hit-fields
 * @notify_search: Whether or not to call xesam_g_search_continue() when the 
 *                 hits are accessed the first time
 * @user_data: User supplied data to associate with this hits object
 * @user_data_destroy: Called to free @user_data when the hits object is finalized.
 *                     Set to %NULL to ignore
 * @returns: A newly allocated sequential hits object
 *
 * Package private method declared in xesam-g-globals-private.h
 *
 * The SequentialHits object is assumed to take ownership of the passed
 * %GPtrArray containing the hit data as well as refering @field_map
 */
XesamGHits*
xesam_g_sequential_hits_new (XesamGSearch	*search, 
							 guint			batch_offset,
							 GHashTable		*field_map,
							 GPtrArray		*hits_data,
							 gboolean		notify_search,
							 gpointer		user_data,
							 GDestroyNotify user_data_destroy)
{	
	XesamGSequentialHits	*hits;
	
	g_return_val_if_fail (XESAM_IS_G_SEARCH (search), NULL);
	g_return_val_if_fail (field_map != NULL, NULL);
	g_return_val_if_fail (hits_data != NULL, NULL);
	
	hits = XESAM_G_SEQUENTIAL_HITS (g_object_new (XESAM_TYPE_G_SEQUENTIAL_HITS,
												  "search", search,
												  "field-map", field_map,
												  "hits-data", hits_data,
												  "notify-search", notify_search,
												  "user-data", user_data,
												  "user-data-destroy", user_data_destroy,
												  NULL));
	hits->priv->batch_offset = batch_offset;
		
	/* Our parent class XesamGHit will ref the search object */
	
	return XESAM_G_HITS (hits);
}

