/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2007 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include "xesam-g-session.h"
#include "xesam-g-search.h"
#include "xesam-g-searcher.h"
#include "xesam-glib-globals.h"
#include "xesam-g-globals-private.h"
#include "xesam-g-debug-private.h"
#include "xesam-g-utils.h"
#include "xesam-marshal.h"

/**
 * SECTION:xesam-g-session
 * @short_description: A client connection to a search engine
 *
 * #XesamGSession is the client side representation of a connection to a search
 * engine. Every Xesam client application must first set up session before any
 * searches can be spawned. Given a session searches are spawned via
 * xesam_g_session_new_search() or xesam_g_session_new_search_from_text().
 *
 * Before any searches are spawned the client should set any relevant properties
 * on the session. The session properties are used to control how searches behave
 * and once the first search has been spawned the session properties can not be
 * changed again on that session.
 *
 * #XesamGSession is available since 1.0
 */

/* 
 * Used to cache session properties, we use value pointers instead of real
 * types to allow NULL to indicate non-looked up values.
 */
typedef struct {
	GValue	*live;					// Boolean
	GValue	*hit_fields;			// GStrv
	GValue	*hit_fields_extended;	// GStrv
	GValue	*hit_snippet_length;	// UInt
	GValue	*sort_primary;			// String
	GValue	*sort_secondary;		// String
	GValue	*sort_order;			// String
	GValue	*vendor_id;				// String
	GValue	*vendor_version;		// UInt
	GValue	*vendor_display;		// String
	GValue	*vendor_xesam;			// UInt
	GValue	*vendor_ontology_fields;  // GStrv
	GValue	*vendor_ontology_contents;// GStrv
	GValue	*vendor_ontology_sources; // GStrv
	GValue	*vendor_extensions;		// GStrv
	GValue	*vendor_ontologies;		// GPtrArray(GStrv)
	GValue	*vendor_maxhits;		// UInt
} XesamGSessionProperties;

struct _XesamGSessionPrivate {
	XesamGSessionProperties *props;
	XesamGSearcher			*searcher;
	GError					*last_error;
	gchar 					*id;
	gboolean				ready;	
	gboolean				closed;
	gboolean				has_spawned_search;
	gulong					on_ready_close_handler;
};

typedef enum {
	PRQ_TYPE_SET,
	PRQ_TYPE_GET,
	PRQ_TYPE_SYNC
} PropRequestOpType;

typedef struct {
	XesamGSession		*session;
	GParamSpec			*pspec;
	GValue				*value;		// Used for async "set" ops when session is not ready
	gulong				handler_id; // Used if we need to disconnect from a signal
	guint				prop_id;
	PropRequestOpType   op_type;
} PropRequestCtx;

/* Property ids */
enum
{
	PROP_0,

	PROP_LIVE,
	PROP_HIT_FIELDS,
	PROP_HIT_FIELDS_EXTENDED,
	PROP_HIT_SNIPPET_LENGTH,
	PROP_SORT_PRIMARY,
	PROP_SORT_SECONDARY,
	PROP_SORT_ORDER,
	PROP_VENDOR_ID,
	PROP_VENDOR_VERSION,
	PROP_VENDOR_DISPLAY,
	PROP_VENDOR_XESAM,
	PROP_VENDOR_ONTOLOGY_FIELDS,
	PROP_VENDOR_ONTOLOGY_CONTENTS,
	PROP_VENDOR_ONTOLOGY_SOURCES,
	PROP_VENDOR_EXTENSIONS,
	PROP_VENDOR_ONTOLOGIES,
	PROP_VENDOR_MAXHITS,
	
	LAST_PROPERTY
};

/* Property names, we need these to be able to do reverse lookups
 * properpty-name -> property-id, and this functionality
 * is not exposed via gobjects api. This is used in sync_property 
 */
static const gchar *PROP_NAMES[] = {
	"PROP_0",
	
	"search-live",
	"hit-fields",
	"hit-fields-extended",
	"hit-snippet-length",
	"sort-primary",
	"sort-secondary",
	"sort-order",
	"vendor-id",
	"vendor-version",
	"vendor-display",
	"vendor-xesam",
	"vendor-ontology-fields",
	"vendor-ontology-contents",
	"vendor-ontology-sources",
	"vendor-extensions",
	"vendor-ontologies",
	"vendor-maxhits",
	
	NULL
};

/* Signals */
enum
{
	READY,
	CLOSED,
	ERROR,
	
	LAST_SIGNAL
};

static guint g_session_signals[LAST_SIGNAL] = { 0 };

/* Private methods */
static void _session_received_cb 					(XesamGSearcher		*searcher,
													 gchar				*session_handle,
													 gpointer			data,
													 GError				*error);

static void _set_session_ready						(XesamGSession		*session);

static void _closed_cb 								(XesamGSearcher		*searcher,
													 gpointer			data,
													 GError				*error);

static PropRequestCtx* 
			_property_context_new					(XesamGSession		*session,
													 GParamSpec			*pspec,
													 guint				prop_id,
													 const GValue		*value,
													 PropRequestOpType  op_type);

static void _property_context_destroy 				(PropRequestCtx		*prop_ctx);

static void _set_property_cb 						(XesamGSearcher		*searcher,
													 gchar				*prop_name,
													 GValue				*value,
													 gpointer			data,
													 GError				*error);

static gint _get_property_id 						(const gchar		*property);

static void _receive_property	 					(PropRequestCtx		*prop_ctx);

static void _sync_required_properties				(XesamGSession		*session);

static gboolean _has_required_properties			(XesamGSession		*session);

static void _set_property_delayed					(PropRequestCtx *prop_ctx);

static void _set_last_error							(XesamGSession  *session,
													 GError			*error);

static void _set_new_error							(XesamGSession	*session,
													 gint			code,
													 const gchar	*format,
													 ...);


#define free_g_value_if_null(val)\
if (val != NULL) {\
		g_value_unset (val);\
		g_free (val);\
}

G_DEFINE_TYPE (XesamGSession, xesam_g_session, G_TYPE_OBJECT);

static void
xesam_g_session_finalize (GObject *session_obj)
{
	XesamGSession 			*session;
	XesamGSessionPrivate	*priv;
	XesamGSessionProperties *props;
	
	session = XESAM_G_SESSION (session_obj);
	priv = session->priv;
	props = priv->props;
	
	xesam_g_debug_object (session, "Finalizing");
	
	if (!priv->closed) {
		xesam_g_debug_object (session, "Session not closed in finalization."
							  " Closing now.");
		/* Just fire a close request on the searcher, ignoring the response */
		xesam_g_searcher_close_session (priv->searcher, priv->id,
										NULL, NULL);
	}
	
	/* Free private members */
	if (priv->searcher)
		g_object_unref (priv->searcher);
	if (priv->id)
		g_free (priv->id);
	
	if (priv->last_error)
		g_error_free (priv->last_error);
	
	/* Free properties */
	free_g_value_if_null (props->live);
	free_g_value_if_null (props->hit_fields);
	free_g_value_if_null (props->hit_fields_extended);
	free_g_value_if_null (props->hit_snippet_length);
	free_g_value_if_null (props->sort_primary);
	free_g_value_if_null (props->sort_secondary);
	free_g_value_if_null (props->sort_order);
	free_g_value_if_null (props->vendor_id);
	free_g_value_if_null (props->vendor_version);
	free_g_value_if_null (props->vendor_display);
	free_g_value_if_null (props->vendor_xesam);
	free_g_value_if_null (props->vendor_ontology_fields);
	free_g_value_if_null (props->vendor_ontology_contents);
	free_g_value_if_null (props->vendor_ontology_sources);
	free_g_value_if_null (props->vendor_extensions);
	free_g_value_if_null (props->vendor_ontologies);
	free_g_value_if_null (props->vendor_maxhits);
	
	
	/* Free fundamental structs */
	g_free (props);
	g_free (priv);
	
	G_OBJECT_CLASS (xesam_g_session_parent_class)->finalize (session_obj);
}

static void
xesam_g_session_set_property (GObject 		*object, guint 		prop_id, 
							  const GValue 	*value,  GParamSpec *pspec)
{	
	XesamGSession 	*session;
	PropRequestCtx 	*prop_ctx;
	gchar			*prop_name;
	
	g_return_if_fail (XESAM_IS_G_SESSION (object));
	
	if (prop_id <= PROP_0 || prop_id >= LAST_PROPERTY) {
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		return;
	}
	
	session = XESAM_G_SESSION (object);
	if (session->priv->has_spawned_search) {
		g_critical ("It is illegal to change session properties"
					" after the first search has been spawned.");
		return;
	}
	
	prop_ctx = _property_context_new (session, pspec, prop_id,
									  value, PRQ_TYPE_SET);
	
	/* Block all property change notifications until the callback returns
	 * thawed in _set_property_cb possibly via a roundtrip to 
	 * _set_property_delayed */
	g_object_freeze_notify (object);
	
	/* If the session is not ready yet, wait for ready signal
	 * and set the property then */
	if (!xesam_g_session_is_ready(session)) {
		prop_ctx->handler_id = g_signal_connect_swapped (session, "ready",
														 (GCallback) _set_property_delayed,
														 prop_ctx);
		return;
	}
	
	prop_name = g_property_to_xesam_property (pspec->name);
	
	xesam_g_searcher_set_property (session->priv->searcher,
								   session->priv->id,
								   prop_name,
								   value,
								   _set_property_cb,
								   prop_ctx);
	
	g_free (prop_name);
}

static void
xesam_g_session_get_property (GObject		*object,
							  guint			prop_id,
							  GValue		*value,
							  GParamSpec	*pspec)
{
	XesamGSession			*session;
	XesamGSessionProperties *props;
	
	g_return_if_fail (XESAM_IS_G_SESSION (object));
	
	session = XESAM_G_SESSION (object);
	props = session->priv->props;
	
	if (!session->priv->ready) {
		g_critical ("Getting property '%s', but session is not ready",
					pspec->name);
		return;
	}
	
	/* The idea in this method is to check if we have the property value cached.
	 * Nobody is going to change the properties remotely, so we can be sure
	 * that all changes goes through us - hence the cache is always consistent.
	 * 
	 * If we don't have a cached copy we are screwed because it is a design
	 * policy of xesam-glib to never do blocking remote calls.
	 */
	
	if (!session->priv->id) {
		/* We haven't received the session id yet. We need the id 
		 * in the GetProperty method. */
		g_critical ("get_property on non-ready session. You must wait for the "
					"'ready' signal before calling get_property on a "
					"XesamGSession");
		return;
	}
	
	switch (prop_id)
	{
	case PROP_LIVE:
		if (props->live) {
			g_value_copy (props->live, value);
			return;
		}
		break;	
	case PROP_HIT_FIELDS:
		if (props->hit_fields) {
			g_value_copy (props->hit_fields, value);
			return;
		}
		break;
	case PROP_HIT_FIELDS_EXTENDED:
		if (props->hit_fields_extended) {
			g_value_copy (props->hit_fields_extended, value);
			return;
		}
		break;
	case PROP_HIT_SNIPPET_LENGTH:
		if (props->hit_snippet_length) {
			g_value_copy (props->hit_snippet_length, value);
			return;
		}
		break;
	case PROP_SORT_PRIMARY:
		if (props->sort_primary) {
			g_value_copy (props->sort_primary, value);
			return;
		}
		break;
	case PROP_SORT_SECONDARY:
		if (props->sort_secondary) {
			g_value_copy (props->sort_secondary, value);
			return;
		}
		break;
	case PROP_SORT_ORDER:
		if (props->sort_order) {
			g_value_copy (props->sort_order, value);
			return;
		}
		break;
	case PROP_VENDOR_ID:
		if (props->vendor_id) {
			g_value_copy (props->vendor_id, value);
			return;
		}
		break;
	case PROP_VENDOR_VERSION:
		if (props->vendor_version) {
			g_value_copy (props->vendor_version, value);
			return;
		}
		break;
	case PROP_VENDOR_DISPLAY:
		if (props->vendor_display) {
			g_value_copy (props->vendor_display, value);
			return;
		}
		break;
	case PROP_VENDOR_XESAM:
		if (props->vendor_xesam) {
			g_value_copy (props->vendor_xesam, value);
			return;
		}
		break;
	case PROP_VENDOR_ONTOLOGY_FIELDS:
		if (props->vendor_ontology_fields) {
			g_value_copy (props->vendor_ontology_fields, value);
			return;
		}
		break;
	case PROP_VENDOR_ONTOLOGY_CONTENTS:
		if (props->vendor_ontology_contents) {
			g_value_copy (props->vendor_ontology_contents, value);
			return;
		}
		break;
	case PROP_VENDOR_ONTOLOGY_SOURCES:
		if (props->vendor_ontology_sources) {
			g_value_copy (props->vendor_ontology_sources, value);
			return;
		}
		break;
	case PROP_VENDOR_EXTENSIONS:
		if (props->vendor_extensions) {
			g_value_copy (props->vendor_extensions, value);
			return;
		}
		break;
	case PROP_VENDOR_ONTOLOGIES:
		if (props->vendor_ontologies) {
			g_value_copy (props->vendor_ontologies, value);
			return;
		}
		break;
	case PROP_VENDOR_MAXHITS:
		if (props->vendor_maxhits) {
			g_value_copy (props->vendor_maxhits, value);
			return;
		}
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		return;
		break;
	}
	
	g_critical ("Attempt to retrieve non-synchronized property '%s'."
				"Properties should not be read until a notification on that"
				"property has been emitted. Call fx. sync_property() first.",
				PROP_NAMES[prop_id]);
}

/**
 * xesam_g_session_sync_property
 * @session: Session on which to synchronize a property
 * @property: Name of the property to synchronize
 *
 * Asynchronously update the cached value of a session property.
 * When the requested property has been retrieved a notification
 * will be emitted for this property.
 *
 * It is not legal to to call g_object_get() on a session before a
 * notification has been emitted on the requested property.
 *
 * You may invoke this method even though the session is not ready. The request
 * will just be queued until the session becomes ready.
 *
 * Calling g_object_set() will also synchronize a property. To test whether
 * a property has been synced call xesam_g_session_test_property().
 */
void
xesam_g_session_sync_property (XesamGSession *session, gchar *property)
{
	PropRequestCtx 	*prop_ctx;
	GParamSpec		*pspec;
	GObjectClass	*session_class;
	gchar			*prop_name;
	gint			prop_id;
	
	g_return_if_fail (XESAM_IS_G_SESSION(session));
	g_return_if_fail (property != NULL);
	
	prop_id = _get_property_id (property);
	session_class = G_OBJECT_CLASS(XESAM_G_SESSION_GET_CLASS(session));
	pspec = g_object_class_find_property (session_class, property);
	
	if (prop_id == -1) {
		g_critical ("Request to sync unknown property '%s'", property);
		return;
	}
	
	/* If we don't have a session id enqueue the request.
	 * We cannot check for session->priv->ready, since that implies that we
	 * have synched the required properties too. This would cause a deadlock.*/
	if (!session->priv->id) {
		g_object_ref (session);
		g_signal_connect_data (session, "ready",
							   G_CALLBACK(xesam_g_session_sync_property),
							   g_strdup (property),
							   (GClosureNotify)g_free,
							   G_CONNECT_AFTER);
		return;
	}
	
	/* Make sure we disconnect any signal handlers we set up if the session
	 * was not ready to begin with (ie this is a delayed call) */
	int num_disconn =
		g_signal_handlers_disconnect_by_func (session,
											  xesam_g_session_sync_property,
											  property);
	
	if (num_disconn == 1)
		g_object_unref (session);
	else if (num_disconn > 1) {
		g_critical ("%d signal handlers connected to delayed sync of "
					"property '%s'. Should only be 1. This is a bug in "
					"xesam-glib.", num_disconn, property);
		return;
	}
	
	prop_ctx = _property_context_new (session, pspec, prop_id, 
									  NULL, PRQ_TYPE_SYNC);
	
	prop_name = g_property_to_xesam_property (pspec->name);
	
	/* Asynchronously read the property from the searcher */
	xesam_g_searcher_get_property (session->priv->searcher,
								   session->priv->id,
								   prop_name,
								   _set_property_cb,
								   prop_ctx);
	
	g_free (prop_name);
}

/**
 * xesam_g_session_test_property
 * @session: The session on which to test a property
 * @property: Name of property to test for
 *
 * Test whether a given property has been synchronized from the server.
 * It is illegal to call g_object_get() on a property
 * that has not been synced.
 *
 * A property is synchronized if you have either written to it or if you have
 * called xesam_g_session_sync_property() on it.
 *
 * Returns: %TRUE if and only if the property is ready for reading with 
 *          g_object_get ().
 */
gboolean
xesam_g_session_test_property (XesamGSession 	*session,
							   gchar 			*property)
{
	int						prop_id;
	XesamGSessionProperties *props;
	
	g_return_val_if_fail (XESAM_IS_G_SESSION (session), FALSE);
	g_return_val_if_fail (property != NULL, FALSE);
	
	props = session->priv->props;
	prop_id = _get_property_id (property);
	
	if (prop_id == -1) {
		g_critical ("Can not test unknown property '%s'", property);
		return FALSE;
	}
	
	switch (prop_id)
	{
	case PROP_LIVE:
		return props->live != NULL;
		break;	
	case PROP_HIT_FIELDS:
		return props->hit_fields != NULL;
		break;
	case PROP_HIT_FIELDS_EXTENDED:
		return props->hit_fields_extended != NULL;
		break;
	case PROP_HIT_SNIPPET_LENGTH:
		return props->hit_snippet_length != NULL;
		break;
	case PROP_SORT_PRIMARY:
		return props->sort_primary != NULL;
		break;
	case PROP_SORT_SECONDARY:
		return props->sort_secondary != NULL;
		break;
	case PROP_SORT_ORDER:
		return props->sort_order != NULL;
		break;
	case PROP_VENDOR_ID:
		return props->vendor_id != NULL;
		break;
	case PROP_VENDOR_VERSION:
		return props->vendor_version != NULL;
		break;
	case PROP_VENDOR_DISPLAY:
		return props->vendor_display != NULL;
		break;
	case PROP_VENDOR_XESAM:
		return props->vendor_xesam != NULL;
		break;
	case PROP_VENDOR_ONTOLOGY_FIELDS:
		return props->vendor_ontology_fields != NULL;
		break;
	case PROP_VENDOR_ONTOLOGY_CONTENTS:
		return props->vendor_ontology_contents != NULL;
		break;
	case PROP_VENDOR_ONTOLOGY_SOURCES:
		return props->vendor_ontology_sources != NULL;
		break;
	case PROP_VENDOR_EXTENSIONS:
		return props->vendor_extensions != NULL;
		break;
	case PROP_VENDOR_ONTOLOGIES:
		return props->vendor_ontologies != NULL;
		break;
	case PROP_VENDOR_MAXHITS:
		return props->vendor_maxhits != NULL;
		break;
	default:
		g_critical ("Can not test invalid property '%s'", property);
		return FALSE;
		break;
	}
}

static void
xesam_g_session_class_init (XesamGSessionClass *klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = xesam_g_session_finalize;
	object_class->set_property = xesam_g_session_set_property;
	object_class->get_property = xesam_g_session_get_property;	
	
	/**
	 * XesamGSession:search-live:
	 * 
	 * If %TRUE any  #XesamGSearch spawned on this session will continue to
	 * monitor the query and notify about changes via the relevant signal
	 * on the search object.
	 *
	 * The #XesamGSearch objects will only use the #XesamGSearch::hits-modified
	 * and #XesamGSearch::hits-removed signals if this property is %TRUE.
	 */
	g_object_class_install_property (object_class,
	                                 PROP_LIVE,
	                                 g_param_spec_boolean ("search-live",
														   "Search live",
														   "Whether the search should do live updates to the result set",
															FALSE,
															G_PARAM_READWRITE));
	
	g_object_class_install_property (object_class,
	                                 PROP_HIT_FIELDS,
	                                 g_param_spec_boxed ("hit-fields",
	                                                     "Hit data",
	                                                     "Fields to return for each hit",
	                                                     G_TYPE_STRV,
														 G_PARAM_READWRITE));

	
	g_object_class_install_property (object_class,
	                                 PROP_HIT_FIELDS_EXTENDED,
	                                 g_param_spec_boxed ("hit-fields-extended",
														 "Extended hit data",
														 "Extra fields fetched on request",
														 G_TYPE_STRV,
														 G_PARAM_READWRITE));
	
	g_object_class_install_property (object_class,
	                                 PROP_HIT_SNIPPET_LENGTH,
	                                 g_param_spec_int ("hit-snippet-length",
													   "Length of hit summaries",
													   "Length of hit summaries in characters",
													   0, 1000, 200,
													   G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
	                                 PROP_SORT_PRIMARY,
	                                 g_param_spec_string ("sort-primary",
														  "Primary sort key",
														  "Name of field to sort hits after",
														  "xesam:score",
														  G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
	                                 PROP_SORT_SECONDARY,
	                                 g_param_spec_string ("sort-secondary",
														  "Secondary sort key",
														  "Name of field used to sort if primary sort keys equal",
														  "",
														  G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
	                                 PROP_SORT_ORDER,
	                                 g_param_spec_string ("sort-order",
														  "Sort order",
														  "The order in whic to sort results",
														  "descending",
														  G_PARAM_READWRITE));
	
	g_object_class_install_property (object_class,
	                                 PROP_VENDOR_ID,
	                                 g_param_spec_string ("vendor-id",
														 "Unique vendor id",
														 "A string uniquely identofying the search engine vendor",
														 "Unknown",
														 G_PARAM_READABLE));
	g_object_class_install_property (object_class,
	                                 PROP_VENDOR_VERSION,
	                                 g_param_spec_uint ("vendor-version",
													    "Vendor version number",
													    "Integer representation of the vendors version number",
													    0, G_MAXUINT, 0,
														G_PARAM_READABLE));
	g_object_class_install_property (object_class,
	                                 PROP_VENDOR_DISPLAY,
	                                 g_param_spec_string ("vendor-display",
														  "Short description of vendor",
														  "A short string describing the vendor suitable for display",
														  "Unknown",
														  G_PARAM_READABLE));
	g_object_class_install_property (object_class,
	                                 PROP_VENDOR_XESAM,
	                                 g_param_spec_uint ("vendor-xesam",
														"Xesam version of search engine",
														"The XESAM search specification version number implemented by the search engine",
														0, G_MAXUINT, 1,
														G_PARAM_READABLE));
	g_object_class_install_property (object_class,
	                                 PROP_VENDOR_ONTOLOGY_FIELDS,
	                                 g_param_spec_boxed ("vendor-ontology-fields",
														 "Fields supported by search engine",
														 "A list of all field names supported by the search engine",
														 G_TYPE_STRV,
														 G_PARAM_READABLE));
	g_object_class_install_property (object_class,
	                                 PROP_VENDOR_ONTOLOGY_CONTENTS,
	                                 g_param_spec_boxed ("vendor-ontology-contents",
														 "Content types supported by the search engine",
														 "A list of all content types supported by the search engine",
														 G_TYPE_STRV,
														 G_PARAM_READABLE));
	g_object_class_install_property (object_class,
	                                 PROP_VENDOR_ONTOLOGY_SOURCES,
	                                 g_param_spec_boxed ("vendor-ontology-sources",
														 "Source types supported by the search engine",
														 "A list of all source types supported by the search engine",
														 G_TYPE_STRV,
														 G_PARAM_READABLE));
	g_object_class_install_property (object_class,
	                                 PROP_VENDOR_EXTENSIONS,
	                                 g_param_spec_boxed ("vendor-extensions",
														 "Xesam query extensions supported by vendor",
														 "A list of all xesam query language extensions supported by the search engine",
														 G_TYPE_STRV,
														 G_PARAM_READABLE));
	
	/**
	 * XesamGSession:vendor-ontologies
	 * 
	 * A %GPtrArray of GStrv<!-- -->s. Each %GStrv has three elements
	 * and describes an ontology. The members of the ontology descriptors are
	 * unique_name, version, and path.
	 */
	g_object_class_install_property (object_class,
	                                 PROP_VENDOR_ONTOLOGIES,
									 g_param_spec_pointer ("vendor-ontologies",
														   "Ontologies supported by the vendor",
														   "An array of ontology description triples (unique_name, version, path)",
														   G_PARAM_READABLE));
	g_object_class_install_property (object_class,
	                                 PROP_VENDOR_MAXHITS,
	                                 g_param_spec_uint ("vendor-maxhits",
														"Maximum number of hits retrievable",
														"The max number of hits retrievable from the search engine. The actual number of hits might be bigger than this.",
														0, G_MAXUINT, 1,
														G_PARAM_READABLE));
	
	/**
	 * XesamGSession::ready
	 * 
	 * Emitted when the session handle has been retrieved from the underlying
	 * #XesamGSearcher.
	 *
	 * All operations on a session require a session handle, so you should
	 * generally wait for this signal before doing anything with your session.
	 */
	g_session_signals[READY] =
		g_signal_new ("ready",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__VOID,
		              G_TYPE_NONE, 0);
	
	g_session_signals[CLOSED] =
		g_signal_new ("closed",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__VOID,
		              G_TYPE_NONE, 0);
	
	/**
	 * XesamGSession::error
	 *
	 * Emitted when an error has occured. Only errors related to communications
	 * with the search engine is reported this way. Internal- or programming
	 * errors are not reported via this signal.
	 *
	 * To inspect the last error call xesam_g_session_get_last_error().
	 */
	g_session_signals[ERROR] =
		g_signal_new ("error",
		              G_OBJECT_CLASS_TYPE (klass),
		              G_SIGNAL_RUN_LAST,
		              0,
		              NULL, NULL,
		              g_cclosure_marshal_VOID__VOID,
		              G_TYPE_NONE, 0);
}

static void
xesam_g_session_init (XesamGSession *session)
{
	XesamGSessionPrivate 	*priv;
	XesamGSessionProperties *props;
	
	priv = g_new0 (XesamGSessionPrivate, 1);
	props = g_new0 (XesamGSessionProperties, 1);
	
	priv->ready = FALSE;
	priv->closed = FALSE;
	
	priv->id = NULL;
	
	priv->props = props;
	session->priv = priv;
	
	xesam_g_debug_object (session, "Created");
}

static PropRequestCtx*
_property_context_new (XesamGSession *session, GParamSpec 	*pspec, 
					   guint 		 prop_id,  const GValue *value,
					   PropRequestOpType op_type)
{
	PropRequestCtx *prop_ctx;
	
	g_return_val_if_fail (XESAM_IS_G_SESSION(session), NULL);
	g_return_val_if_fail (G_IS_PARAM_SPEC(pspec), NULL);
	
	if (op_type == PRQ_TYPE_SET)
		g_return_val_if_fail (G_IS_VALUE(value), NULL);
	
	g_param_spec_ref (pspec);
	g_object_ref (session);
	
	prop_ctx = g_new0 (PropRequestCtx, 1);
	prop_ctx->session = session;
	prop_ctx->pspec = pspec;
	prop_ctx->prop_id = prop_id;
	prop_ctx->op_type = op_type;
	
	if (value) {
		prop_ctx->value = g_new0 (GValue, 1);
		prop_ctx->value = g_value_init (prop_ctx->value, G_VALUE_TYPE (value));
		g_value_copy (value, prop_ctx->value);
	}
	
	return prop_ctx;
}

static void
_property_context_destroy (PropRequestCtx *prop_ctx)
{
	g_return_if_fail (prop_ctx != NULL);
	
	if (prop_ctx->handler_id)
		g_signal_handler_disconnect (prop_ctx->session, prop_ctx->handler_id);
	
	g_param_spec_unref (prop_ctx->pspec);
	g_object_unref (prop_ctx->session);
	
	if (prop_ctx->value) {
		g_value_unset (prop_ctx->value);
		g_free (prop_ctx->value);
	}
	
	g_free (prop_ctx);
}

/*
 * This method *must* only be called as a callback with a fully qualified
 * property context - ie ctx->value and ctx->handler_id must be set,
 * as they are in xesam_g_session_set_property when the session is not ready.
 */
static void
_set_property_delayed (PropRequestCtx *prop_ctx)
{
	gchar 			*prop_name;
	
	g_return_if_fail (prop_ctx != NULL);
	g_return_if_fail (xesam_g_session_is_ready (prop_ctx->session));
	g_return_if_fail (prop_ctx->value != NULL);
	
	prop_name = g_property_to_xesam_property (prop_ctx->pspec->name);
	
	xesam_g_searcher_set_property (prop_ctx->session->priv->searcher,
								   prop_ctx->session->priv->id,
								   prop_name,
								   prop_ctx->value,
								   _set_property_cb,
								   prop_ctx);
	
	g_free (prop_name);
}

/*
 * Callback for when a session property has been set
 */
static void
_receive_property (PropRequestCtx *prop_ctx)
{
	XesamGSession					*session;
	XesamGSessionProperties			*props;
	GValue							*target_prop;
	
	g_return_if_fail (prop_ctx != NULL);
	g_return_if_fail (XESAM_IS_G_SESSION(prop_ctx->session));
	g_return_if_fail (G_IS_PARAM_SPEC(prop_ctx->pspec));
	
	session = prop_ctx->session;
	props = session->priv->props;
	
	switch (prop_ctx->prop_id)
	{
	case PROP_LIVE:
		target_prop = init_value_if_null (&(props->live), G_TYPE_BOOLEAN);
		break;	
	case PROP_HIT_FIELDS:
		target_prop = init_value_if_null (&(props->hit_fields), G_TYPE_STRV);
		break;
	case PROP_HIT_FIELDS_EXTENDED:
		target_prop = init_value_if_null (&(props->hit_fields_extended), G_TYPE_STRV);
		break;
	case PROP_HIT_SNIPPET_LENGTH:
		target_prop = init_value_if_null (&(props->hit_snippet_length), G_TYPE_UINT);
		break;
	case PROP_SORT_PRIMARY:
		target_prop = init_value_if_null (&(props->sort_primary), G_TYPE_STRING);
		break;
	case PROP_SORT_SECONDARY:
		target_prop = init_value_if_null (&(props->sort_secondary), G_TYPE_STRING);
		break;
	case PROP_SORT_ORDER:
		target_prop = init_value_if_null (&(props->sort_order), G_TYPE_STRING);
		break;
	case PROP_VENDOR_ID:
		target_prop = init_value_if_null (&(props->vendor_id), G_TYPE_STRING);
		break;
	case PROP_VENDOR_VERSION:
		target_prop = init_value_if_null (&(props->vendor_version), G_TYPE_UINT);
		break;
	case PROP_VENDOR_DISPLAY:
		target_prop = init_value_if_null (&(props->vendor_display), G_TYPE_STRING);
		break;
	case PROP_VENDOR_XESAM:
		target_prop = init_value_if_null (&(props->vendor_xesam), G_TYPE_UINT);
		break;
	case PROP_VENDOR_ONTOLOGY_FIELDS:
		target_prop = init_value_if_null (&(props->vendor_ontology_fields), G_TYPE_STRV);
		break;
	case PROP_VENDOR_ONTOLOGY_CONTENTS:
		target_prop = init_value_if_null (&(props->vendor_ontology_contents), G_TYPE_STRV);
		break;
	case PROP_VENDOR_ONTOLOGY_SOURCES:
		target_prop = init_value_if_null (&(props->vendor_ontology_sources), G_TYPE_STRV);
		break;
	case PROP_VENDOR_EXTENSIONS:
		target_prop = init_value_if_null (&(props->vendor_extensions), G_TYPE_STRV);
		break;
	case PROP_VENDOR_ONTOLOGIES:
		target_prop = init_value_if_null (&(props->vendor_ontologies), G_TYPE_POINTER);
		break;
	case PROP_VENDOR_MAXHITS:
		target_prop = init_value_if_null (&(props->vendor_maxhits), G_TYPE_UINT);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (session, prop_ctx->prop_id, 
										   prop_ctx->pspec);
		return;
		break;
	}
	
	/* All props except vendor.ontologies can be handled easily. This is because
	 * an array of arrays of strings is not mapped by the GLib type system.
	 * dbus-glib views it as a custom type, GPtrArray_GStrv, which in turn
	 * is a boxed type. Go figure. */
	if (prop_ctx->prop_id != PROP_VENDOR_ONTOLOGIES)
		g_value_copy (prop_ctx->value, target_prop);
	else
		g_value_set_pointer (target_prop, g_value_get_boxed (prop_ctx->value));
	
	if (!session->priv->ready &&
		_has_required_properties (session))
			_set_session_ready (session);
}

/* async callback for property request
 *
 * IMPORTANT: This method is a callback from a XesamGSearcher which is an
 *            untrusted source, we should be careful to check the args */
static void
_set_property_cb (XesamGSearcher	*searcher,
				  gchar				*prop_name,
				  GValue			*value,
				  gpointer			data,
				  GError			*error)
{
	PropRequestCtx  *ctx;
	
	ctx = (PropRequestCtx*) data;
	
	g_return_if_fail (ctx != NULL);
	g_return_if_fail (XESAM_IS_G_SESSION(ctx->session));
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	
	if (error) {
		_set_last_error (ctx->session, error);
		goto clean_up;
	}
	
	if (prop_name == NULL) {
		_set_new_error (ctx->session,
						XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
						"When setting property '%s' server returned NULL "
						"property name",
						ctx->pspec->name);
	}
	
	/* FIXME: We could check that:
	          prop_name == g_property_to_xesam_property(ctx->pspec->name) */
	
	if (value == NULL) {
		_set_new_error (ctx->session,
						XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
						"When setting property '%s' server returned NULL value",
						ctx->pspec->name);
	}
	
	/* Assume new value in the context */
	if (ctx->value) {
		g_value_unset (ctx->value);
		g_free (ctx->value);
	}
	ctx->value = value;
	
	_receive_property (ctx);
	
	if (ctx->op_type == PRQ_TYPE_SET) {
		/* Unfreeze notifications to emit "notify::<prop_name>"
		 * frozen in xesam_g_session_set_property */
		g_object_thaw_notify (G_OBJECT(ctx->session));
	} else if (ctx->op_type == PRQ_TYPE_SYNC) {
		g_object_notify (G_OBJECT(ctx->session), ctx->pspec->name);
	} else if (ctx->op_type == PRQ_TYPE_GET) {
		/* nothing */
	} else {
		g_critical ("Invalid op_type '%d' for property request context",
					ctx->op_type);
	}
	
	
	
	clean_up:
		if (prop_name) g_free (prop_name);
		_property_context_destroy (ctx);
		/* don't free value, it is owned by ctx at this point */
}

/* get the property id for a given property name 
 * FIXME: Is it really true that glib does not expose this? */
static gint
_get_property_id (const gchar *property)
{
	gint i;
	for (i = 0; i < LAST_PROPERTY; i++) {
		if (g_str_equal(property, PROP_NAMES[i]))
			return i;
	}
	return -1;
}


/**
 * xesam_g_session_new
 * @searcher: Searcher to pass communications through
 * @returns: A newly created #XesamGSession on the provided searcher or %NULL
 *           on error
 *
 * Create a new session on the provided searcher.
 * 
 * Then session is ready for operation when the #XesamGSession::ready signal
 * is emitted on it.
 */
XesamGSession*
xesam_g_session_new (XesamGSearcher *searcher)
{
	XesamGSession   *session;
	
	g_return_val_if_fail (XESAM_IS_G_SEARCHER(searcher), NULL);
	
	session = XESAM_G_SESSION (g_object_new (XESAM_TYPE_G_SESSION, NULL));
	
	session->priv->searcher = searcher;
	g_object_ref (searcher);
	
	/* Start acquiring a session handle */
	xesam_g_searcher_new_session (searcher, _session_received_cb, session);
	
	return session;
}

/* Async callback for receiving a session id from the server
 *
 * IMPORTANT: This is a callback from a XesamGSearcher which is an untrusted
			source. We should be careful to check the args */
static void
_session_received_cb (XesamGSearcher	*searcher,
					  gchar				*session_handle,
					  gpointer			data,
					  GError			*error)
{
	XesamGSession 			*session;
	XesamGSessionPrivate	*priv;
	
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	g_return_if_fail (XESAM_IS_G_SESSION(data));
	
	session = XESAM_G_SESSION (data);
	priv = session->priv;
	
	if (error) {
		g_warning ("Error receiving session handle: %s", error->message);
		_set_last_error (session, error);
		return;
	}
	
	if (session_handle == NULL) {
		_set_new_error (session, XESAM_G_SEARCHER_ERROR_BAD_RESPONSE,
						"When requesting session handle, searcher returned NULL");
		return;
	}
	
	priv->id = session_handle;
	
	/* We are first ready when we have received all required properties */
	_sync_required_properties (session);
}

static void
_sync_required_properties (XesamGSession *session)
{
	g_return_if_fail (XESAM_IS_G_SESSION(session));
	
	/* hit-fields are needed to build the field-map(s) */
	xesam_g_session_sync_property (session, "hit-fields");
	xesam_g_session_sync_property (session, "hit-fields-extended");
}

static gboolean
_has_required_properties (XesamGSession *session)
{
	g_return_val_if_fail (XESAM_IS_G_SESSION(session), FALSE);
	
	return (session->priv->props->hit_fields != NULL &&
			session->priv->props->hit_fields_extended != NULL);
}

static void
_set_session_ready (XesamGSession *session)
{
	g_return_if_fail (XESAM_IS_G_SESSION(session));
	
	g_return_if_fail (!session->priv->ready
					  && _has_required_properties(session));	
		
	session->priv->ready = TRUE;
	xesam_g_debug_object (session, "Ready. Got session handle '%s'",
						  session->priv->id);
	g_signal_emit (session, g_session_signals[READY], 0);
	
}

/**
 * xesam_g_session_get_searcher
 * @session: The session to get the #XesamGSearcher object for.
 * @returns: #XesamGSearcher used by the session
 */
XesamGSearcher*
xesam_g_session_get_searcher (XesamGSession *session)
{
	g_return_val_if_fail (XESAM_IS_G_SESSION (session), NULL);
	
	return session->priv->searcher;
}

/**
 * xesam_g_session_get_id
 * @session: The session to retrieve the unique id for
 * @returns: The unique id for the session. If the session is not ready this
 *           method will return %NULL.
 *
 * Each session is identified on the server via a session id.
 * Use this method to get it.
 *
 * Note that this method returning non-%NULL does not imply that the
 * session is ready.
 */
gchar*
xesam_g_session_get_id (XesamGSession *session)
{
	g_return_val_if_fail (XESAM_IS_G_SESSION (session), NULL);
	
	return session->priv->id;
}

/**
 * xesam_g_session_is_ready
 * @session: The session to check for readiness
 * @returns: %TRUE if and only if the session has received a session id from
 *           the server. This means that %FALSE is returned on error.
 *
 * Check whether a session is ready to spawn searches and/or get and set
 * properties. 
 */
gboolean
xesam_g_session_is_ready (XesamGSession *session)
{
	g_return_val_if_fail (XESAM_IS_G_SESSION (session), FALSE);
	
	return session->priv->ready;
}

/**
 * xesam_g_session_is_closed
 * @session: The session to check
 * @returns: %TRUE if and only if the session has been closed by a call to
 *           #xesam_g_session_close.
 *
 * Check if a session has been closed. Returns %TRUE on error.
 */
gboolean
xesam_g_session_is_closed (XesamGSession *session)
{
	g_return_val_if_fail (XESAM_IS_G_SESSION (session), TRUE);
	
	return session->priv->closed;
}

/**
 * xesam_g_session_close
 * @session: Session to close
 *
 * Close a session. Spawning searches and setting/getting properties on a closed
 * session is illegal.
 */
void
xesam_g_session_close (XesamGSession *session)
{
	XesamGSessionPrivate	*priv;
	
	g_return_if_fail (XESAM_IS_G_SESSION (session));
	
	priv = session->priv;
	
	if (priv->closed && priv->on_ready_close_handler == 0) {
		xesam_g_debug_object (session, "Got close() request, but is already closed");
		return;
	}
	
	priv->closed = TRUE;
	priv->ready = FALSE;	
	
	/* Defer close request if we don't have an id yet */
	if (!priv->id) {
		if (priv->on_ready_close_handler != 0) {
			g_warning ("Session '%s' already closed or has close request pending",
					   "BORK");//priv->id ? priv->id : "<unknown>");
			return;
		}
				
		xesam_g_debug_object (session, "Defering close(), no session handle yet");
		
		priv->on_ready_close_handler =
			g_signal_connect_after (session, "ready",
									G_CALLBACK(xesam_g_session_close),
									NULL);
		g_object_ref (session);
		return;
	}
	
	/* Disconnect signal handler if this is a defered close call */
	if (priv->on_ready_close_handler) {
		g_signal_handler_disconnect (session, priv->on_ready_close_handler);
		priv->on_ready_close_handler = 0;
		g_object_unref (session);
	}
	
	xesam_g_debug_object (session, "Sending Close('%s')", priv->id);
	xesam_g_searcher_close_session (priv->searcher,
									priv->id,
									_closed_cb,
									session);
}

/**
 * xesam_g_session_get_last_error
 * @session: The #XesamGSession to inspect for errors
 * @returns: The last #GError registered by the session or %NULL if no error
 *           has occured
 *
 * When an error occurs it is reported by the #XesamGSession::error signal,
 * after which you can inspect it with this method.
 *
 * Only communication errors with the search engine is reported this way.
 * Programming errors are not reported this way.
 */
const GError*
xesam_g_session_get_last_error  (XesamGSession  *session)
{
	g_return_val_if_fail (XESAM_IS_G_SESSION(session), NULL);
	
	return session->priv->last_error;
}

/**
 * xesam_g_session_new_search
 * @session: Session to spawn a search on
 * @query: Query the search should match
 * @returns: A newly created search object.
 *
 * Create a new #XesamGSearch from a #XesamGQuery.
 */
XesamGSearch*
xesam_g_session_new_search (XesamGSession *session, XesamGQuery *query)
{
	XesamGSearch *search;
	
	session->priv->has_spawned_search = TRUE;
	
	search = xesam_g_search_new (session, query);
	
	return search;
}

/**
 * xesam_g_session_new_search_from_text
 * @session: Session to spawn search on
 * @search_text: A string encoded in the XESAM User Search Language.
 * @returns: A newly created search object
 *
 * Create a new #XesamGSearch from a string composed in the Xesam
 * User Search Language.
 *
 * This method is suitable for use on raw user input.
 */
XesamGSearch*
xesam_g_session_new_search_from_text (XesamGSession *session,
									  gchar *search_text)
{
	session->priv->has_spawned_search = TRUE;
	
	return xesam_g_search_new_from_text (session, search_text);
}

/* Callback invoked when the searcher has closed our session 
 *
 * IMPORTANT: This is a callback from a XesamGSearcher which is an untrusted
 *            source. We should be careful to check args */
static void
_closed_cb (XesamGSearcher  *searcher,
			gpointer		data,
			GError			*error)
{
	XesamGSession   *session;
	
	g_return_if_fail (XESAM_IS_G_SEARCHER(searcher));
	g_return_if_fail (XESAM_IS_G_SESSION(data));
	
	session = XESAM_G_SESSION (data);
	
	if (error) {
		g_warning ("Error closing session: %s", error->message);
		_set_last_error (session, error);
		return;
	}
	
	session->priv->closed = TRUE;
	session->priv->ready = FALSE;
	
	xesam_g_debug_object (session, "Closed");
	
	g_signal_emit (session, g_session_signals[CLOSED], 0);
}

static void
_set_new_error (XesamGSession	*session,
				gint			code,
				const gchar		*format,
				...)
{	
	GError	*error;
	va_list args;
	gchar   *message;
	
	g_return_if_fail (XESAM_IS_G_SESSION(session));
	
	va_start (args, format);
	g_vasprintf (&message, format, args);
	va_end (args);
	
	/* Yes, we use searcher error codes */
	error = g_error_new (XESAM_G_SEARCHER_ERROR, code, message);
	
	g_free (message);
	_set_last_error (session, error);
	
}


/* Update session's last_error, free the incoming one,
 * and emit an "error" signal */
static void 
_set_last_error	(XesamGSession  *session,
				 GError			*error)
{
	XesamGSessionPrivate	*priv;
	
	g_return_if_fail (XESAM_IS_G_SESSION(session));
	g_return_if_fail (error != NULL);
	
	priv = session->priv;
	
	if (priv->last_error)
		g_error_free (priv->last_error);
	priv->last_error = g_error_copy (error);
	g_error_free (error);
	
	g_signal_emit (session, g_session_signals[ERROR], 0);
}
