/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2007 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include "xesam-g-hits.h"
#include "xesam-g-sparse-hits.h"
#include "xesam-g-search.h"
#include "xesam-g-globals-private.h"
#include "xesam-g-utils.h"
#include "xesam-g-debug-private.h"

/**
 * SECTION:xesam-g-sparse-hits
 * @short_description: Package private container used to model a collection of non-sequential hits
 *
 * Container for conveniently handling a collection of non-sequential #XesamGHit
 * objects.
 *
 * #XesamGSparseHits is available since 1.0
 */

struct _XesamGSparseHitsPrivate {
	GArray		*id_map;
	GHashTable	*field_map;
	guint		last_id_offset;
	guint		last_id;
};

/* Private methods */

static guint		xesam_g_sparse_hits_get_hit_id_real	(XesamGHits *base,
														 guint		offset);

static XesamGHit*	xesam_g_sparse_hits_next_real	 	(XesamGHits *base,
														 XesamGHit	*hit);

G_DEFINE_TYPE (XesamGSparseHits, xesam_g_sparse_hits, XESAM_TYPE_G_HITS);

static void
xesam_g_sparse_hits_init (XesamGSparseHits *hits)
{
	XesamGSparseHitsPrivate	*priv;
	
	priv = g_new0 (XesamGSparseHitsPrivate, 1);
	priv->last_id_offset = 0;
	priv->last_id = 0;
	
	hits->priv = priv;
}

static void
xesam_g_sparse_hits_finalize (GObject *object)
{
	XesamGSparseHits		*hits;
	XesamGSparseHitsPrivate	*priv;

	hits = XESAM_G_SPARSE_HITS (object);
	priv = hits->priv;
	
	g_array_free (priv->id_map, TRUE);
	g_hash_table_unref (priv->field_map);
	
	g_free (priv);
	
	G_OBJECT_CLASS (xesam_g_sparse_hits_parent_class)->finalize (object);
}

static void
xesam_g_sparse_hits_class_init (XesamGSparseHitsClass *klass)
{
	GObjectClass	*object_class = G_OBJECT_CLASS (klass);
	XesamGHitsClass	*hits_class = XESAM_G_HITS_CLASS (klass);

	object_class->finalize = xesam_g_sparse_hits_finalize;
	
	hits_class->get_hit_id = xesam_g_sparse_hits_get_hit_id_real;
	hits_class->next = xesam_g_sparse_hits_next_real;
}

static guint
xesam_g_sparse_hits_get_hit_id_real (XesamGHits	*base,
									 guint		offset)
{
	XesamGSparseHits	*hits;
	
	g_return_val_if_fail (XESAM_IS_G_SPARSE_HITS (base), 0);
	
	/* Bounds checking of @offset done in Hits constructor */
	
	hits = XESAM_G_SPARSE_HITS (base);
	
	return g_array_index (hits->priv->id_map, guint, offset);
}

static XesamGHit*
xesam_g_sparse_hits_next_real (XesamGHits	*base,
							   XesamGHit	*hit)
{	
	XesamGSparseHitsPrivate	*priv;
	XesamGSparseHits		*hits;
	XesamGHit				*next_hit;
	guint					hit_id;
	int						offset;
	
	g_return_val_if_fail (XESAM_IS_G_SPARSE_HITS (base), NULL);
	
	if (hit == NULL) {
		offset = 0;
		goto found;
	}
	
	g_return_val_if_fail (XESAM_IS_G_HIT (hit), NULL);
	
	hits = XESAM_G_SPARSE_HITS (base);
	priv = hits->priv;
	hit_id = xesam_g_hit_get_id (hit);
	next_hit = NULL;
	
	/* If we have matching hit id offset cached then use it to look up
	 * the next hit */
	if (priv->last_id == hit_id) {
		offset = priv->last_id_offset + 1;
		
		/* Check if we've reached the end of the list */
		if (offset >= xesam_g_hits_get_count (base)) 
			goto fail;
		
		goto found;
	}
	
	/* We don't have a matching hit id offset cached. Find it */
	for (offset = 0; offset < xesam_g_hits_get_count(base); offset++) {
		if (g_array_index (priv->id_map, guint, offset) == hit_id)
			break;
	}
	
	/* If we did not find the id, the user passed a hit that is not in this
	 * SparseHits object. Warn and return NULL */
	if (offset == xesam_g_hits_get_count(base)) {
		g_warning ("In call to xesam_g_hits_next() the passed Hit object is not"
				   " found in the SparseHits object");
		goto fail;
	}
	
	/* Check if the next offset is available. Otherwise we've reached the
	 * end of the list */
	++offset;
	if (offset >= xesam_g_hits_get_count (base))
		goto fail;
	
	goto found;
	
	found:
		next_hit = xesam_g_hits_get (base, offset);
		priv->last_id_offset = offset;
		priv->last_id = xesam_g_hit_get_id (next_hit);
		return next_hit;
	
	fail:
		priv->last_id = 0;
		priv->last_id_offset = 0;
		return NULL;
}

/**
 * xesam_g_sparse_hits_new
 * @search: Search to create a hits object for
 * @id_map: An array mapping each index to a hit id. Ie. the contents of the
 *         array are the hit ids.
 * @hits_data: An array of %GPtrArray<!-- -->s with hit data as per
 *             #XesamGSession:hit-fields
 * @user_data: User supplied data to associate with this hits object
 * @user_data_destroy: Called to free @user_data when the hits object is finalized.
 *                     Set to %NULL to ignore
 * @returns: A newly allocated sparse hits object
 *
 * Package private method declared in xesam-g-globals-private.h
 *
 * The SparseHits object is assumed to take ownership of the passed
 * %GArray of id maps and the %GPtrArray containing the hit data.
 */
XesamGHits*
xesam_g_sparse_hits_new (XesamGSearch	*search, 
						 GArray			*id_map,
						 GHashTable		*field_map,
						 GPtrArray		*hits_data,
						 gpointer		user_data,
						 GDestroyNotify user_data_destroy)
{	
	XesamGSparseHits	*hits;
	
	g_return_val_if_fail (XESAM_IS_G_SEARCH (search), NULL);
	g_return_val_if_fail (id_map != NULL, NULL);
	g_return_val_if_fail (field_map != NULL, NULL);
	g_return_val_if_fail (hits_data != NULL, NULL);
	g_return_val_if_fail (id_map->len == hits_data->len, NULL);
	
	hits = XESAM_G_SPARSE_HITS (g_object_new (XESAM_TYPE_G_SPARSE_HITS,
											  "search", search,
											  "field-map", field_map,
											  "hits_data", hits_data,
											  "user-data", user_data,
											  "user-data-destroy", user_data_destroy,
											  NULL));
	hits->priv->id_map = id_map;
		
	/* Our parent class XesamGHits will ref the search object and the field_map */
	
	return hits;
}

