/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2008 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <math.h>

#include "xesam-g-user-search-parser.h"
#include "xesam-g-query-builder.h"
#include "xesam-g-debug-private.h"
#include "xesam-g-utils.h"

typedef enum {
	TOKEN_TYPE_INVALID, /* Used to flag non-tokens */
	TOKEN_TYPE_PHRASE,  /* "bad wolf"c */
	TOKEN_TYPE_SELECT,  /* tag:hacking */
	TOKEN_TYPE_WORD,    /* foo */
	TOKEN_TYPE_COLLECT, /* AND or OR */
	TOKEN_TYPE_CLOSE    /* Marks end of clause or query */
} TokenType;

typedef struct _Token Token;

struct _Token {
	TokenType			type;
	XesamGQueryToken	type_detail;
	const gchar			*text;
	guint				text_len;
	const gchar			*extra; /* For phrase extra contains the modifiers, for selects it contains the relation */
	guint				extra_len;
	Token				*select_value;
	gboolean			negate;
};

struct _XesamGUserSearchParserPrivate {
	GSList					*builders;
	GSList					*tokens;
	GSList					*query;
	const gchar				*text;
	gboolean				emit_query_tokens;
};

static void		reset_parser			(XesamGUserSearchParser	*parser);

static Token*	token_new				(TokenType				type,
										 XesamGQueryToken		type_detail,
										 const gchar			*text,
										 guint					text_len,
										 const gchar			*extra,
										 guint					extra_len,
										 gboolean				negate);

static void		token_destroy			(Token					*tok);

static gboolean	tokenize				(XesamGUserSearchParser	*self,
										 GError					**error);

static gboolean	build_query				(XesamGUserSearchParser	*self,
										 GError					**error);

static gboolean	notify_builders			(XesamGUserSearchParser	*self,
										 GError					**error);

static const gchar*
				get_token_type_name		(TokenType				type);

enum  {
	PROP_0,
	
	PROP_EMIT_QUERY_TOKENS
};


static gpointer xesam_g_user_search_parser_parent_class = NULL;

/**
 * SECTION:xesam-g-user-search-parser
 * @short_description: A general purpose parser for the
 *                     Xesam User Search Language
 *
 * #XesamGUserSearchParser is a general purpose parser for the Xesam User Search
 * Language version 1.0.
 * 
 * To actually build a query that is machine understandable a
 * #XesamGQueryBuilder should be implemented by the consumer and added to the
 * parser via xesam_g_user_search_parser_add_builder().
 *
 * #XesamGUserSearchParser is available since 1.0
 */

/**
 * xesam_g_user_search_parser_add_builder
 * @self: The parser to add the builder to
 * @builder: The builder to add
 *
 * Add a #XesamGQueryBuilder to a user search parser. When the parser has build
 * a query from the user string the relevant methods on all registered builders
 * will be invoked. More than one builder may be registered via this function.
 *
 * The consumer need to add a way to retrieve the compiled query from @builder
 * itself.
 */
void
xesam_g_user_search_parser_add_builder (XesamGUserSearchParser	*self,
										XesamGQueryBuilder		*builder)
{
	g_return_if_fail (XESAM_IS_G_USER_SEARCH_PARSER (self));
	g_return_if_fail (XESAM_IS_G_QUERY_BUILDER (builder));
	
	self->priv->builders = g_slist_prepend (self->priv->builders, builder);
	g_object_ref (builder);
}

/**
 * xesam_g_user_search_parser_parse_string
 * @self: The parser to use
 * @user_search: A %NULL-terminated string in the Xesam UserSearch Language
 * @error: An optional place to store a #GError or %NULL to ignore errors
 *
 * Parse a %NULL-terminated string formatted in the Xesam 1.0
 * User Search Language. Returns %TRUE if and only if the parsing
 * complete without errors.
 *
 * When this method returns succesfully the compiled query can be retrieved
 * from the #XesamGQueryBuilder<!-- -->s installed with
 * xesam_g_user_search_parser_add_builder(). The application consuming the
 * parser needs to add a way to retrieve the query from the builder it self.
 *
 * Returns: %TRUE if and only if the parsing completed without errors
 */
gboolean
xesam_g_user_search_parser_parse_string (XesamGUserSearchParser	*self,
										 const char				*text,
										 GError					**error)
{
	XesamGUserSearchParserPrivate	*priv;
	gboolean						result;
	
	g_return_val_if_fail (XESAM_IS_G_USER_SEARCH_PARSER (self), FALSE);
	g_return_val_if_fail (text != NULL, FALSE);
	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
	
	xesam_g_debug_object (self, "Parsing string:\n%s", text);
	
	priv = self->priv;
	
	if (priv->text) {
		g_set_error (error, XESAM_G_USER_SEARCH_PARSER_ERROR,
					 XESAM_G_USER_SEARCH_PARSER_ERROR_CONCURRENT_PARSE,
					 "Got concurrent parse request");
		return FALSE;
	}
	
	reset_parser (self);
	priv->text = text;
	
	/* code here */
	result = tokenize (self, error);
	if (!result) return result;
	
	result = build_query (self, error);
	if (!result) return result;
	
	result = notify_builders (self, error);
	if (!result) return result;
	
	priv->text = NULL;
	return result;
}


GQuark
xesam_g_user_search_parser_error_quark (void)
{
	return g_quark_from_static_string ("xesam_g_user_search_parser_error-quark");
}


XesamGUserSearchParser*
xesam_g_user_search_parser_new (void)
{
	XesamGUserSearchParser * self;
	self = g_object_newv (XESAM_TYPE_G_USER_SEARCH_PARSER, 0, NULL);
	
	xesam_g_debug_object (self, "Created");
	
	return self;
}

static void
xesam_g_user_search_parser_set_property (GObject 		*object,
										 guint 			prop_id, 
										 const GValue 	*value,
										 GParamSpec 	*pspec)
{
	XesamGUserSearchParserPrivate	*priv;
	
	g_return_if_fail (XESAM_IS_G_USER_SEARCH_PARSER (object));
	
	priv = XESAM_G_USER_SEARCH_PARSER(object)->priv;
	
	switch (prop_id)
	{
	case PROP_EMIT_QUERY_TOKENS:
		priv->emit_query_tokens = g_value_get_boolean (value);
		break;	
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		return;
		break;
	}
}

static void
xesam_g_user_search_parser_get_property (GObject	*object,
										 guint		prop_id,
										 GValue		*value,
										 GParamSpec	*pspec)
{
	XesamGUserSearchParserPrivate	*priv;
	
	g_return_if_fail (XESAM_IS_G_USER_SEARCH_PARSER (object));
	
	priv = XESAM_G_USER_SEARCH_PARSER(object)->priv;
	
	switch (prop_id)
	{
	case PROP_EMIT_QUERY_TOKENS:
		g_value_set_boolean (value, priv->emit_query_tokens);
		break;	
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		return;
		break;
	}
}

static void
xesam_g_user_search_parser_init (XesamGUserSearchParser * self)
{
	XesamGUserSearchParserPrivate *priv;
	
	priv = g_slice_new (XesamGUserSearchParserPrivate);
	
	priv->tokens = NULL;
	priv->query = NULL;
	priv->builders = NULL;
	priv->text = NULL;
	priv->emit_query_tokens = TRUE;
		
	self->priv = priv;
}

static void
xesam_g_user_search_parser_finalize (GObject *self)
{
	XesamGUserSearchParserPrivate	*priv;
	XesamGQueryBuilder				*builder;
	GSList							*iter;
	
	xesam_g_debug_object (self, "Finalizing");
	
	priv = XESAM_G_USER_SEARCH_PARSER(self)->priv;
	
	reset_parser (XESAM_G_USER_SEARCH_PARSER(self));
	
	for (iter = priv->builders; iter; iter = iter->next) {
		builder = XESAM_G_QUERY_BUILDER (iter->data); /* We don't need this, but it's here for easier debugging */
		g_object_unref (builder);
	}
	
	G_OBJECT_CLASS(xesam_g_user_search_parser_parent_class)->finalize (self);
}

static void
xesam_g_user_search_parser_class_init (XesamGUserSearchParserClass * klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);
	
	xesam_g_user_search_parser_parent_class = g_type_class_peek_parent (klass);
	
	object_class->finalize = xesam_g_user_search_parser_finalize;
	object_class->set_property = xesam_g_user_search_parser_set_property;
	object_class->get_property = xesam_g_user_search_parser_get_property;
	
	/**
	 * XesamGUserSearchParser:emit-query-tokens:
	 * 
	 * If %TRUE xesam_g_query_builder_start_query() and
	 * xesam_g_query_builder_close_query() will be invoked on the query builders
	 * installed with xesam_g_user_search_parser_add_builder().
	 *
	 * If this property is %FALSE only the tokens not of a query type will
	 * be emitted. This is useful for embedding queries from user search strings
	 * inside queries constructed from other sources (such as xml).
	 */
	g_object_class_install_property (object_class,
	                                 PROP_EMIT_QUERY_TOKENS,
	                                 g_param_spec_boolean ("emit-query-tokens",
														   "Emit query tokens",
														   "Whether or not to include the query tokens when invoking the query builders",
															TRUE,
															G_PARAM_READWRITE));
}

GType
xesam_g_user_search_parser_get_type (void)
{
	static GType xesam_g_user_search_parser_type_id = 0;
	
	if (G_UNLIKELY (xesam_g_user_search_parser_type_id == 0)) {
		static const GTypeInfo g_define_type_info = { 
			sizeof (XesamGUserSearchParserClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) xesam_g_user_search_parser_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (XesamGUserSearchParser),
			0,
			(GInstanceInitFunc) xesam_g_user_search_parser_init
		};
		
		xesam_g_user_search_parser_type_id = g_type_register_static (G_TYPE_OBJECT,
																	 "XesamGUserSearchParser",
																	 &g_define_type_info,
																	 0);
	}
	
	return xesam_g_user_search_parser_type_id;
}

static void
reset_parser (XesamGUserSearchParser	*self)
{
	GSList 							*iter;
	XesamGUserSearchParserPrivate	*priv;
	
	g_return_if_fail (XESAM_IS_G_USER_SEARCH_PARSER(self));
	
	priv = self->priv;
	
	/* Free any collected tokens */
	for (iter = priv->tokens; iter; iter = iter->next) {
		token_destroy ((Token*)iter->data);
	}
	
	for (iter = priv->query; iter; iter = iter->next) {
		token_destroy ((Token*)iter->data);
	}
	
	if (priv->tokens)
		g_slist_free (priv->tokens);
	priv->tokens = NULL;
	
	if (priv->query)
		g_slist_free (priv->query);
	priv->query = NULL;
	
	/* priv->text is also used to indicate that we are parsing, so clear it */
	priv->text = NULL;
}

static Token*
token_new (TokenType		type,
		   XesamGQueryToken	type_detail,
		   const gchar		*text,
		   guint			text_len,
		   const gchar		*extra,
		   guint			extra_len,
		   gboolean			negate)
{
	Token *tok;
	
	g_return_val_if_fail (text != NULL, NULL);
	
	tok = g_slice_new (Token);
	tok->type = type;
	tok->type_detail = type_detail;
	tok->text = text;
	tok->text_len = text_len;
	tok->extra = extra;
	tok->extra_len = extra_len;
	tok->select_value = NULL;
	tok->negate = negate;
	
	return tok;
}

static void
token_destroy (Token *tok)
{
	g_return_if_fail (tok != NULL);
	
	if (tok->select_value)
		token_destroy (tok->select_value);
	
	/* The strings are pointers into the search string, so don't free those */
	g_slice_free (Token, tok);
}

/* Skips past the next token if it is the relation part of a select */
static gboolean
check_relation (const gchar			*text,
				guint				pos,
				TokenType			*type,
				XesamGQueryToken	*type_detail,
				guint				*extra_start,
				guint				*extra_end)
{
	gchar		next;
	guint		i = pos;
	
	while (text[i] != '\0') {
		switch (text[i]) {
			//g_debug ("Check rel: %c", text[*pos]);
			case ' ':
				++i;
				//g_debug ("CHECK REL SKIP");
				break;
			case '<':
				next = text[i+1];
				*extra_start = i;
				if (next == '=') {
					*type_detail = XESAM_G_QUERY_TOKEN_LESS_THAN_EQUALS;
					i++;
				} else
					*type_detail = XESAM_G_QUERY_TOKEN_LESS_THAN;
				*extra_end = i;
				*type = TOKEN_TYPE_SELECT;
				return TRUE;
			case '>':
				next = text[i+1];
				*extra_start = i;
				if (next == '=') {
					*type_detail = XESAM_G_QUERY_TOKEN_GREATER_THAN_EQUALS;
					i++;
				} else
					*type_detail = XESAM_G_QUERY_TOKEN_GREATER_THAN;
				*extra_end = i;
				*type = TOKEN_TYPE_SELECT;
				return TRUE;
			case '=':
				*extra_start = i;
				*extra_end = i;
				*type_detail = XESAM_G_QUERY_TOKEN_EQUALS;
				*type = TOKEN_TYPE_SELECT;
				return TRUE;
			case ':':
				*extra_start = i;
				*extra_end = i;
				*type_detail = XESAM_G_QUERY_TOKEN_CONTAINS;
				*type = TOKEN_TYPE_SELECT;
				return TRUE;
			default:
				return FALSE;
		}
	}
	
	return FALSE;
}

/* Check if the string between mark and pos is a collector token and set
 * type accordingly */
static gboolean
check_collector (const gchar 		*text,
				 guint				mark,
				 guint				pos,
				 TokenType			*type,
				 XesamGQueryToken	*type_detail)
{
	/* FIXME: Should we should take i18n into account here? Or should the
	 * queries be preprocessed by the app? */
	
	switch (text[mark]) {
		case 'A':
		case 'a':
			if (pos == mark+2 &&
				(text[mark+1] == 'n' || text[mark+1] == 'N') &&
				(text[pos] == 'd' || text[pos] == 'D')) {
				*type = TOKEN_TYPE_COLLECT;
				*type_detail = XESAM_G_QUERY_TOKEN_AND;
				return TRUE;
			}
			break;
		case 'O':
		case 'o':
			if (pos == mark+1 &&
				(text[mark+1] == 'r' || text[mark+1] == 'R')) {
				*type = TOKEN_TYPE_COLLECT;
				*type_detail = XESAM_G_QUERY_TOKEN_OR;
				return TRUE;
			}
			break;
		case '&':
			if (pos == mark+1 && text[mark+1] == '&') {
				*type = TOKEN_TYPE_COLLECT;
				*type_detail = XESAM_G_QUERY_TOKEN_AND;
				return TRUE;
			}
			break;
		case '|':
			if (pos == mark+1 && text[mark+1] == '|') {
				*type = TOKEN_TYPE_COLLECT;
				*type_detail = XESAM_G_QUERY_TOKEN_OR;
				return TRUE;
			}
			break;
		default:
			return FALSE;
	}
	
	return FALSE;
}

/* Read until first non-character appears. Modifiers can only be ascii chars. */
static void
check_phrase_mods (const gchar		*text,
				   guint 			pos,
				   guint			*extra_start,
				   guint			*extra_end,
				   XesamGQueryToken	*type_detail)
{
	*extra_start = pos;
	while (text[pos] != '\0' && text[pos] >= 'a' && text[pos] <= 'z') {
		switch (text[pos]) {
			case 'r':
				*type_detail = XESAM_G_QUERY_TOKEN_REG_EXP;
				break;
			case 'p':
				*type_detail = XESAM_G_QUERY_TOKEN_PROXIMITY;
				break;
		}
		
		++pos;
	}
	
	/* If we have no modifiers we have extra_end == extra_start */
	*extra_end = pos;
}

/*
 * Macros used in tokenize()
 */

/* Read any number of blanks */
#define SKIP_BLANKS while (text[++pos] == ' ') ;//g_debug("SKIP");

/* Add a token to our list. We filter out all explicit ANDs, it is eaiser
 * to work with the tokens without them. Note that some code may depend
 * on us filtering out the ANDs at this level */
#define EMIT_TOKEN \
G_STMT_START{\
if (last_type == TOKEN_TYPE_SELECT)\
	type_detail = XESAM_G_QUERY_TOKEN_STRING;\
else if (type_detail == XESAM_G_QUERY_TOKEN_NONE)\
	type_detail = XESAM_G_QUERY_TOKEN_FULL_TEXT;\
if (type_detail != XESAM_G_QUERY_TOKEN_AND) {\
	token = token_new (type, type_detail, text+mark, pos - mark, text+extra_start, extra_end - extra_start, negate_token);\
	priv->tokens = g_slist_prepend (priv->tokens, token);\
}\
negate_token = FALSE;\
last_type = type;\
last_type_detail = type_detail;\
type = TOKEN_TYPE_INVALID;\
type_detail = XESAM_G_QUERY_TOKEN_NONE;\
mark = pos;\
}G_STMT_END

#define HAD_TWO_CHAR_SELECT() \
(last_type_detail == XESAM_G_QUERY_TOKEN_LESS_THAN_EQUALS ||\
 last_type_detail == XESAM_G_QUERY_TOKEN_GREATER_THAN_EQUALS)

#define INC_POS() \
G_STMT_START{\
	++pos;\
	extra_start = pos;\
	extra_end = pos;\
}G_STMT_END

/* Tokenize the current priv->text and store all tokens in priv->tokens */
/* FIXME: Add support for + and - operators from the USL */
static gboolean
tokenize (XesamGUserSearchParser	*self,
		  GError					**error)
{
	XesamGUserSearchParserPrivate *priv;
	Token				*token;
	guint				pos, mark;
	guint				extra_start, extra_end;
	const gchar			*text;
	gboolean			negate_token;
	TokenType			type, last_type;
	XesamGQueryToken	type_detail, last_type_detail;
	
	g_return_val_if_fail (XESAM_IS_G_USER_SEARCH_PARSER (self), FALSE);
	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
	g_return_val_if_fail (self->priv->text != NULL, FALSE);
	
	priv = self->priv;
	
	/* Initial setup */
	pos = 0;
	mark = 0;
	extra_start = 0;
	extra_end = 0;
	text = self->priv->text;
	negate_token = FALSE;
	type = TOKEN_TYPE_INVALID;
	last_type = TOKEN_TYPE_INVALID;
	type_detail = XESAM_G_QUERY_TOKEN_NONE;
	last_type_detail = XESAM_G_QUERY_TOKEN_NONE;
	
	/* Actual tokenization
	 * Note: 'continue' statements below does not affect the switches only
	 *       the while loop
	 */
	while (text[pos] != '\0') {
		//g_debug ("SCAN '%c' (%d)", text[pos], pos);
		switch (text[pos]) {
			case ' ':
				//g_debug ("CASE ' ' (%d)", pos);
				/* Ignore leading blanks */
				switch (type) {
					case TOKEN_TYPE_PHRASE:
						SKIP_BLANKS;
						break;
					case TOKEN_TYPE_INVALID:
						SKIP_BLANKS;
						mark = pos;
						continue; /* We're now at a non-blank, so continue iteration directly */
					case TOKEN_TYPE_WORD:
						/* See if a relation is next. In that case the token is a select */
						check_relation (text, pos, &type, &type_detail,
										&extra_start, &extra_end);
						if (type == TOKEN_TYPE_WORD)
							check_collector (text, mark, pos-1, &type, &type_detail);
						EMIT_TOKEN;
						
						/* We have to update the pos and marks and continue
						 * because we hit the extra check_relation() below
						 * otherrwise */
						pos = extra_end;
						INC_POS();
						mark = pos;
						continue;
					case TOKEN_TYPE_SELECT:
					case TOKEN_TYPE_COLLECT:
					case TOKEN_TYPE_CLOSE:
						g_set_error (error, XESAM_G_USER_SEARCH_PARSER_ERROR,
									 XESAM_G_USER_SEARCH_PARSER_ERROR_INTERNAL,
									 "Internal parser error when parsing '%s'. "
									 "Unexpected %s-token at position %d.",
									 text, get_token_type_name(type), pos);
						return FALSE;
						
				}
				break;
			case '"':
				//g_debug ("CASE '\"' (%d)", pos);
				switch (type) {
					case TOKEN_TYPE_PHRASE:
						/* This is the end of the phrase */
						check_phrase_mods (text, pos+1, &extra_start, &extra_end, &type_detail);
						EMIT_TOKEN;
						pos = extra_end - 1;
						INC_POS();
						mark = pos;
						continue;
					case TOKEN_TYPE_INVALID:
						mark = pos + 1;
						type = TOKEN_TYPE_PHRASE;
						break;
					case TOKEN_TYPE_WORD:
						/* A double-quote inside a w"ord, just include as part of word */
						break;
					case TOKEN_TYPE_SELECT:
					case TOKEN_TYPE_COLLECT:
					case TOKEN_TYPE_CLOSE:
						g_set_error (error, XESAM_G_USER_SEARCH_PARSER_ERROR,
									 XESAM_G_USER_SEARCH_PARSER_ERROR_INTERNAL,
									 "Internal parser error when parsing '%s'. "
									 "Unexpected double quotes in %s-state at "
									 "position %d.",
									 text, get_token_type_name(type), pos);
						return FALSE;
				}
				break;
			case '+' :
				if (type == TOKEN_TYPE_INVALID) {
					negate_token = FALSE;
					INC_POS();
					continue;
				}
			case '-' :
				if (type == TOKEN_TYPE_INVALID) {
					negate_token = TRUE;
					INC_POS();
					continue;
				}
			default:
				//g_debug ("CASE default '%c' (%d)", text[pos], pos);
				switch (type) {
					case TOKEN_TYPE_INVALID:
						if (text[pos] == '+') {
							/* Ignore '+' as it means AND, which is implicit */
							INC_POS();
							mark = pos;
						} else {
							type = TOKEN_TYPE_WORD;
							mark = pos;
							INC_POS();
						}
						continue;
					case TOKEN_TYPE_PHRASE:
					case TOKEN_TYPE_WORD:
						break;
					case TOKEN_TYPE_SELECT:
					case TOKEN_TYPE_COLLECT:
					case TOKEN_TYPE_CLOSE:
						g_set_error (error, XESAM_G_USER_SEARCH_PARSER_ERROR,
									 XESAM_G_USER_SEARCH_PARSER_ERROR_INTERNAL,
									 "Internal parser error when parsing '%s'. "
									 "Unexpected %s-state when reading character "
									 "at position %d.",
									 text, get_token_type_name(type), pos);
						return FALSE;
				}
				break;
		}
		
		/* See if we have hit the relation part of a select, so we don't mistake
		 * it for a WORD */
		if (check_relation (text, pos, &type, &type_detail,
							&extra_start, &extra_end)) {
			EMIT_TOKEN;
			pos = extra_end;
			mark = pos;
		}
		
		/* Inc. position */
		INC_POS();
		
	}
	
	/* Any non-parsed trailing stuff is just a word... */
	if (mark < pos) {
		type = TOKEN_TYPE_WORD;
		EMIT_TOKEN;
	}
	
	/* We have reversed the tokens because we a prepending them (for efficiency) */
	priv->tokens = g_slist_reverse (priv->tokens);
	
	/* Can you believe we made it through the switch-maze? */
	return TRUE;
}

static const gchar*
get_token_type_name (TokenType type)
{
	switch (type) {
		case TOKEN_TYPE_WORD: return "WORD";
		case TOKEN_TYPE_SELECT: return "SELECT";
		case TOKEN_TYPE_COLLECT: return "COLLECT";
		case TOKEN_TYPE_PHRASE: return "PHRASE";
		case TOKEN_TYPE_CLOSE : return "CLOSE";
		case TOKEN_TYPE_INVALID: return "INVALID";
	}
	return "ERROR";
}

#define NEW_CLOSE_TOKEN(close_type_detail) token_new (TOKEN_TYPE_CLOSE, close_type_detail, "", 0, NULL, 0, FALSE)
#define NEW_CLOSE_AND_TOKEN() NEW_CLOSE_TOKEN(XESAM_G_QUERY_TOKEN_AND)
#define NEW_CLOSE_OR_TOKEN() NEW_CLOSE_TOKEN(XESAM_G_QUERY_TOKEN_OR)
#define NEW_AND_TOKEN() token_new (TOKEN_TYPE_COLLECT, XESAM_G_QUERY_TOKEN_AND, "AND", 3, NULL, 0, FALSE)
#define NEW_OR_TOKEN() token_new (TOKEN_TYPE_COLLECT, XESAM_G_QUERY_TOKEN_OR, "OR", 2, NULL, 0, FALSE)
#define HANDLE_SELECT(_tok)\
G_STMT_START{\
if (_tok->type == TOKEN_TYPE_SELECT) {\
	if (iter->next) {\
		_tok->select_value = (Token*) iter->next->data;\
		iter = iter->next;\
	} else {\
		/* There is no value to select on. Force token into a word */\
		xesam_g_debug_object (self, "No select value. Forcing token into word");\
		tmp_tok->type = TOKEN_TYPE_WORD;\
		tmp_tok->type_detail = XESAM_G_QUERY_TOKEN_FULL_TEXT;\
	}\
}\
}G_STMT_END


static gboolean
build_query (XesamGUserSearchParser	*self,
			 GError					**error)
{
	XesamGUserSearchParserPrivate 	*priv;
	Token							*tok, *tmp_tok;
	Token							*root_tok;
	XesamGQueryToken				prev_type_detail;
	GSList							*iter, *prev;
	GSList							*query;
	
	g_return_val_if_fail (XESAM_IS_G_USER_SEARCH_PARSER (self), FALSE);
	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
	g_return_val_if_fail (self->priv->text != NULL, FALSE);
	
	priv = self->priv;
	query = NULL;
	
	if (!priv->tokens) {
		g_set_error (error, XESAM_G_USER_SEARCH_PARSER_ERROR,
					 XESAM_G_USER_SEARCH_PARSER_ERROR_EMPTY_QUERY,
					 "The search string '%s' maps to an empty query",
					 priv->text);
		return FALSE;
	}
	
	/* Run through the tokens and pass them to the builders.
	 * Inject any implicit AND tokens on the fly */
	prev_type_detail = XESAM_G_QUERY_TOKEN_NONE;
	for (iter = priv->tokens; iter; iter = iter->next) {
		tok = (Token*)iter->data;
		
		if (query == NULL) {
			query = g_slist_prepend (query, tok);
			prev = iter;
			HANDLE_SELECT(tok);
			continue;
		}
		
		root_tok = (Token*)query->data;
		
		/* Root token is an AND */
		if (root_tok->type_detail == XESAM_G_QUERY_TOKEN_AND) {
			/* If the head token is an AND it means that all tokens should
			 * be ANDed - meaning that we can just prepend the token to the
			 * outer AND collector - unless the token is an OR collector
			 * then we need to make that token the new query root */
			
			if (tok->type_detail == XESAM_G_QUERY_TOKEN_AND)
				g_critical ("Unexpected explicit AND collector");
			else if (tok->type_detail == XESAM_G_QUERY_TOKEN_OR) {
				/* Include the next token after the OR if it is there,
				 * if not just pretend that this token regular WORD */
				if (iter->next) {						
					/* If the next token is again a collector, force it 
					 * into a word */
					tmp_tok = (Token*) iter->next->data;
					iter = iter->next;
					if (tmp_tok->type == TOKEN_TYPE_COLLECT) {
						tmp_tok->type = TOKEN_TYPE_WORD;
						tmp_tok->type_detail = XESAM_G_QUERY_TOKEN_FULL_TEXT;
					}
					
					query = g_slist_prepend (query, tmp_tok);
					query = g_slist_prepend (query, NEW_CLOSE_OR_TOKEN());
					query = g_slist_append (query, tok);
				} else {
					tok->type = TOKEN_TYPE_WORD;
					tok->type_detail = XESAM_G_QUERY_TOKEN_FULL_TEXT;
				}
			} else {
				/* If the token is a select add the next token (if any) as 
				 * its value */
				HANDLE_SELECT(tok);
				query = g_slist_insert_before (query, query->next, tok);
			}
		}
		
		/* Root token is an OR */
		else if (root_tok->type_detail == XESAM_G_QUERY_TOKEN_OR) {
			if (tok->type_detail == XESAM_G_QUERY_TOKEN_AND)
				g_critical ("Unexpected explicit AND collector");
			else if (tok->type_detail == XESAM_G_QUERY_TOKEN_OR) {
				if (iter->next) {
					/* If the next token is again a collector, force it 
					 * into a word */
					tmp_tok = (Token*) iter->next->data;
					iter = iter->next;
					
					/* We need to check if the new token is a selector */
					HANDLE_SELECT(tmp_tok);
					
					if (tmp_tok->type == TOKEN_TYPE_COLLECT) {
						tmp_tok->type = TOKEN_TYPE_WORD;
						tmp_tok->type_detail = XESAM_G_QUERY_TOKEN_FULL_TEXT;
					}
					
					query = g_slist_insert_before (query, query->next, tmp_tok);
				} else {
					g_critical ("Unexpected end of OR collector");
				}
			} else {
				/* If the first child of the root OR is an AND, go into that.
				 * Else we should replace the first child with an AND
				 * containing that child and the new token */
				if (query->next) {
					tmp_tok = (Token*) query->next->data;
					
					HANDLE_SELECT(tok);
					
					if (tmp_tok->type_detail == XESAM_G_QUERY_TOKEN_AND) {
						query = g_slist_insert_before (query,
													   query->next->next,
													   tok);
					} else {
						GSList *tmp_list = NULL;
						tmp_list = g_slist_prepend (tmp_list, NEW_AND_TOKEN());
						tmp_list = g_slist_prepend (tmp_list, tmp_tok);
						tmp_list = g_slist_prepend (tmp_list, tok);
						tmp_list = g_slist_prepend (tmp_list,
													NEW_CLOSE_AND_TOKEN());
						tmp_list = g_slist_concat (tmp_list, query->next->next);
						query->next = tmp_list;
					}
				} else {
					g_critical ("Unexpected empty OR collector");
				}
			}
		}
		
		/* Root token is an WORD, PHRASE, or SELECT */
		else {
			/* We have a NON-collector as head. This means that there should
			 * be only this single token in the query. Surround it with an
			 * appropriate collector according to the token */
			if (query->next) {
				g_critical ("Multiple unnested selector tokens");
				continue;
			}
			
			HANDLE_SELECT(tok);
			
			/* Find out what to do with the token */
			Token *collector;
			if (tok->type_detail == XESAM_G_QUERY_TOKEN_AND) {
				g_critical ("Unexpected explicit AND when updating root WORD");
				continue;
			} else if (tok->type_detail == XESAM_G_QUERY_TOKEN_OR) {
				if (!iter->next) {
					/* Force this OR into a WORD */
					tok->type = TOKEN_TYPE_WORD;
					tok->type_detail = XESAM_G_QUERY_TOKEN_FULL_TEXT;
					collector = NEW_AND_TOKEN();
				} else {
					collector = tok;
					tok = (Token*)iter->next->data;
					iter = iter->next;
					
					/* We need to check if the new token is a selector */
					HANDLE_SELECT(tok);
					
					/* If next token is another collector force it into a WORD */
					if (tok->type == TOKEN_TYPE_COLLECT) {
						tok->type = TOKEN_TYPE_WORD;
						tok->type_detail = XESAM_G_QUERY_TOKEN_FULL_TEXT;
					}
				}
			} else
				collector = NEW_AND_TOKEN();
			
			/* Update query with what we learned above */
			query = g_slist_append (query,
									collector);
			query = g_slist_prepend (query, tok);
			query = g_slist_prepend (query, 
									 collector->type_detail == XESAM_G_QUERY_TOKEN_AND ?
									 NEW_CLOSE_AND_TOKEN() : NEW_CLOSE_OR_TOKEN());
			
		}
		
		prev = iter;
	}
	
	/* query is build reversed */
	query = g_slist_reverse (query);
	
	priv->query = query;
	g_slist_free (priv->tokens);
	priv->tokens = NULL;
	
	return TRUE;
}

static guint
count_words (const gchar *str, guint len)
{
	guint count;
	
	for (count = 0; count < len; ++str) {
		if (str[0] == ' ') ++count;
	}
	
	return count;
}

/* The maximum number of attributes to read. The Xesam USL defines 14 modifiers */
#define MAX_MODS 14

/* Fill in all string element attributes need with const strings. Also fill
 * in the boost string if needed */
static void
calc_phrase_attributes (Token		*tok,
						gchar		**attr_names,
						gchar		**attr_values,
						gchar		**boost)
{
	int		i;
	int		last_attr;
	double	slack_d;
	guint	slack;
	
	last_attr = 0;
	
	for (i = 0; i < tok->extra_len;  i++) {
		if (i >= MAX_MODS) {
			g_warning ("Skipping %d modifiers of %d",
					   tok->extra_len - i, tok->extra_len);
			return;
		} else if (tok->extra[i] == ' ') {
			g_critical ("Unexpected white space when reading phrase modifiers. "
						"This is a bug in xesam-glib");
			return;
		}
		
		
		/* Note : This switch omits 'p' and 'r' modifiers because they can not
		 * be represented as attributes. They require a whole other selector
		 * to be applied. This is handled in check_phrase_mods() */
		switch (tok->extra[i]) {
			case 'b' :
				attr_names[last_attr] = "boost";
				attr_values[last_attr] = "10.0";
				*boost = "10.0";
				++last_attr;
				break;
			case 'c' :
				attr_names[last_attr] = "caseSensitive";
				attr_values[last_attr] = "true";
				++last_attr;
				break;
			case 'C' :
				attr_names[last_attr] = "caseSensitive";
				attr_values[last_attr] = "false";
				++last_attr;
				break;
			case 'd' :
				attr_names[last_attr] = "diacriticSensitive";
				attr_values[last_attr] = "true";
				++last_attr;
				break;
			case 'D' :
				attr_names[last_attr] = "diacriticSensitive";
				attr_values[last_attr] = "false";
				++last_attr;
				break;
			case 'e' :
				/* 'e' is an alias for 'cdl' */
				attr_names[last_attr] = "caseSensitive";
				attr_values[last_attr] = "true";
				++last_attr;
				attr_names[last_attr] = "diacriticSensitive";
				attr_values[last_attr] = "true";
				++last_attr;
				attr_names[last_attr] = "enableStemming";
				attr_values[last_attr] = "false";
				++last_attr;
				break;
			case 'f' :
				attr_names[last_attr] = "fuzzy";
				attr_values[last_attr] = "0.40";
				++last_attr;
				break;
			case 'l' :
				attr_names[last_attr] = "enableStemming";
				attr_values[last_attr] = "false";
				++last_attr;
				break;
			case 'L' :
				attr_names[last_attr] = "enableStemming";
				attr_values[last_attr] = "true";
				++last_attr;
				break;
			case 'o' :
				attr_names[last_attr] = "ordered";
				attr_values[last_attr] = "true";
				++last_attr;
				break;
			case 'p':
			case 'r':
				/* Ignore. See above. */
				break;
			case 's':				
				slack_d = sqrt(count_words(tok->text, tok->text_len));
				slack = slack_d > 1.1 ? floor(slack_d) : floor(slack_d) - 1;
				slack = CLAMP (slack, 0, 9);
				attr_names[last_attr] = "slack";
				attr_values[last_attr] = digit_as_const_char (slack);
				++last_attr;
				break;
			case 'w' :
				attr_names[last_attr] = "wordBreak";
				attr_values[last_attr] = "true";
				++last_attr;
				break;
			default :
				g_debug ("Ignoring uknown phrase modifier '%c'", tok->extra[i]);
				break;
			
		}
	}
	
	attr_names[last_attr] = NULL;
	attr_values[last_attr] = NULL;
}

static gboolean
notify_builders (XesamGUserSearchParser	*self,
				 GError					**error)
{
	XesamGUserSearchParserPrivate	*priv;
	GSList							*iter, *biter;
	Token							*tok;
	gboolean						negate;
	gchar							*boost;
	gboolean						result;
	GString							*value;
	GString							*select_field;
	gchar							*attr_names[MAX_MODS + 1];
	gchar							*attr_values[MAX_MODS + 1];
	
	g_return_val_if_fail (XESAM_IS_G_USER_SEARCH_PARSER (self), FALSE);
	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
	
	priv = self->priv;
	value = g_string_sized_new (512);
	select_field = g_string_sized_new (128);
	
	if (priv->emit_query_tokens) {
		for (biter = priv->builders; biter; biter = biter->next) {
			xesam_g_query_builder_start_query (XESAM_G_QUERY_BUILDER(biter->data),
											   XESAM_G_QUERY_TOKEN_QUERY,
											   NULL,
											   NULL,
											   error);
			
			if (error && *error) {
				result = FALSE;
				goto cleanup;
			}
			
		}
	}
	
	for (iter = priv->query; iter; iter = iter->next) {
		tok = (Token*) iter->data;
		attr_names[0] = NULL;
		attr_values[0] = NULL;
		boost = NULL;
		
		if (!xesam_g_query_token_is_clause_type (tok->type_detail)) {
			g_critical ("When notifying query builders - Unexpected token "
						"'%s' clause type expected",
						xesam_g_query_token_get_name(tok->type_detail));
			return FALSE;
		}
		
		switch (tok->type) {
			case TOKEN_TYPE_SELECT:
				/* FIXME: What to do with 'type' selector of usl?" */
				
				if (!tok->select_value) {
					g_set_error (error, XESAM_G_USER_SEARCH_PARSER_ERROR,
								 XESAM_G_USER_SEARCH_PARSER_ERROR_INTERNAL,
								 "When notifying query builders - selector "
								 "token without a value. This is a bug in "
								 "xesam-glib");
						return FALSE;
				}
				
				/* If either the token or the select value is negated, then 
				 * we negate the whole clause */
				if (tok->select_value->negate)
					tok->negate = TRUE;
				
				if (tok->select_value->type == TOKEN_TYPE_PHRASE) {
					calc_phrase_attributes (tok->select_value,
											attr_names, attr_values, &boost);
				}
				
				g_string_assign (select_field, "xesam:");
				g_string_append_len (select_field, tok->text, tok->text_len);
				
				g_string_overwrite_len (value, 0, tok->select_value->text,
										tok->select_value->text_len);
				g_string_truncate (value, tok->select_value->text_len);
				
				for (biter = priv->builders; biter; biter = biter->next) {
					
					/* We need to assert that the token is not a fullText selector
					 * because fullText selectors does not have a field-type child 
					 * which we are going to add in a sec */
					if (tok->type_detail == XESAM_G_QUERY_TOKEN_FULL_TEXT) {
						g_set_error (error, XESAM_G_USER_SEARCH_PARSER_ERROR,
									 XESAM_G_USER_SEARCH_PARSER_ERROR_INTERNAL,
									 "When notifying query builders - found  "
									 "'fullText' selector in select token. This"
									 " is a bug in xesam-glib");
						return FALSE;
					}
					
					xesam_g_query_builder_add_clause (XESAM_G_QUERY_BUILDER(biter->data),
													  tok->type_detail,
													  boost,
													  &tok->negate,
													  NULL,
													  NULL,
													  error);
					
					if (error && *error) {
						result = FALSE;
						goto cleanup;
					}
					
					xesam_g_query_builder_add_field (XESAM_G_QUERY_BUILDER(biter->data),
													 XESAM_G_QUERY_TOKEN_FIELD,
													 select_field->str,
													 error);
					
					if (error && *error) {
						result = FALSE;
						goto cleanup;
					}
					
					/* FIXME: If we knew the ontology we could detect if the
					 * value type should be date, integer, or something other 
					 * than string */
					xesam_g_query_builder_add_value (XESAM_G_QUERY_BUILDER(biter->data),
													 XESAM_G_QUERY_TOKEN_STRING,
													 value->str,
													 (const gchar**)attr_names,
													 (const gchar**)attr_values,
													 error);
					
					if (error && *error) {
						result = FALSE;
						goto cleanup;
					}
					
					xesam_g_query_builder_close_clause (XESAM_G_QUERY_BUILDER(biter->data),
														error);
					
					if (error && *error) {
						result = FALSE;
						goto cleanup;
					}
				}
				break;
			case TOKEN_TYPE_PHRASE:
				calc_phrase_attributes (tok, attr_names, attr_values, &boost);
			case TOKEN_TYPE_WORD:
				g_string_overwrite_len (value, 0, tok->text, tok->text_len);
				g_string_truncate (value, tok->text_len);
				
				for (biter = priv->builders; biter; biter = biter->next) {
					
					xesam_g_query_builder_add_clause (XESAM_G_QUERY_BUILDER(biter->data),
													  tok->type_detail,
													  boost,
													  &tok->negate,
													  NULL,
													  NULL,
													  error);
					if (error && *error) {
						result = FALSE;
						goto cleanup;
					}
					
					/* If this is not a fullText clause then specify that we
					 * query fullTextFields. The most likely case is that the
					 * phrase mods 'p' or 'r' where used in which case this is
					 * a 'proximity' or 'regExp' clause */
					if (tok->type_detail != XESAM_G_QUERY_TOKEN_FULL_TEXT) {
						xesam_g_query_builder_add_field (XESAM_G_QUERY_BUILDER(biter->data),
														 XESAM_G_QUERY_TOKEN_FULL_TEXT_FIELDS,
														 NULL,
														 error);
						
						if (error && *error) {
							result = FALSE;
							goto cleanup;
						}
					}
					
					/* FIXME: If we knew the ontology we could detect if the
					 * value type should be date, integer, or something other 
					 * than string */
					xesam_g_query_builder_add_value (XESAM_G_QUERY_BUILDER(biter->data),
													 XESAM_G_QUERY_TOKEN_STRING,
													 value->str,
													 (const gchar**)attr_names,
													 (const gchar**)attr_values,
													 error);
					
					if (error && *error) {
						result = FALSE;
						goto cleanup;
					}
					
					xesam_g_query_builder_close_clause (XESAM_G_QUERY_BUILDER(biter->data),
														error);
					
					if (error && *error) {
						result = FALSE;
						goto cleanup;
					}
				}
				break;
			case TOKEN_TYPE_COLLECT:
				for (biter = priv->builders; biter; biter = biter->next) {
					
					if (!xesam_g_query_token_is_collector_type(tok->type_detail)) {
						g_warning ("When notifying query builders - expecting "
								   "collector type but got '%s'",
								   xesam_g_query_token_get_name(tok->type_detail));
						result = FALSE;
						goto cleanup;
					}
					
					xesam_g_query_builder_add_clause (XESAM_G_QUERY_BUILDER(biter->data),
													  tok->type_detail,
													  NULL,
													  NULL,
													  NULL,
													  NULL,
													  error);
					
					if (error && *error) {
						result = FALSE;
						goto cleanup;
					}
				}
				break;
			case TOKEN_TYPE_CLOSE:
				for (biter = priv->builders; biter; biter = biter->next) {
					xesam_g_query_builder_close_clause (XESAM_G_QUERY_BUILDER(biter->data),
														error);
				}
				
				if (error && *error) {
					result = FALSE;
					goto cleanup;
				}
				
				break;
			default:
				g_critical ("Bad token type '%s'", get_token_type_name(tok->type));
				result = FALSE;
				goto cleanup;
		}
	}
	
	if (priv->emit_query_tokens) {
		for (biter = priv->builders; biter; biter = biter->next) {
			xesam_g_query_builder_close_query (XESAM_G_QUERY_BUILDER(biter->data),											   
											   error);
			
			if (error && *error)
				return FALSE;
		}
	}
	
	cleanup:
		g_string_free (value, TRUE);
		g_string_free (select_field, TRUE);
	
	return result;
}
