/***************************************************************************
 *            xesam-g-utils.c
 *
 *  Thu Jul 26 21:22:10 2007
 *  Copyright  2007  Mikkel Kamstrup Erlandsen
 *  <mikkel.kamstrup@gmail.com>
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with main.c; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */

#include "xesam-g-utils.h"
#include "xesam-g-debug-private.h"

/*
 * Initialize a value if it points to NULL
 */
GValue*
init_value_if_null (GValue **value, GType value_type)
{
    if (*value == NULL) {
        *value = g_new0 (GValue, 1);
        g_value_init (*value, value_type);
    }
    
    return *value;
        
}

/**
 * xesam_g_clone_value
 * @orig: %GValue to clone
 *
 * Return value: A newly allocated and initialized %GValue with a copy of the
 *               contents of @orig.
 */
GValue*
xesam_g_clone_value (const GValue *orig)
{
	GValue  *clone;
	
	clone = g_new0 (GValue, 1);
	g_value_init (clone, G_VALUE_TYPE(orig));
	g_value_copy (orig, clone);
	
	return clone;
}

/**
 * g_property_to_xesam_property
 * @g_prop_name: A string formatted as a standard GObject property (or signal)
 *               delimited by '_' or '-'.
 *
 * Return value: A newly allocated string with '-' and '_'
 * converted to '.'. Primarily for computing xesam session property
 * names from gobject property names.
 */
gchar*
g_property_to_xesam_property (gchar *g_prop_name)
{
    gchar *prop_name;
    
    prop_name = g_strdup (g_prop_name);
	g_strdelimit (prop_name, "-_", '.');
    
    return prop_name;
}

static void
_free_value_func (GValue *value, gpointer data) {
    g_value_unset (value);
    g_free (value);
}

void
free_ptr_array_of_values (GPtrArray *array)
{
    g_ptr_array_foreach (array, (GFunc)_free_value_func, NULL);
    g_ptr_array_free (array, TRUE);
}

const gchar*
digit_as_const_char (guint			digit)
{
	switch (digit) {
		case 0 : return "0";
		case 1 : return "1";
		case 2 : return "2";
		case 3 : return "3";
		case 4 : return "4";
		case 5 : return "5";
		case 6 : return "6";
		case 7 : return "7";
		case 8 : return "8";
		case 9 : return "9";
		default :
			g_warning ("Not a digit '%d'", digit);
			return "0";
	}
}

/*
 * Calculate a new field map from an array of strings. The field map is just
 * hash from field names to uint offsets of the field in @fields.
 */
GHashTable*
calc_field_map (GStrv fields)
{
	GHashTable 				*map;
	guint					idx, length;
	
	g_return_val_if_fail (fields != NULL, NULL);
	
	map = g_hash_table_new_full ((GHashFunc) g_str_hash,
								 (GEqualFunc) g_str_equal,
								 (GDestroyNotify) g_free,
								 NULL);
	
	/* Copy (field_name, array_offset) into map */
	length = g_strv_length (fields);
	for (idx = 0; idx < length; idx ++) {
		xesam_g_debug ("Field map - %s = %d",
					   fields[idx], idx);
		g_hash_table_insert (map,
							 g_strdup (fields[idx]),
							 GUINT_TO_POINTER(idx));
	}
	
	return map;
}


/*
 * Calculate a new field map by looking up a GStrv property on an object
 */
GHashTable*
calc_field_map_from_property (GObject *object, const gchar *prop_name)
{
	GStrv		fields = NULL;
	GHashTable	*map;
	
	g_return_val_if_fail (G_IS_OBJECT(object), NULL);
	g_return_val_if_fail (prop_name != NULL, NULL);
	
	g_object_get (object, prop_name, &fields, NULL);
	
	g_return_val_if_fail (fields != NULL, NULL);
	map = calc_field_map (fields);
	
	return map;
}

/*
 * Look up the field map associated with an object. Returns NULL if there
 * is no field map.
 */
GHashTable*
get_field_map (GObject *object)
{
	g_return_val_if_fail (G_IS_OBJECT(object), NULL);
	
	return (GHashTable*) g_object_get_data (object,
											"xesam-glib-private:FieldMap");
}

/*
 * Set the field map associated with an object.
 * Any old field map will be unreffed.
 */
GHashTable*
set_field_map (GObject *object, GHashTable *field_map)
{
	g_return_val_if_fail (G_IS_OBJECT(object), NULL);
	g_return_val_if_fail (field_map != NULL, NULL);
	
	g_object_set_data_full (object,
							"xesam-glib-private:FieldMap",
							g_hash_table_ref (field_map),
							(GDestroyNotify)g_hash_table_unref);
	
	return field_map;
}

/*
 * Make sure @object has an attached field map. If it doesn't look one up on
 * @cache. If @cache is NULL or does not have one, calculate one from the
 * @prop_name property on @source.
 */
GHashTable*
ensure_field_map(GObject 		*object,
				 GObject 		*cache,
				 GObject 		*source,
				 const gchar	*prop_name)
{
	GHashTable *map;
	
	g_return_val_if_fail (G_IS_OBJECT (object), NULL);
	
	map = get_field_map (object);
	
	if (map == NULL) {
		if (cache != NULL)
			map = get_field_map (cache);
		if (map == NULL && source != NULL && prop_name != NULL)
			map = calc_field_map_from_property (source, prop_name);					
	} else
		return map;
	
	return set_field_map (object, map);
}

GQuark
xesam_g_error_quark (void)
{
  return g_quark_from_static_string ("xesam-g-error-quark");
}
