/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * xesam-glib
 * Copyright (C) Mikkel Kamstrup Erlandsen 2008 <mikkel.kamstrup@gmail.com>
 * 
 * xesam-glib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * xesam-glib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with xesam-glib.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include "xesam-glib/xesam-g-xml-query-builder.h"
#include "xesam-glib/xesam-g-query-token.h"
#include "xesam-glib/xesam-g-query-builder.h"
#include "xesam-glib/xesam-g-utils.h"
#include "xesam-glib/xesam-g-debug-private.h"

#define XML_LEAD "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
#define XML_LEAD_LEN 38

/**
 * SECTION:xesam-g-xml-query-builder
 * @short_description: A #XesamGQueryBuilder implementation capable of producing
 *                     and XML representation of the built query
 *
 * %XesamGXmlQueryBuilder implements a #XesamGQueryBuilder capable of producing
 * an XML representation of the built query. To obtain a copy of the built
 * string call xesam_g_xml_query_builder_get_xml().
 *
 * See also the example on <link linkend="build-search">building and executing
 * a query</link>.
 *
 * #XesamGXmlQueryBuilder is available since 1.0
 */

enum  {
	PROP_0,
	
	PROP_VALIDATING,
	PROP_FRAGMENT_ONLY
};

struct _XesamGXmlQueryBuilderPrivate {
	GString				*buffer;
	XesamGQueryToken	query_type;
	GSList				*clause_stack;
	gboolean			validating;
	gboolean			fragment_only;
};

/* Macro to access the emidiate parent clause token type, 
 * or XESAM_G_QUERY_TOKEN_NONE if the clause stack is empty */
#define ACTIVE_CLAUSE() (self->priv->clause_stack ? \
    GPOINTER_TO_UINT (self->priv->clause_stack->data) : XESAM_G_QUERY_TOKEN_NONE)

/* Pop the newest element from the clause stack */
#define CLAUSE_STACK_POP() \
self->priv->clause_stack = g_slist_delete_link (self->priv->clause_stack, \
												self->priv->clause_stack)

/* Put a token on top of the clause stack */
#define CLAUSE_STACK_PUSH(token) \
self->priv->clause_stack = g_slist_prepend (self->priv->clause_stack, \
											GUINT_TO_POINTER (token))

/* Evaluates to TRUE if there are no clauses on the stack */
#define CLAUSE_STACK_EMPTY() (self->priv->clause_stack == NULL)

/* Assert that there is a query type defined in the current context,
 * and throw a GError in case there isn't */
#define VALIDATE_QUERY_TYPE() \
G_STMT_START { \
	if (self->priv->validating && \
		self->priv->query_type == XESAM_G_QUERY_TOKEN_NONE) { \
		g_set_error (error, \
					 XESAM_G_QUERY_BUILDER_ERROR, \
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_PARENT, \
					 "Adding element to unstarted query"); \
		return FALSE; \
	} \
} G_STMT_END 

/* Throw a GError if @stmt is FALSE and the builder is validating */
#define VALIDATE_PARENT_CLAUSE(stmt, msg) \
G_STMT_START { \
	if (self->priv->validating && !(stmt)) { \
		g_set_error (error, \
					 XESAM_G_QUERY_BUILDER_ERROR, \
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_PARENT, \
					 "Invalid parent: "#msg); \
		return FALSE; \
	} \
} G_STMT_END 

static gboolean xesam_g_xml_query_builder_real_start_query (XesamGQueryBuilder	*self,
															XesamGQueryToken	token,
															const gchar			*content_cat,
															const gchar			*source_cat,
															GError				**error);

static gboolean xesam_g_xml_query_builder_real_add_clause (XesamGQueryBuilder	*self,
														   XesamGQueryToken		token,
														   const gchar			*boost,
														   const gboolean		*negate,
														   const gchar			**attr_names,
														   const gchar			**attr_vals,
														   GError				**error);

static gboolean xesam_g_xml_query_builder_real_add_field (XesamGQueryBuilder	*self,
														  XesamGQueryToken		token,
														  const char			*name,
														  GError				**error);

static gboolean xesam_g_xml_query_builder_real_add_value (XesamGQueryBuilder	*self,
														  XesamGQueryToken		token,
														  const gchar			*value,
														  const gchar			**attr_names,
														  const gchar			**attr_vals,
														  GError				**error);

static gboolean xesam_g_xml_query_builder_real_close_clause
														  (XesamGQueryBuilder	*self,
														   GError				**error);

static gboolean xesam_g_xml_query_builder_real_close_query
														  (XesamGQueryBuilder	*self,
														   GError				**error);

static gpointer xesam_g_xml_query_builder_parent_class = NULL;
static XesamGQueryBuilderIface* xesam_g_xml_query_builder_parent_iface = NULL;


static gboolean
xesam_g_xml_query_builder_real_start_query (XesamGQueryBuilder		*base,
											 XesamGQueryToken		token,
											 const gchar			*content_cat,
											 const gchar			*source_cat,
											 GError					**error)
{
	XesamGXmlQueryBuilder	*self;
	
	g_return_val_if_fail (XESAM_IS_G_XML_QUERY_BUILDER (base), FALSE);
	
	self = XESAM_G_XML_QUERY_BUILDER (base);
	
	/* Make sure this is a 'query' and not a 'userQuery' */
	if (token != XESAM_G_QUERY_TOKEN_QUERY) {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_TOKEN_TYPE,
					 "XMLQueryBuilder only supports the <query> element as"
					 " query root. Got '%s'",
					 xesam_g_query_token_get_name (token));
		return FALSE;
	}
	
	/* Reset string buffer */
	g_string_truncate (self->priv->buffer, 0);
	
	if (!self->priv->fragment_only) {
		g_string_append (self->priv->buffer,
						 XML_LEAD"<request xmlns=\"http://freedesktop.org/standards/xesam/1.0/query\">");
	} else {
		
	}
	
	/* Clear the clause stack */
	if (self->priv->clause_stack) {
		g_slist_free (self->priv->clause_stack);
		self->priv->clause_stack = NULL;
	}
		
	
	if (content_cat != NULL && source_cat != NULL)
		g_string_append_printf (self->priv->buffer,
								"<query content=\"%s\" source=\"%s\">",
								content_cat, source_cat);
	else if (content_cat == NULL && source_cat != NULL)
		g_string_append_printf (self->priv->buffer,
								"<query source=\"%s\">",
								source_cat);
	else if (content_cat != NULL && source_cat == NULL)
		g_string_append_printf (self->priv->buffer,
								"<query content=\"%s\">",
								content_cat);
	else
		g_string_append_printf (self->priv->buffer, "<query>");
		
	self->priv->query_type = XESAM_G_QUERY_TOKEN_QUERY;
		
	
	
	return TRUE;
}

static gboolean
xesam_g_xml_query_builder_real_add_clause (XesamGQueryBuilder		*base,
										   XesamGQueryToken			token,
										   const gchar				*boost,
										   const gboolean			*negate,
										   const gchar				**attr_names,
										   const gchar				**attr_vals,
										   GError					**error)
{
	XesamGXmlQueryBuilder	*self;
	int						i;
	
	g_return_val_if_fail (XESAM_IS_G_XML_QUERY_BUILDER (base), FALSE);
	
	self = XESAM_G_XML_QUERY_BUILDER (base);
	
	if (!xesam_g_query_token_is_clause_type (token)) {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_TOKEN_TYPE,
					 "Expected token of clause type, but got '%s'",
					 xesam_g_query_token_get_name (token));
		return FALSE;
	}
	
	VALIDATE_QUERY_TYPE();
	VALIDATE_PARENT_CLAUSE (CLAUSE_STACK_EMPTY() ||
							xesam_g_query_token_is_collector_type(ACTIVE_CLAUSE()),
							"Parent is not a query or collector clause");
	
	/* Append element name */
	g_string_append_printf (self->priv->buffer, "<%s",
							xesam_g_query_token_get_name(token));
	
	/* Append boost and negate */
	if (boost != NULL)
		g_string_append_printf (self->priv->buffer, " boost=\"%s\"", boost);
	
	if (negate != NULL)
		g_string_append_printf (self->priv->buffer, " negate=\"%s\"",
								*negate ? "true" : "false");
	
	/* Append any extra attributes */
	if (attr_names && attr_vals) {		
		for (i = 0; attr_names[i] != NULL; i++) {
			
			if (attr_vals[i] == NULL) {
				g_set_error (error,
							 XESAM_G_QUERY_BUILDER_ERROR,
							 XESAM_G_QUERY_BUILDER_ERROR_INVALID_ATTRIBUTE_LIST,
							 "Attribute '%s' at position %d with no value in"
							 " clause element '%s'",
							 attr_names[i], i, xesam_g_query_token_get_name (token));
				return FALSE;
			}
			
			/* Don't append boost or negate attributes if they are set
			 * directly in the method args */
			if ((g_str_equal (attr_names[i], "boost") && boost != NULL) ||
				(g_str_equal (attr_names[i], "negate") && negate != NULL))
				continue;
			
			g_string_append_printf (self->priv->buffer, " %s=\"%s\"",
									attr_names[i], attr_vals[i]);
		}
	} else if (attr_vals != NULL && attr_names == NULL) {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_ATTRIBUTE_LIST,
					 "Attribute values, but no attribute names in clause element"
					 "'%s'",
					 xesam_g_query_token_get_name (token));
		return FALSE;
	} else if (attr_vals == NULL && attr_names != NULL) {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_ATTRIBUTE_LIST,
					 "Attribute names, but no attribute values in clause element"
					 "'%s'",
					 xesam_g_query_token_get_name (token));
		return FALSE;
	}
	
	g_string_append (self->priv->buffer, ">");
							
	CLAUSE_STACK_PUSH(token);
	
	return TRUE;
}

static gboolean
xesam_g_xml_query_builder_real_add_field (XesamGQueryBuilder		*base,
										  XesamGQueryToken			token,
										  const char				*name,
										  GError					**error)
{
	XesamGXmlQueryBuilder	*self;
	
	g_return_val_if_fail (XESAM_IS_G_XML_QUERY_BUILDER (base), FALSE);
	
	self = XESAM_G_XML_QUERY_BUILDER (base);
	
	if (token == XESAM_G_QUERY_TOKEN_FIELD) {
		if (name)
			g_string_append_printf (self->priv->buffer,
									"<field name=\"%s\"/>", name);
		else {
			g_set_error (error,
						 XESAM_G_QUERY_BUILDER_ERROR,
						 XESAM_G_QUERY_BUILDER_ERROR_INVALID_FIELD,
						 "Field token with null-name");
			return FALSE;
		}
		
	} else if (token == XESAM_G_QUERY_TOKEN_FULL_TEXT_FIELDS) {
		g_string_append (self->priv->buffer, "<fullTextFields/>");
		if (name)
			g_warning ("Trying to add 'name' attribue to <fullTextFields/>. "
					   "This is not allowed");
	} else {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_TOKEN_TYPE,
					 "Expected token of field type, but got '%s'",
					 xesam_g_query_token_get_name (token));
		return FALSE;
	}
	
	return TRUE;
}

static gboolean
xesam_g_xml_query_builder_real_add_value (XesamGQueryBuilder		*base,
										   XesamGQueryToken			token,
										   const gchar				*value,
										   const gchar				**attr_names,
										   const gchar				**attr_vals,
										   GError					**error)
{
	XesamGXmlQueryBuilder	*self;
	int 					i;
	
	g_return_val_if_fail (XESAM_IS_G_XML_QUERY_BUILDER (base), FALSE);
	
	self = XESAM_G_XML_QUERY_BUILDER (base);
	
	if (!xesam_g_query_token_is_value_type (token)) {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_TOKEN_TYPE,
					 "Expected token of value type, but got '%s'",
					 xesam_g_query_token_get_name (token));
		return FALSE;
	}
	
	if (value == NULL) {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_VALUE,
					 "Value token '%s' with NULL value",
					 xesam_g_query_token_get_name (token));
		return FALSE;
	}
	
	/* Append element name */
	g_string_append_printf (self->priv->buffer, "<%s",
							xesam_g_query_token_get_name(token));
	
	/* Append attributes */
	if (attr_names && attr_vals) {		
		for (i = 0; attr_names[i] != NULL; i++) {
			
			if (attr_vals[i] == NULL) {
				g_set_error (error,
							 XESAM_G_QUERY_BUILDER_ERROR,
							 XESAM_G_QUERY_BUILDER_ERROR_INVALID_ATTRIBUTE_LIST,
							 "Attribute '%s' at position %d with no value in"
							 "value element '%s'",
							 attr_names[i], i, xesam_g_query_token_get_name (token));
				return FALSE;
			}
			
			g_string_append_printf (self->priv->buffer, " %s=\"%s\"",
									attr_names[i], attr_vals[i]);
		}
	} else if (attr_vals != NULL && attr_names == NULL) {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_ATTRIBUTE_LIST,
					 "Attribute values, but no attribute names in value element"
					 "'%s'",
					 xesam_g_query_token_get_name (token));
		return FALSE;
	} else if (attr_vals == NULL && attr_names != NULL) {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_ATTRIBUTE_LIST,
					 "Attribute names, but no attribute values in value element"
					 "'%s'",
					 xesam_g_query_token_get_name (token));
		return FALSE;
	}
	
	/* Close tag, append value, and append closing element */
	g_string_append_printf (self->priv->buffer, ">%s</%s>",
							value, xesam_g_query_token_get_name (token));
	
	return TRUE;
}

static gboolean
xesam_g_xml_query_builder_real_close_clause (XesamGQueryBuilder	*base,
											 GError			**error)
{
	XesamGXmlQueryBuilder	*self;
	XesamGQueryToken		token;
	
	g_return_val_if_fail (XESAM_IS_G_XML_QUERY_BUILDER (base), FALSE);
	
	self = XESAM_G_XML_QUERY_BUILDER (base);
	token = ACTIVE_CLAUSE();
	
	if (token == XESAM_G_QUERY_TOKEN_NONE) {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_PARENT,
					 "No parent clause to close");
		return FALSE;
	}
	
	/* token must be a clause, or else xesam-glib messed up... */		
	if (!xesam_g_query_token_is_clause_type (token)) {
		/* We log this error because we somehow messed up inside xesam-glib,
		 * ie. this is not a user error */
		g_critical ("Internal xesam-glib error: Clause stack contains non-clause"
					"token '%s'", xesam_g_query_token_get_name (token));
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_PARENT,
					 "Internal xesam-glib error: Clause stack contains non-clause");
		return FALSE;
	}
	
	g_string_append_printf (self->priv->buffer, "</%s>",
							xesam_g_query_token_get_name (token));
	
	CLAUSE_STACK_POP ();
	
	return TRUE;
}

static gboolean
xesam_g_xml_query_builder_real_close_query (XesamGQueryBuilder	*base,
											 GError				**error)
{
	XesamGXmlQueryBuilder	*self;
	
	g_return_val_if_fail (XESAM_IS_G_XML_QUERY_BUILDER (base), FALSE);
	
	self = XESAM_G_XML_QUERY_BUILDER (base);
	
	/* Check for token type correctness whether or not we are validating */
	if (!xesam_g_query_token_is_query_type (self->priv->query_type)) {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_PARENT,
					 "No query root defined when closing query");
		return FALSE;
	}
	
	/* If we are validating check that we have closed all clauses */
	if (self->priv->validating && !CLAUSE_STACK_EMPTY()) {
		g_set_error (error,
					 XESAM_G_QUERY_BUILDER_ERROR,
					 XESAM_G_QUERY_BUILDER_ERROR_INVALID_PARENT,
					 "Unclosed clause '%s' when closing query",
					 xesam_g_query_token_get_name (ACTIVE_CLAUSE()));
		return FALSE;
	}
	
	g_string_append_printf (self->priv->buffer, "</%s>",
							xesam_g_query_token_get_name (self->priv->query_type));
	
	if (!self->priv->fragment_only) {
		g_string_append (self->priv->buffer, "</request>");
	}
	
	return TRUE;
}


static void
xesam_g_xml_query_builder_finalize (GObject *obj)
{
	XesamGXmlQueryBuilderPrivate	*priv;
	
	xesam_g_debug_object (obj, "Finalizing");
	
	priv = XESAM_G_XML_QUERY_BUILDER(obj)->priv;
	
	g_string_free (priv->buffer, TRUE);
	
	if (priv->clause_stack)
		g_slist_free (priv->clause_stack);
	
	g_free (priv);
}

static void
xesam_g_xml_query_builder_set_property (GObject 		*object,
										guint 			prop_id, 
										const GValue 	*value,
										GParamSpec		*pspec)
{
	XesamGXmlQueryBuilderPrivate	*priv;
	
	g_return_if_fail (XESAM_IS_G_XML_QUERY_BUILDER (object));
	
	priv = XESAM_G_XML_QUERY_BUILDER(object)->priv;
	
	switch (prop_id)
	{
	case PROP_VALIDATING:
		priv->validating = g_value_get_boolean (value);
		break;
	case PROP_FRAGMENT_ONLY:
		priv->fragment_only = g_value_get_boolean (value);
		break;	
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		return;
		break;
	}
}

static void
xesam_g_xml_query_builder_get_property (GObject		*object,
										guint		prop_id,
										GValue		*value,
										GParamSpec	*pspec)
{
	XesamGXmlQueryBuilderPrivate	*priv;
	
	g_return_if_fail (XESAM_IS_G_XML_QUERY_BUILDER (object));
	
	priv = XESAM_G_XML_QUERY_BUILDER(object)->priv;
	
	switch (prop_id)
	{
	case PROP_VALIDATING:
		g_value_set_boolean (value, priv->validating);
		break;
	case PROP_FRAGMENT_ONLY:
		g_value_set_boolean (value, priv->fragment_only);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		return;
		break;
	}
}

static void
xesam_g_xml_query_builder_class_init (XesamGXmlQueryBuilderClass * klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);
	
	xesam_g_xml_query_builder_parent_class = g_type_class_peek_parent (klass);
	
	object_class->finalize     = xesam_g_xml_query_builder_finalize;
	object_class->set_property = xesam_g_xml_query_builder_set_property;
	object_class->get_property = xesam_g_xml_query_builder_get_property;
	
	/**
	 * XesamGXmlQueryBuilder:validating:
	 * 
	 * If %TRUE the query builder will use a weak form of query validation.
	 * Do not trust the builder to generate valid queries just because this
	 * switch is on. The builder only does checks that do not effect performance
	 * in a measurable way.
	 *
	 * Generally it is safe to leave the validation switch on. The prime
	 * reason for disabling it would be if a program wanted to generate snippets
	 * of queries to use elsewhere. See also the
	 * #XesamGXmlQueryBuilder:fragment-only property if you are interested in this.
	 */
	g_object_class_install_property (object_class,
	                                 PROP_VALIDATING,
	                                 g_param_spec_boolean ("validating",
														   "Validating",
														   "Whether or not the builder should perform a weak form of validation",
															TRUE,
															G_PARAM_READWRITE));
	
	/**
	 * XesamGXmlQueryBuilder:fragment-only
	 * 
	 * If %FALSE (the default) the query builder will add
	 * &lt;?xml version="1.0" encoding="utf-8"?&gt;&lt;request&gt;
	 * and a closing &lt;/request&gt; tag to the XML returned by the builder.
	 *
	 * Default is %FALSE. This is required for queries to be compatible with the
	 * Xesam XML Query Language.
	 *
	 * If you want to generate snippets of queries to use elsewhere you probably
	 * want to turn this switch to %TRUE and set the
	 * #XesamGXmlQueryBuilder:validating property to %FALSE
	 */
	g_object_class_install_property (object_class,
	                                 PROP_FRAGMENT_ONLY,
	                                 g_param_spec_boolean ("fragment-only",
														   "Fragment only",
														   "Whether or not to only return XML fragments",
															FALSE,
															G_PARAM_READWRITE));
}


static void
xesam_g_xml_query_builder_interface_init (XesamGQueryBuilderIface * iface)
{
	xesam_g_xml_query_builder_parent_iface = g_type_interface_peek_parent (iface);
	
	iface->start_query = xesam_g_xml_query_builder_real_start_query;
	iface->add_clause = xesam_g_xml_query_builder_real_add_clause;
	iface->add_field = xesam_g_xml_query_builder_real_add_field;
	iface->add_value = xesam_g_xml_query_builder_real_add_value;
	iface->close_clause = xesam_g_xml_query_builder_real_close_clause;
	iface->close_query = xesam_g_xml_query_builder_real_close_query;
}

static void
xesam_g_xml_query_builder_init (XesamGXmlQueryBuilder *self)
{	
	self->priv = g_new0 (XesamGXmlQueryBuilderPrivate, 1);
	
	self->priv->buffer = g_string_sized_new (2048);	
	self->priv->query_type = XESAM_G_QUERY_TOKEN_NONE;
	self->priv->clause_stack = NULL;
	self->priv->validating = TRUE;
	self->priv->fragment_only = FALSE;
	
	g_string_append (self->priv->buffer, XML_LEAD);
}

GType
xesam_g_xml_query_builder_get_type (void)
{
	static GType xesam_g_xml_query_builder_type_id = 0;
	
	if (G_UNLIKELY (xesam_g_xml_query_builder_type_id == 0)) {
		
		static const GTypeInfo g_define_type_info = {
			sizeof (XesamGXmlQueryBuilderClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) xesam_g_xml_query_builder_class_init,
			(GClassFinalizeFunc) NULL,
			NULL,
			sizeof (XesamGXmlQueryBuilder),
			0,
			(GInstanceInitFunc) xesam_g_xml_query_builder_init
		};
		
		static const GInterfaceInfo xesam_g_query_builder_info = {
			(GInterfaceInitFunc) xesam_g_xml_query_builder_interface_init,
			(GInterfaceFinalizeFunc) NULL,
			NULL
		};
		
		xesam_g_xml_query_builder_type_id = g_type_register_static (G_TYPE_OBJECT,
																	 "XesamGXmlQueryBuilder",
																	 &g_define_type_info,
																	 0);
		
		g_type_add_interface_static (xesam_g_xml_query_builder_type_id,
									 XESAM_TYPE_G_QUERY_BUILDER,
									 &xesam_g_query_builder_info);
	}
	
	return xesam_g_xml_query_builder_type_id;
}

/**
 * xesam_g_xml_query_builder_new
 * @returns:  a newly created xml query builder
 *
 * Create a new xml query builder
 */
XesamGXmlQueryBuilder*
xesam_g_xml_query_builder_new (void)
{
	XesamGXmlQueryBuilder	*self;
	
	self = g_object_new (XESAM_TYPE_G_XML_QUERY_BUILDER, NULL);
	
	return self;
}

/**
 * xesam_g_xml_query_builder_get_xml
 * @self: The query builder to extract xml from
 * @returns: a string representation of the query consisting of well formed xml.
 *           The returned string should be freed with g_free()
 *
 * Compile the xml representation of the build query.
 *
 * You may call this method on unfinished queries. In that case the returned
 * xml might not constitute a full query.
 */
gchar*
xesam_g_xml_query_builder_get_xml (XesamGXmlQueryBuilder *self)
{
	g_return_val_if_fail (XESAM_IS_G_XML_QUERY_BUILDER (self), NULL);
	
	return g_strdup (self->priv->buffer->str);
}

/**
 * xesam_g_xml_query_builder_get_query
 * @self: The query builder to extract a query from
 * @returns: A newly constructed #XesamGQuery representing the build query.
 *           Note that this is a floating reference since #XesamGQuery is a
 *			 subclass of #GInitiallyUnowned.
 *
 * You may call this method on unfinished queries. In that case the returned
 * query might not contain valid XML.
 *
 * Since the returned reference to the #XesamGQuery is floating you can pass
 * it on to a #XesamGSearch with xesam_g_session_new_search() without
 * g_object_unref()<!-- -->ing it. Like so:
 * |[
 * XesamGQuery           *query;
 * XesamGXmlQueryBuilder *builder;
 * XesamGSession         *session;
 *
 * // Instantiate builder and session. Build a query with the builder
 * 
 * xesam_g_session_new_search (session,
 *                             xesam_g_xml_query_builder_get_query(builder));
 * ]|
 */
XesamGQuery*
xesam_g_xml_query_builder_get_query (XesamGXmlQueryBuilder*	self)
{
	g_return_val_if_fail (XESAM_IS_G_XML_QUERY_BUILDER (self), NULL);
	
	return xesam_g_query_new_from_xml (self->priv->buffer->str);
}
